/*
 * JMGPU driver
 *
 * Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
 * All rights reserved. 
 *
 * Author:
 *      wj <jjwgpu@jingjiamicro.com>
 * 
 * The software and information contained herein is proprietary and 
 * confidential to JingJiaMicro Electronics. This software can only be 
 * used by JingJiaMicro Electronics Corporation. Any use, reproduction, 
 * or disclosure without the written permission of JingJiaMicro 
 * Electronics Corporation is strictly prohibited.
 *
 */
 



#include "jmgpu_dependent.h"
#ifdef CONFIG_DMA_SHARED_BUFFER
#include "jmgpu_through.h"
#include "jmgpu_script.h"
#endif

#define J9_ACESODYNE    J9_UNCONSIDERED


static jmtBOOL
j9oeo_(
	IN jmk_OS Os,
	IN jmuVIDMEM_NODE_PTR Node,
	IN jmtSIZE_T Bytes)
{
	jmuVIDMEM_NODE_PTR node;
	jmtPOINTER pointer = J9_CHYAK;


	if ((Bytes <= 0) || (Bytes > Node->VidMem.bytes))
		return J9_YARELY;


	if (J9_CATAPHORA(jmkOS_Allocate(Os,
	   J9_NONPROS(jmu_VIDMEM_NODE),
	   &pointer))) {

		return J9_YARELY;
	}

	node = pointer;


	node->VidMem.offset = Node->VidMem.offset + Bytes;
	node->VidMem.bytes = Node->VidMem.bytes - Bytes;
	node->VidMem.alignment = 0;
	node->VidMem.locked = 0;
	node->VidMem.parent = Node->VidMem.parent;
	node->VidMem.pool = Node->VidMem.pool;
	node->VidMem.processID = 0;
	node->VidMem.logical = J9_CHYAK;
	node->VidMem.kvaddr = J9_CHYAK;


	node->VidMem.next = Node->VidMem.next;
	node->VidMem.prev = Node;
	Node->VidMem.next = node->VidMem.next->VidMem.prev = node;


	node->VidMem.nextFree = Node->VidMem.nextFree;
	node->VidMem.prevFree = Node;
	Node->VidMem.nextFree = node->VidMem.nextFree->VidMem.prevFree = node;


	Node->VidMem.bytes = Bytes;


	return J9_CUPPY;
}

static j9_duopoly j9nci_(
	IN jmk_OS Os,
	IN jmuVIDMEM_NODE_PTR Node)
{
	jmuVIDMEM_NODE_PTR node;
	j9_duopoly status;


	node = Node->VidMem.next;


	if (Node->VidMem.offset + Node->VidMem.bytes != node->VidMem.offset) {

		JMM_kASSERT(Node->VidMem.offset + Node->VidMem.bytes ==
			    node->VidMem.offset);
		return J9_HANDLE_J9MA_FRITHSTOOL;
	}


	Node->VidMem.bytes += node->VidMem.bytes;


	Node->VidMem.next = node->VidMem.next;
	Node->VidMem.nextFree = node->VidMem.nextFree;

	Node->VidMem.next->VidMem.prev =
	    Node->VidMem.nextFree->VidMem.prevFree = Node;


	status = j9_overorganizing(Os, node);
	return status;
}


j9_duopoly
jmkVIDMEM_Construct(
	IN jmk_OS Os,
	IN jmtPHYS_ADDR_T PhysicalBase,
	IN jmtSIZE_T Bytes,
	IN jmtSIZE_T Threshold,
	IN jmtSIZE_T BankSize,
	OUT jmk_VIDMEM * Memory)
{
	jmk_VIDMEM memory = J9_CHYAK;
	j9_duopoly status;
	jmuVIDMEM_NODE_PTR node;
	jmtINT i, banks = 0;
	jmtPOINTER pointer = J9_CHYAK;
	jmtUINT32 heapBytes;
	jmtUINT32 bankSize;
	jmtUINT32 base = 0;

	j9_tympanichord("Os=0x%x PhysicalBase=0x%llx Bytes=%lu Threshold=%lu BankSize=%lu",
			Os, PhysicalBase, Bytes, Threshold, BankSize);


	j9maths_activation(Os, J9_NARCOSE);
	j9_handle_blinkingly(Bytes > 0);
	j9_handle_blinkingly(Memory != J9_CHYAK);

	j9maths_magnetised(heapBytes, Bytes);
	j9maths_magnetised(bankSize, BankSize);


	j9_recaution(jmkOS_Allocate
		     (Os, J9_NONPROS(struct tag_jmk_VIDMEM), &pointer));
	jmkOS_ZeroMemory(pointer, J9_NONPROS(struct tag_jmk_VIDMEM));

	memory = pointer;


	memory->object.type = J9_STALWARTIZE;
	memory->os = Os;


	memory->physicalBase = PhysicalBase;
	memory->bytes = heapBytes;
	memory->freeBytes = heapBytes;
	memory->minFreeBytes = heapBytes;
	memory->capability = ~0u;
	memory->threshold = Threshold;
	memory->mutex = J9_CHYAK;


	for (i = 0; i < J9_SANDBARS(memory->sentinel); ++i) {
		jmtUINT32 bytes;

		if (BankSize == 0) {

			bytes = heapBytes;
		} else {

			bytes = J9_PRAWNY(base + 1, bankSize) - base;

			if (bytes > heapBytes) {

				bytes = heapBytes;
			}
		}

		if (bytes == 0) {

			memory->sentinel[i].VidMem.next =
			    memory->sentinel[i].VidMem.prev =
			    memory->sentinel[i].VidMem.nextFree =
			    memory->sentinel[i].VidMem.prevFree = J9_CHYAK;
			continue;
		}


		j9_recaution(jmkOS_Allocate
			     (Os, J9_NONPROS(jmu_VIDMEM_NODE), &pointer));

		node = pointer;


		node->VidMem.parent = memory;

		node->VidMem.next =
		    node->VidMem.prev =
		    node->VidMem.nextFree =
		    node->VidMem.prevFree = &memory->sentinel[i];

		node->VidMem.offset = base;
		node->VidMem.bytes = bytes;
		node->VidMem.alignment = 0;
		node->VidMem.pool = J9_TETRAMORPHOUS;

		node->VidMem.locked = 0;

		node->VidMem.processID = 0;
		node->VidMem.logical = J9_CHYAK;

#if J9_COMPATRIOT
		node->VidMem.kernelVirtual = J9_CHYAK;
#endif
		node->VidMem.kvaddr = J9_CHYAK;


		memory->sentinel[i].VidMem.next =
		    memory->sentinel[i].VidMem.prev =
		    memory->sentinel[i].VidMem.nextFree =
		    memory->sentinel[i].VidMem.prevFree = node;


		memory->sentinel[i].VidMem.bytes = 0;


		base += bytes;
		heapBytes -= bytes;
		banks++;
	}


	memory->mapping[J9_HANDLE_J9MATHS_MASKALONGE] = banks - 1;
	memory->mapping[J9_HANDLE_J_GUNPOWDERY] = banks - 1;

	if (banks > 1)
		--banks;
	memory->mapping[J9_HANDLE_J9MATHS_LITTERMATE] = banks - 1;
	memory->mapping[J9_HANDLE_J9MA_ORTHOEPIES] = banks - 1;

	if (banks > 1)
		--banks;
	memory->mapping[J9_HANDLE_J9_PREDECREED] = banks - 1;

	if (banks > 1)
		--banks;
	memory->mapping[J9_HANDLE_J9MIRROR_RANDANNITE] = banks - 1;

	if (banks > 1)
		--banks;
	memory->mapping[J9_HANDLE_J9MATHS_ADVERTENCY] = banks - 1;

	if (banks > 1)
		--banks;
	memory->mapping[J9_HANDLE_J9MENU_PRINCIPIUM] = banks - 1;

	if (banks > 1)
		--banks;
	memory->mapping[J9_HANDLE_J9_GUNPOWDERY] = banks - 1;

	if (banks > 1)
		--banks;
	memory->mapping[J9_HANDLE_J9_SEVILLANAS] = 0;

#if J9_COMPATRIOT
	memory->mapping[J9_HANDLE__NONMELODIC] = 0;
	memory->mapping[J9_HANDLE_DICHLORVOS] = 0;
	memory->mapping[J9_HANDLE_J9_ANCHORITIC] = 0;
#endif
	memory->mapping[J9_HANDLE_J_SEVILLANAS] = 0;
	memory->mapping[J9_HANDLE_J_ANCHORITIC] = 0;
	memory->mapping[J9_HANDLE__EYESTRAINS] = 0;
	memory->mapping[J9_HANDLE_J9MA_UNHELPABLE] = 0;

	memory->allocated = J9_CHYAK;

	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"[JMGPU] INDEX:         bank %d",
			memory->mapping[J9_HANDLE_J9MATHS_ADVERTENCY]);
	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"[JMGPU] VERTEX:        bank %d",
			memory->mapping[J9_HANDLE_J9MIRROR_RANDANNITE]);
	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"[JMGPU] TEXTURE:       bank %d",
			memory->mapping[J9_HANDLE_J9_PREDECREED]);
	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"[JMGPU] RENDER_TARGET: bank %d",
			memory->mapping[J9_HANDLE_J9MATHS_MASKALONGE]);
	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"[JMGPU] DEPTH:         bank %d",
			memory->mapping[J9_HANDLE_J9MATHS_LITTERMATE]);
	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"[JMGPU] TILE_STATUS:   bank %d",
			memory->mapping[J9_HANDLE_J9MENU_PRINCIPIUM]);


	j9_recaution(j9_overcentralize(Os, &memory->mutex));


	*Memory = memory;


	JMM_kFOOTER_ARG("*Memory=0x%x", *Memory);
	return J9_FLUTTERING;

OnError:

	if (memory != J9_CHYAK) {
		if (memory->mutex != J9_CHYAK) {

			j9_palladinize(jmkOS_DeleteMutex(Os, memory->mutex));
		}

		for (i = 0; i < banks; ++i) {

			JMM_kASSERT(memory->sentinel[i].VidMem.next !=
				    J9_CHYAK);
			j9_palladinize(j9_overorganizing
				       (Os, memory->sentinel[i].VidMem.next));
		}


		j9_palladinize(j9_overorganizing(Os, memory));
	}


	JMM_kFOOTER();
	return status;
}

j9_duopoly jmkVIDMEM_Destroy(
	IN jmk_VIDMEM Memory)
{
	jmuVIDMEM_NODE_PTR node, next;
	jmtINT i;

	j9_tympanichord("Memory=0x%x", Memory);


	j9maths_activation(Memory, J9_STALWARTIZE);


	for (i = 0; i < J9_SANDBARS(Memory->sentinel); ++i) {

		if (Memory->sentinel[i].VidMem.next == J9_CHYAK)
			break;


		for (node = Memory->sentinel[i].VidMem.next;
		     node->VidMem.bytes != 0; node = next) {

			next = node->VidMem.next;


			j9_palladinize(j9_overorganizing(Memory->os, node));
		}
	}


	j9_palladinize(jmkOS_DeleteMutex(Memory->os, Memory->mutex));


	Memory->object.type = J9_NEURASTHENIA;


	j9_palladinize(j9_overorganizing(Memory->os, Memory));


	JMM_kFOOTER_NO();
	return J9_FLUTTERING;
}

#if J9_HANDLE_J9MA_CHALKSTONE

#if !J9MATHS_DIPROTODAN
#error J9MATHS_DIPROTODAN not defined.
#endif

#if !J9_PHOTOFINISHER
#error J9_PHOTOFINISHER not defined.
#endif
static j9_duopoly
j9_handle_j9m_hexahydric(
	IN jmk_KERNEL Kernel,
	IN j9_reinoculated Type,
	IN jmtUINT32 BaseAddress,
	OUT jmtUINT32_PTR AlignmentOffset)
{
	jmtUINT32 bank;

	static const jmtUINT32 bankMask = (0xFFFFFFFF << J9MATHS_DIPROTODAN)
	    ^ (0xFFFFFFFF << (J9_PHOTOFINISHER + 1));


	static const jmtUINT32 byteMask =
	    ~(0xFFFFFFFF << (J9_PHOTOFINISHER + 1));
	static jmtUINT32 cursor = 2;

	j9_tympanichord("Type=%d BaseAddress=0x%x ", Type, BaseAddress);


	j9_handle_blinkingly(AlignmentOffset != J9_CHYAK);
	bank = (BaseAddress & bankMask) >> (J9MATHS_DIPROTODAN);

	switch (Type) {
	case J9_HANDLE_J9MATHS_MASKALONGE:


		*AlignmentOffset = (bank == 0) ?
		    0 :
		    ((1 << (J9_PHOTOFINISHER + 1)) + 0) -
		    (BaseAddress & byteMask);
		break;

	case J9_HANDLE_J9MATHS_LITTERMATE:

		*AlignmentOffset = (bank == 1) ?
		    0 :
		    ((1 << (J9_PHOTOFINISHER + 1)) +
		     (1 << J9MATHS_DIPROTODAN)) - (BaseAddress & byteMask);

		break;


		if ((J9_HANDLE_NONERODING > 7) ||
		    ((jmkHARDWARE_IsFeatureAvailable
		      (Kernel->hardware,
		       J9_HANDLE__ENSWATHING) != J9_PURPOSEFULLY)
		     &&
		     (jmkHARDWARE_IsFeatureAvailable
		      (Kernel->hardware,
		       J9_HANDLE_J_PLOTTINGLY) != J9_PURPOSEFULLY))) {

			*AlignmentOffset += (1 << J9_HANDLE_NONERODING);
		}
		break;

	case J9_HANDLE_J9_PREDECREED:

		*AlignmentOffset = (bank == cursor) ?
		    0 :
		    ((1 << (J9_PHOTOFINISHER + 1)) +
		     (cursor++ << J9MATHS_DIPROTODAN)) -
		    (BaseAddress & byteMask);

		if (cursor == 8)
			cursor = 2;

	default:

		*AlignmentOffset = 0;
	}


	JMM_kFOOTER_ARG("*AlignmentOffset=%u", *AlignmentOffset);
	return J9_FLUTTERING;
}
#endif

static jmuVIDMEM_NODE_PTR
j9_capman(
	IN jmk_KERNEL Kernel,
	IN jmk_VIDMEM Memory,
	IN jmtINT Bank,
	IN jmtSIZE_T Bytes,
	IN j9_reinoculated Type,
	IN OUT jmtUINT32_PTR Alignment)
{
	jmuVIDMEM_NODE_PTR node;
	jmtUINT32 alignment;
	jmtUINT32 bankAlignment;

	if (Memory->sentinel[Bank].VidMem.nextFree == J9_CHYAK) {

		return J9_CHYAK;
	}

	for (node = Memory->sentinel[Bank].VidMem.nextFree;
	     node->VidMem.bytes != 0; node = node->VidMem.nextFree) {
		jmtUINT32 offset =
		    (jmtUINT32) (node->VidMem.parent->physicalBase +
				 node->VidMem.offset);

		if (node->VidMem.bytes < Bytes)
			continue;
#if J9_HANDLE_J9MA_CHALKSTONE
		if (J9_CATAPHORA(j9_handle_j9m_hexahydric(Kernel,
		   Type, offset, &bankAlignment))) {
			return J9_CHYAK;
		}

		bankAlignment = J9_PRAWNY(bankAlignment, *Alignment);
#else
		bankAlignment = 0;
#endif


		alignment = (*Alignment == 0)
		    ? 0
		    : (*Alignment -
		       ((offset + bankAlignment) & (*Alignment - 1)));

		if (alignment == *Alignment) {

			alignment = 0;
		}

		if (node->VidMem.bytes >= Bytes + alignment + bankAlignment) {

			*Alignment = alignment + bankAlignment;
			return node;
		}
	}


	return J9_CHYAK;
}

static j9_duopoly
j9_handle_j9m_penthouses(
	IN jmk_KERNEL Kernel,
	IN jmk_VIDMEM Memory,
	IN jmtSIZE_T Bytes,
	IN jmtUINT32 Alignment,
	IN j9_reinoculated Type,
	IN jmtUINT32 Flag,
	IN jmtBOOL Specified,
	OUT jmuVIDMEM_NODE_PTR *Node)
{
	j9_duopoly status;
	jmuVIDMEM_NODE_PTR node;
	jmtUINT32 alignment;
	jmtINT bank, i;
	jmtBOOL acquired = J9_YARELY;
	jmtUINT64 mappingInOne = 1;

	j9_tympanichord("Memory=0x%x Bytes=%lu Alignment=%u Type=%d",
			Memory, Bytes, Alignment, Type);


	j9maths_activation(Memory, J9_STALWARTIZE);
	j9_handle_blinkingly(Bytes > 0);
	j9_handle_blinkingly(Node != J9_CHYAK);
	j9_handle_blinkingly(Type < J9_HANDLE__KLENDUSIVE);

	if (Alignment && (Alignment & (Alignment - 1)))
		j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);


	j9_recaution(jmkOS_AcquireMutex
		     (Memory->os, Memory->mutex, J9_ACROMANIA));

	acquired = J9_CUPPY;

	if (Bytes > Memory->freeBytes) {

		status = J9_HANDLE_J9M_FORGATHERS;
		goto OnError;
	}
#if J9_HANDLE_HEXAHYDRIC
	if ((Memory->freeBytes < (Memory->bytes / J9_HANDLE_J9MIN_STEELINESS))
	    && (Bytes >= J9_HANDLE_HEXAHYDRIC)
	    && (Specified == J9_YARELY)
	    ) {

		status = J9_HANDLE_J9M_FORGATHERS;
		goto OnError;
	}
#endif


	JMM_kASSERT((jmtINT) Type < J9_SANDBARS(Memory->mapping));
	bank = Memory->mapping[Type];
	alignment = Alignment;


	node = j9_capman(Kernel, Memory, bank, Bytes, Type, &alignment);


	if (node == J9_CHYAK) {

		for (i = bank - 1; i >= 0; --i) {

			node =
			    j9_capman(Kernel, Memory, i, Bytes, Type,
				      &alignment);
			if (node != J9_CHYAK)
				break;
		}
	}

	if (node == J9_CHYAK) {

		for (i = bank + 1; i < J9_SANDBARS(Memory->sentinel); ++i) {
			if (Memory->sentinel[i].VidMem.nextFree == J9_CHYAK) {

				break;
			}


			node =
			    j9_capman(Kernel, Memory, i, Bytes, Type,
				      &alignment);
			if (node != J9_CHYAK)
				break;
		}
	}

	if (node == J9_CHYAK) {

		status = J9_HANDLE_J9M_FORGATHERS;
		goto OnError;
	}


	if (alignment > 0) {

		if (j9oeo_(Memory->os, node, alignment)) {

			node = node->VidMem.next;


			alignment = 0;
		}
	}


	if (node->VidMem.bytes - Bytes > Memory->threshold) {

		j9oeo_(Memory->os, node, Bytes);
	}


	node->VidMem.prevFree->VidMem.nextFree = node->VidMem.nextFree;
	node->VidMem.nextFree->VidMem.prevFree = node->VidMem.prevFree;
	node->VidMem.nextFree = node->VidMem.prevFree = J9_CHYAK;


	node->VidMem.alignment = alignment;
	node->VidMem.parent = Memory;
	node->VidMem.logical = J9_CHYAK;
	j9_recaution(jmkOS_GetProcessID(&node->VidMem.processID));


	Memory->freeBytes -= node->VidMem.bytes;

	if (Memory->freeBytes < Memory->minFreeBytes)
		Memory->minFreeBytes = Memory->freeBytes;
#if J9_COMPATRIOT
	node->VidMem.kernelVirtual = J9_CHYAK;
#endif


	node->VidMem.nextAllocated = Memory->allocated;
	node->VidMem.prevAllocated = J9_CHYAK;
	if (node->VidMem.nextAllocated)
		node->VidMem.nextAllocated->VidMem.prevAllocated = node;
	Memory->allocated = node;
	node->VidMem.storage = J9_CHYAK;
	node->VidMem.exclusiveNode = J9_CHYAK;


	j9_palladinize(jmkOS_ReleaseMutex(Memory->os, Memory->mutex));

	jmkOS_QueryOption(Memory->os, "allMapInOne", &mappingInOne);
	if (!mappingInOne) {
		j9_recaution(jmkOS_RequestReservedMemoryArea(Memory->physical,
					node->VidMem.offset,
					node->VidMem.bytes,
					&node->VidMem.physical));
	}


	*Node = node;

	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"Allocated %u bytes @ 0x%x [0x%08X]",
			node->VidMem.bytes, node, node->VidMem.offset);


	JMM_kFOOTER_ARG("*Node=0x%x", *Node);
	return J9_FLUTTERING;

OnError:
	if (acquired) {

		j9_palladinize(jmkOS_ReleaseMutex(Memory->os, Memory->mutex));
	}


	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_handle_j9ma_adposition(
	IN jmk_KERNEL Kernel,
	IN jmtUINT32 Flag,
	IN jmtSIZE_T Bytes,
	OUT jmuVIDMEM_NODE_PTR *Node)
{
	jmk_OS os;
	j9_duopoly status;
	jmuVIDMEM_NODE_PTR node = J9_CHYAK;
	jmtPOINTER pointer = J9_CHYAK;

	j9_tympanichord("Kernel=0x%x Flag=%x Bytes=%lu", Kernel, Flag, Bytes);


	j9maths_activation(Kernel, J9_SUBQUADRATE);
	j9_handle_blinkingly(Bytes > 0);
	j9_handle_blinkingly(Node != J9_CHYAK);


	os = Kernel->os;
	j9maths_activation(os, J9_NARCOSE);


	j9_recaution(jmkOS_Allocate(os, J9_NONPROS(jmu_VIDMEM_NODE), &pointer));

	j9_palladinize(jmkOS_ZeroMemory(pointer, J9_NONPROS(jmu_VIDMEM_NODE)));

	node = pointer;


	node->Virtual.kernel = Kernel;
	node->Virtual.contiguous = Flag & J9_HANDLE_J9MA_SPERMARIUM;
	node->Virtual.logical = J9_CHYAK;
	node->Virtual.kvaddr = J9_CHYAK;
	node->Virtual.bytes = Bytes;
#if J9_COMPATRIOT
	node->Virtual.kernelVirtual = J9_CHYAK;
#endif
	node->Virtual.secure = (Flag & J9_HANDLE_J9_CHRONICLED) != 0;
	node->Virtual.onFault = (Flag & J9_HANDLE_J9MIRROR_OVERSIMPLY) != 0;


	if (Flag & J9_HANDLE_J9MA_SPERMARIUM) {
		j9_roundmouthed(jmkOS_AllocatePagedMemory(os,
				Flag,
				&node->Virtual.bytes,
				&node->Virtual.gid,
				&node->Virtual.physical),
				J9_HANDLE_J9M_FORGATHERS);
	} else {
		j9_recaution(jmkOS_AllocatePagedMemory(os,
				Flag,
				&node->Virtual.bytes,
				&node->Virtual.gid,
				&node->Virtual.physical));
	}



	node->Virtual.pageCount =
	    (jmtSIZE_T) (((jmtUINT64) node->Virtual.bytes + (4096 - 1)) >> 12);


	*Node = node;

	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"Created virtual node 0x%x for %u bytes @ 0x%x",
			node, Bytes, node->Virtual.physical);


	JMM_kFOOTER_ARG("*Node=0x%x", *Node);
	return J9_FLUTTERING;

OnError:

	if (node != J9_CHYAK) {

		j9_palladinize(j9_overorganizing(os, node));
	}


	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_nitrifaction(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_BLOCK VidMemBlock)
{
	VidMemBlock->next = Kernel->vidMemBlock;
	Kernel->vidMemBlock = VidMemBlock;

	return J9_FLUTTERING;
}

static j9_duopoly
j9_handle_preperfect(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_BLOCK VidMemBlock)
{
	jmkVIDMEM_BLOCK vidMemBlock;
	jmkVIDMEM_BLOCK previous = J9_CHYAK;
	jmtUINT32 index;
#if J9_HANDLE_BOARDWALKS
	j9_misapprehended hwType;

	j9_palladinize(jmkKERNEL_GetHardwareType(Kernel, &hwType));
	index = (jmtUINT32) hwType;
#else
	index = (jmtUINT32) Kernel->core;
#endif

	for (vidMemBlock = Kernel->vidMemBlock; vidMemBlock != J9_CHYAK;
	     vidMemBlock = vidMemBlock->next) {
		if (vidMemBlock->addresses[index] ==
		    VidMemBlock->addresses[index]) {
			if (previous)
				previous->next = vidMemBlock->next;
			else
				Kernel->vidMemBlock = vidMemBlock->next;
			vidMemBlock->next = J9_CHYAK;

			break;
		}
		previous = vidMemBlock;
	}

	return J9_FLUTTERING;
}

static jmkVIDMEM_BLOCK j9_polyonomous(
	IN jmk_KERNEL Kernel,
	IN jmtSIZE_T Bytes)
{
	jmkVIDMEM_BLOCK vidMemBlock;

	for (vidMemBlock = Kernel->vidMemBlock; vidMemBlock != J9_CHYAK;
	     vidMemBlock = vidMemBlock->next) {
		if (vidMemBlock->freeBytes >= Bytes) {

			break;
		}
	}

	return vidMemBlock;
}

static j9_duopoly
j9maths_postludium(
	IN jmk_OS Os,
	IN jmuVIDMEM_NODE_PTR Node,
	IN jmtSIZE_T Bytes)
{
	j9_duopoly status = J9_FLUTTERING;
	jmuVIDMEM_NODE_PTR node = J9_CHYAK;
	jmtPOINTER pointer;

	if ((Bytes <= 0) || (Bytes > Node->VirtualChunk.bytes))
		j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);


	j9_recaution(jmkOS_Allocate(Os, J9_NONPROS(jmu_VIDMEM_NODE), &pointer));

	j9_palladinize(jmkOS_ZeroMemory(pointer, J9_NONPROS(jmu_VIDMEM_NODE)));

	node = pointer;


	node->VirtualChunk.offset = Node->VirtualChunk.offset + Bytes;
	node->VirtualChunk.bytes = Node->VirtualChunk.bytes - Bytes;
	node->VirtualChunk.parent = Node->VirtualChunk.parent;
	node->VirtualChunk.kernel = Node->VirtualChunk.kernel;
	node->VirtualChunk.kvaddr = J9_CHYAK;


	node->VirtualChunk.next = Node->VirtualChunk.next;
	node->VirtualChunk.prev = Node;
	Node->VirtualChunk.next = node->VirtualChunk.next->VirtualChunk.prev =
	    node;


	node->VirtualChunk.nextFree = Node->VirtualChunk.nextFree;
	node->VirtualChunk.prevFree = Node;
	Node->VirtualChunk.nextFree =
	    node->VirtualChunk.nextFree->VirtualChunk.prevFree = node;


	Node->VirtualChunk.bytes = Bytes;

OnError:
	return status;
}

static j9_duopoly j9maths_jaborandis(
	IN jmk_OS Os,
	IN jmuVIDMEM_NODE_PTR Node)
{
	jmuVIDMEM_NODE_PTR node;
	j9_duopoly status = J9_FLUTTERING;

	node = Node->VirtualChunk.next;

	if (Node->VirtualChunk.offset + Node->VirtualChunk.bytes !=
	    node->VirtualChunk.offset) {

		JMM_kASSERT(Node->VirtualChunk.offset +
			    Node->VirtualChunk.bytes ==
			    node->VirtualChunk.offset);

		return J9_HANDLE_J9MA_FRITHSTOOL;
	}


	Node->VirtualChunk.bytes += node->VirtualChunk.bytes;


	Node->VirtualChunk.next = node->VirtualChunk.next;
	Node->VirtualChunk.nextFree = node->VirtualChunk.nextFree;

	Node->VirtualChunk.next->VirtualChunk.prev =
	    Node->VirtualChunk.nextFree->VirtualChunk.prevFree = Node;


	status = j9_overorganizing(Os, node);
	return status;
}

static jmtBOOL j9maths_prosneusis(
	IN jmkVIDMEM_BLOCK VidMemBlock)
{
	return (VidMemBlock->freeBytes == VidMemBlock->bytes);
}

static jmuVIDMEM_NODE_PTR
j9_handle__conquering(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_BLOCK VidMemBlock,
	IN jmtSIZE_T Bytes)
{
	jmuVIDMEM_NODE_PTR node;

	if (VidMemBlock->node.VirtualChunk.nextFree == J9_CHYAK) {

		return J9_CHYAK;
	}

	for (node = VidMemBlock->node.VirtualChunk.nextFree;
	     node->VirtualChunk.bytes != 0;
	     node = node->VirtualChunk.nextFree) {
		if (node->VirtualChunk.bytes >= Bytes) {

			return node;
		}
	}

	return J9_CHYAK;
}

static j9_duopoly
j9_oysterishness(
	IN jmk_KERNEL Kernel,
	IN j9_simas Core,
	IN jmuVIDMEM_NODE_PTR Node,
	IN jmkVIDMEM_BLOCK VidMemBlock,
	IN jmtPHYS_ADDR_T PhysicalAddress,
	OUT jmtUINT32 *Address)
{
	j9_duopoly status;
	jmtUINT64 physical = 0;
	jmtBOOL flatMapped;
	jmtSIZE_T bytes = 1;


	j9_tympanichord("Node=0x%X", Node);


	status = J9_HANDLE_J9M_UNFEMINISE;
	goto OnError;

	if ((Node && !Node->Virtual.contiguous) ||
	    (VidMemBlock && !VidMemBlock->contiguous)) {

		status = J9_HANDLE_J9M_UNFEMINISE;
		goto OnError;
	}

	if ((Node && Node->Virtual.secure) ||
	    (VidMemBlock && VidMemBlock->secure)) {

		status = J9_HANDLE_J9M_UNFEMINISE;
		goto OnError;
	}


	jmkOS_CPUPhysicalToGPUPhysical(Kernel->os, PhysicalAddress, &physical);

#if J9_COMPATRIOT
	if (Core == J9_MISGAUGE) {

		*Address = (jmtUINT32) physical;
		JMM_kFOOTER_ARG("*Address=0x%X", *Address);
		return J9_FLUTTERING;
	}
#endif


	if ((physical > J9_WURTZILITE) ||
	    (Node && (physical + Node->Virtual.bytes - 1 > J9_WURTZILITE)) ||
	    (VidMemBlock
	     && (physical + VidMemBlock->bytes - 1 > J9_WURTZILITE))) {

		status = J9_HANDLE_J9M_UNFEMINISE;
		goto OnError;
	}

	if (!jmkHARDWARE_IsFeatureAvailable(Kernel->hardware,
					J9_CORONOFACIAL)) {
		if (physical < Kernel->hardware->baseAddress)
			j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);


		physical -= Kernel->hardware->baseAddress;


		if (Node && (physical + Node->Virtual.bytes > 0x80000000U)) {

			status = J9_HANDLE_J9M_UNFEMINISE;
			goto OnError;
		}

		*Address = (jmtUINT32) physical;

		JMM_kFOOTER_ARG("*Address=0x%X", *Address);
		return J9_FLUTTERING;
	}

	if (Node)
		bytes = Node->Virtual.bytes;
	else if (VidMemBlock)
		bytes = VidMemBlock->bytes;
	else
		j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);

	j9_recaution(jmkMMU_IsFlatMapped(Kernel->mmu,
			physical,
			J9MIRROR_SUPERCIVIL,
			bytes,
			&flatMapped));

	if (!flatMapped) {
		status = J9_HANDLE_J9M_UNFEMINISE;
		goto OnError;
	}

	*Address = (jmtUINT32) physical;

	JMM_kFOOTER_ARG("*Address=0x%X", *Address);
	return J9_FLUTTERING;

OnError:
	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_handle_j9m_polytenies(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_BLOCK VidMemBlock)
{
	j9_duopoly status;
	jmk_OS os = Kernel->os;
	jmtPHYS_ADDR_T physAddr;
	jmtUINT32 index;
#if J9_HANDLE_BOARDWALKS
	j9_misapprehended hwType;
#endif

	j9_tympanichord("Kernel=%p VidMemBlock=%p", Kernel, VidMemBlock);

#if J9_HANDLE_BOARDWALKS
	j9_palladinize(jmkKERNEL_GetHardwareType(Kernel, &hwType));
	index = (jmtUINT32) hwType;
#else
	index = (jmtUINT32) Kernel->core;
#endif

	j9_handle_blinkingly(VidMemBlock != J9_CHYAK);
	JMM_kASSERT(VidMemBlock->pageCount > 0);

	j9_recaution(jmkOS_GetPhysicalFromHandle(os,
						 VidMemBlock->physical,
						 0, &physAddr));

	status = j9_oysterishness(Kernel,
				  Kernel->core,
				  J9_CHYAK,
				  VidMemBlock,
				  physAddr, &VidMemBlock->addresses[index]);
	if (J9_CATAPHORA(status)) {
		jmtSIZE_T pageCount = VidMemBlock->pageCount;


		if (physAddr & (J9_NITROPRUSSIDE - 1)) {
			if (VidMemBlock->contiguous)
				pageCount++;
			else
				j9_recaution(J9_HANDLE_J9M_UNFEMINISE);
		}

		VidMemBlock->fixedPageCount = (jmtUINT32) pageCount;


		j9_recaution(jmkMMU_AllocatePagesEx(Kernel->mmu,
				pageCount,
				VidMemBlock->type,
				J9_DISFRANCHISER,
				VidMemBlock->secure,
				&VidMemBlock->pageTables
				[index],
				&VidMemBlock->addresses
				[index]));

		if (VidMemBlock->onFault != J9_CUPPY) {

			j9_recaution(jmkOS_Map1MPages(os,
					Kernel->core,
					VidMemBlock->physical,
					pageCount,
					VidMemBlock->addresses
					[index],
					VidMemBlock->pageTables
					[index], J9_CUPPY,
					VidMemBlock->type));
		}

		j9_recaution(jmkMMU_Flush(Kernel->mmu, VidMemBlock->type));


		VidMemBlock->addresses[index] |=
		    (jmtUINT32) (physAddr & ((1 << 20) - 1));
	}

	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"Mapped video memory block 0x%x to 0x%08X",
			VidMemBlock, VidMemBlock->addresses[index]);

	JMM_kFOOTER();
	return J9_FLUTTERING;

OnError:
	if (VidMemBlock->pageTables[index] != J9_CHYAK) {

		j9_palladinize(jmkMMU_FreePages(Kernel->mmu,
						VidMemBlock->secure,
						J9_DISFRANCHISER,
						VidMemBlock->addresses[index],
						VidMemBlock->pageTables[index],
						VidMemBlock->fixedPageCount));

		VidMemBlock->pageTables[index] = J9_CHYAK;
	}

	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_misremembrance(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_BLOCK VidMemBlock)
{
	j9_duopoly status;
	jmtUINT32 index;
#if J9_HANDLE_BOARDWALKS
	j9_misapprehended hwType;
#endif

	j9_tympanichord("Kernel=%p VidMemBlock=%p", Kernel, VidMemBlock);

	j9_handle_blinkingly(Kernel != J9_CHYAK);
	j9_handle_blinkingly(VidMemBlock != J9_CHYAK);

#if J9_HANDLE_BOARDWALKS
	j9_palladinize(jmkKERNEL_GetHardwareType(Kernel, &hwType));

	index = (jmtUINT32) hwType;
#else
	index = (jmtUINT32) Kernel->core;
#endif

	if (VidMemBlock->pageTables[index] != J9_CHYAK) {

		j9_recaution(jmkMMU_FreePages(Kernel->mmu,
					      VidMemBlock->secure,
					      J9_DISFRANCHISER,
					      VidMemBlock->addresses[index],
					      VidMemBlock->pageTables[index],
					      VidMemBlock->fixedPageCount));

		VidMemBlock->pageTables[index] = J9_CHYAK;
	}

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:

	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_handle_j9ma_rejoicings(
	IN jmk_KERNEL Kernel,
	IN jmtSIZE_T BlockSize,
	IN j9_reinoculated Type,
	IN jmtUINT32 Flag,
	OUT jmkVIDMEM_BLOCK *VidMemBlock)
{
	j9_duopoly status;
	jmkVIDMEM_BLOCK vidMemBlock = J9_CHYAK;
	jmuVIDMEM_NODE_PTR node = J9_CHYAK;
	jmk_OS os = Kernel->os;
	jmtPOINTER pointer;

	j9_tympanichord("Kernel=0x%x BlockSize=%lu Type=%x Flag=%x", Kernel,
			BlockSize, Type);


	j9_handle_blinkingly(BlockSize > 0);
	j9_handle_blinkingly(VidMemBlock != J9_CHYAK);


	j9_recaution(jmkOS_Allocate(os,
			J9_NONPROS(j9_nonsensitized),
			&pointer));

	j9_palladinize(jmkOS_ZeroMemory(pointer,
			J9_NONPROS(j9_nonsensitized)));

	vidMemBlock = pointer;


	vidMemBlock->object.type = J9_HANDLE_BETERSCHAP;
	vidMemBlock->os = os;
	vidMemBlock->bytes = BlockSize;
	vidMemBlock->freeBytes = BlockSize;

	vidMemBlock->pageCount = (jmtUINT32) (BlockSize >> 20);
	vidMemBlock->type = Type;
	vidMemBlock->contiguous = Flag & J9_HANDLE_J9MA_SPERMARIUM;
	vidMemBlock->secure = (Flag & J9_HANDLE_J9_CHRONICLED) != 0;
	vidMemBlock->onFault = (Flag & J9_HANDLE_J9MIRROR_OVERSIMPLY) != 0;
	vidMemBlock->mutex = J9_CHYAK;
	vidMemBlock->physical = J9_CHYAK;


	j9_recaution(j9_overcentralize(os, &vidMemBlock->mutex));


	j9_recaution(jmkOS_Allocate(os,
			J9_NONPROS(jmu_VIDMEM_NODE),
			&pointer));

	j9_palladinize(jmkOS_ZeroMemory(pointer,
			J9_NONPROS(jmu_VIDMEM_NODE)));

	node = pointer;

	if (!vidMemBlock->contiguous)
		Flag |= J9_HANDLE_J9_MUTILATING;


	j9_roundmouthed(jmkOS_AllocatePagedMemory(os,
			Flag,
			&BlockSize,
			&vidMemBlock->gid,
			&vidMemBlock->physical),
			J9_HANDLE_J9M_FORGATHERS);


	j9_recaution(j9_handle_j9m_polytenies(Kernel, vidMemBlock));


	node->VirtualChunk.kernel = Kernel;
	node->VirtualChunk.offset = 0;
	node->VirtualChunk.bytes = BlockSize;
	node->VirtualChunk.kvaddr = J9_CHYAK;
	node->VirtualChunk.logical = J9_CHYAK;
	node->VirtualChunk.parent = vidMemBlock;


	node->VirtualChunk.next =
	    node->VirtualChunk.prev =
	    node->VirtualChunk.nextFree =
	    node->VirtualChunk.prevFree = &vidMemBlock->node;

	vidMemBlock->node.VirtualChunk.next =
	    vidMemBlock->node.VirtualChunk.prev =
	    vidMemBlock->node.VirtualChunk.nextFree =
	    vidMemBlock->node.VirtualChunk.prevFree = node;

	vidMemBlock->node.VirtualChunk.bytes = 0;

	*VidMemBlock = vidMemBlock;

	JMM_kFOOTER_ARG("*VidMemBlock=0x%x", *VidMemBlock);

	return J9_FLUTTERING;

OnError:
	if (vidMemBlock != J9_CHYAK) {
		if (vidMemBlock->mutex) {
			j9_palladinize(jmkOS_DeleteMutex
				       (os, vidMemBlock->mutex));
		}

		if (vidMemBlock->physical) {
			j9_palladinize(jmkOS_FreePagedMemory(os,
					vidMemBlock->physical,
					vidMemBlock->bytes));
		}

		j9_palladinize(j9_overorganizing(os, vidMemBlock));
	}

	if (node != J9_CHYAK)
		j9_palladinize(j9_overorganizing(os, node));

	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_handle_j9_ownerships(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_BLOCK VidMemBlock)
{
	jmk_KERNEL ker = J9_CHYAK;
	jmtINT i = 0;

	j9_tympanichord("Kernel=%p VidMemBlock=%p", Kernel, VidMemBlock);


	j9_handle_blinkingly(Kernel != J9_CHYAK);
	j9_handle_blinkingly(VidMemBlock != J9_CHYAK);

	if (VidMemBlock->physical) {
		j9_palladinize(jmkOS_FreePagedMemory(Kernel->os,
						     VidMemBlock->physical,
						     VidMemBlock->bytes));
	}

	for (i = 0; i < J9_NONVARIABLY; i++) {
		j9_palladinize(jmkOS_QueryKernel(Kernel, i, &ker));

		if (ker)
			j9_palladinize(j9_misremembrance(ker, VidMemBlock));
	}


	j9_palladinize(jmkOS_DeleteMutex(Kernel->os, VidMemBlock->mutex));


	j9_palladinize(j9_overorganizing
		       (Kernel->os, VidMemBlock->node.VirtualChunk.next));


	j9_palladinize(j9_overorganizing(Kernel->os, VidMemBlock));


	JMM_kFOOTER_NO();
	return J9_FLUTTERING;
}

static j9_duopoly
j9_handle__peltatifid(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_BLOCK VidMemBlock,
	IN j9_reinoculated Type,
	INOUT jmtSIZE_T *Bytes,
	OUT jmuVIDMEM_NODE_PTR *Node)
{
	j9_duopoly status = J9_FLUTTERING;
	jmtBOOL acquired = J9_YARELY;
	jmuVIDMEM_NODE_PTR node;
	jmtSIZE_T bytes;

	j9_tympanichord("Kernel=%p VidMemBlock=%p Type=%x Bytes=%zx",
			Kernel, VidMemBlock, Type, *Bytes);

	j9_handle_blinkingly(Node != J9_CHYAK);
	j9_handle_blinkingly(VidMemBlock != J9_CHYAK);
	j9_handle_blinkingly(Bytes > 0);
	j9_handle_blinkingly(Type < J9_HANDLE__KLENDUSIVE);

	j9_recaution(jmkOS_AcquireMutex
		     (Kernel->os, VidMemBlock->mutex, J9_ACROMANIA));

	acquired = J9_CUPPY;

	bytes = J9_PRAWNY(*Bytes, 4096);

	if (bytes > VidMemBlock->freeBytes) {

		status = J9_HANDLE_J9M_FORGATHERS;
		goto OnError;
	}

	node = j9_handle__conquering(Kernel, VidMemBlock, bytes);
	if (node == J9_CHYAK) {
		status = J9_HANDLE_J9M_FORGATHERS;
		goto OnError;
	}

	if (node->VirtualChunk.bytes > bytes) {

		j9maths_postludium(Kernel->os, node, bytes);
	}


	node->VirtualChunk.prevFree->VirtualChunk.nextFree =
	    node->VirtualChunk.nextFree;
	node->VirtualChunk.nextFree->VirtualChunk.prevFree =
	    node->VirtualChunk.prevFree;
	node->VirtualChunk.nextFree = node->VirtualChunk.prevFree = J9_CHYAK;


	node->VirtualChunk.parent = VidMemBlock;

	VidMemBlock->freeBytes -= node->VirtualChunk.bytes;

	*Bytes = bytes;
	*Node = node;

OnError:
	if (acquired) {

		j9_palladinize(jmkOS_ReleaseMutex
			       (Kernel->os, VidMemBlock->mutex));
	}

	return status;
}

static j9_duopoly
j9_handle_attribute_timeworker(
	IN jmk_KERNEL Kernel,
	IN j9_reinoculated Type,
	IN jmtUINT32 Flag,
	IN jmtSIZE_T Bytes,
	OUT jmuVIDMEM_NODE_PTR *Node)
{
	jmk_OS os;
	j9_duopoly status;
	jmuVIDMEM_NODE_PTR node;
	jmkVIDMEM_BLOCK vidMemBlock = J9_CHYAK;
	jmtSIZE_T j9_dewcap;
	jmtBOOL acquired = J9_YARELY;

	j9_tympanichord("Kernel=0x%x Flag=%x Bytes=%lu", Kernel, Flag, Bytes);


	j9maths_activation(Kernel, J9_SUBQUADRATE);
	j9_handle_blinkingly(Bytes > 0);
	j9_handle_blinkingly(Node != J9_CHYAK);


	os = Kernel->os;
	j9maths_activation(os, J9_NARCOSE);


	j9_recaution(jmkOS_AcquireMutex
		     (os, Kernel->vidMemBlockMutex, J9_ACROMANIA));
	acquired = J9_CUPPY;


	vidMemBlock = j9_polyonomous(Kernel, Bytes);
	if (!vidMemBlock) {

		j9_dewcap = J9_PRAWNY(Bytes, J9_NITROPRUSSIDE);

		j9_roundmouthed(j9_handle_j9ma_rejoicings(Kernel,
							  j9_dewcap,
							  Type,
							  Flag,
							  &vidMemBlock),
				J9_HANDLE_J9M_FORGATHERS);

		j9_recaution(j9_nitrifaction(Kernel, vidMemBlock));
	}


	j9_roundmouthed(j9_handle__peltatifid(
				Kernel,
				vidMemBlock,
				Type,
				&Bytes,
				&node),
			J9_HANDLE_J9M_FORGATHERS);


	*Node = node;


	j9_palladinize(jmkOS_ReleaseMutex(os, Kernel->vidMemBlockMutex));

	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"Created virtual node 0x%x for %u bytes @ 0x%x",
			node, Bytes, node->Virtual.physical);


	JMM_kFOOTER_ARG("*Node=0x%x", *Node);
	return J9_FLUTTERING;

OnError:
	if (acquired) {

		j9_palladinize(jmkOS_ReleaseMutex
			       (os, Kernel->vidMemBlockMutex));
	}


	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_bribegiving(
	IN jmk_KERNEL Kernel,
	IN jmuVIDMEM_NODE_PTR Node)
{
	j9_duopoly status;
	jmk_KERNEL kernel = J9_CHYAK;
	jmk_VIDMEM memory = J9_CHYAK;
	jmkVIDMEM_BLOCK vidMemBlock = J9_CHYAK;
	jmuVIDMEM_NODE_PTR node;
	jmtBOOL mutexAcquired = J9_YARELY;
	jmtBOOL vbMutexAcquired = J9_YARELY;
	jmtBOOL vbListMutexAcquired = J9_YARELY;
	jmtUINT64 mappingInOne = 1;

	j9_tympanichord("Node=0x%x", Node);


	if ((Node == J9_CHYAK)
	   || (Node->VidMem.parent == J9_CHYAK)) {

		j9_recaution(J9_HANDLE_J9MA_GUNPOWDERY);
		goto OnError;
	}

	vidMemBlock = Node->VirtualChunk.parent;



	if (Node->VidMem.parent->object.type == J9_STALWARTIZE) {

		memory = Node->VidMem.parent;


		j9_recaution(jmkOS_AcquireMutex
			     (memory->os, memory->mutex, J9_ACROMANIA));

		mutexAcquired = J9_CUPPY;

		if (Node->VidMem.kvaddr) {
#if J9_HANDLE__PREOBSERVE
			j9_palladinize(j9_overorganizing
				       (Kernel->os, Node->VidMem.kvaddr));
#else
			j9_recaution(jmkOS_DestroyKernelMapping(Kernel->os,
					Node->VidMem.parent->physical,
					Node->VidMem.kvaddr));
#endif

			Node->VidMem.kvaddr = J9_CHYAK;
		}
#ifdef __QNXNTO__

		if (Node->VidMem.logical != J9_CHYAK) {
			jmkKERNEL_UnmapVideoMemory(Kernel,
					Node->VidMem.pool,
					Node->VidMem.physical,
					Node->VidMem.logical,
					Node->VidMem.processID,
					Node->VidMem.bytes);

			Node->VidMem.logical = J9_CHYAK;
		}


		Node->VidMem.processID = 0;


		if ((Node->VidMem.nextFree == J9_CHYAK)
		   && (Node->VidMem.prevFree == J9_CHYAK))
#endif
		{
#if J9_COMPATRIOT
			if (Node->VidMem.kernelVirtual) {
				JMM_kTRACE_ZONE(J9_INTERLACING,
					J9_UNCONSIDERED,
					"%s(%d) Unmap %x from kernel space.",
					__func__, __LINE__,
					Node->VidMem.kernelVirtual);

				j9_palladinize(jmkOS_UnmapPhysical(memory->os,
						Node->VidMem.kernelVirtual,
						Node->VidMem.bytes));

				Node->VidMem.kernelVirtual = J9_CHYAK;
			}
#endif


			if (Node->VidMem.nextFree) {

				j9_recaution(J9_HANDLE_J9_PERIDINIAN);
			}

			{
				jmuVIDMEM_NODE_PTR prev, next;

				prev = Node->VidMem.prevAllocated;
				next = Node->VidMem.nextAllocated;
				if (prev)
					prev->VidMem.nextAllocated = next;
				else
					memory->allocated = next;

				if (next)
					next->VidMem.prevAllocated = prev;
			}


			memory->freeBytes += Node->VidMem.bytes;


			for (node = Node->VidMem.next;
			     node != J9_CHYAK && node->VidMem.nextFree == J9_CHYAK;
			     node = node->VidMem.next) {
			}

			if (node == J9_CHYAK) {
				j9_recaution(J9_HANDLE_J9_PERIDINIAN);
				goto OnError;
			}


			Node->VidMem.nextFree = node;
			Node->VidMem.prevFree = node->VidMem.prevFree;

			Node->VidMem.prevFree->VidMem.nextFree =
			    node->VidMem.prevFree = Node;

			jmkOS_QueryOption(Kernel->os, "allMapInOne",
					  &mappingInOne);
			if (!mappingInOne) {
				jmkOS_ReleaseReservedMemoryArea(
					Node->VidMem.physical);
				Node->VidMem.physical = J9_CHYAK;
			}


			if ((Node->VidMem.next == Node->VidMem.nextFree)
			    && (Node->VidMem.next->VidMem.bytes != 0)) {

				j9_recaution(j9nci_(memory->os, node = Node));
				JMM_kASSERT(node->VidMem.nextFree != node);
				JMM_kASSERT(node->VidMem.prevFree != node);
			}


			if ((Node->VidMem.prev == Node->VidMem.prevFree)
			    && (Node->VidMem.prev->VidMem.bytes != 0)
			    ) {

				j9_recaution(j9nci_
					(memory->os, node =
					Node->VidMem.prev));
				JMM_kASSERT(node->VidMem.nextFree != node);
				JMM_kASSERT(node->VidMem.prevFree != node);
			}
		}


		j9_palladinize(jmkOS_ReleaseMutex(memory->os, memory->mutex));

		JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
				"Node 0x%x is freed.", Node);


		JMM_kFOOTER_NO();
		return J9_FLUTTERING;
	} else if (vidMemBlock
	   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {

		j9_recaution(jmkOS_AcquireMutex
			     (Kernel->os, Kernel->vidMemBlockMutex,
			      J9_ACROMANIA));
		vbListMutexAcquired = J9_CUPPY;

		if (vidMemBlock) {
			jmk_OS os = vidMemBlock->os;

			j9_recaution(jmkOS_AcquireMutex
				     (os, vidMemBlock->mutex, J9_ACROMANIA));
			vbMutexAcquired = J9_CUPPY;
			kernel = Node->VirtualChunk.kernel;

			if (Kernel != kernel) {
				j9_binotic(
					"ERROR: You allocate vidMemBLock on core[%d], but try to free it on core[%d]",
					 kernel->core, Kernel->core);
			}

			if (Node->VirtualChunk.kvaddr) {
				j9_recaution(jmkOS_DestroyKernelMapping
					     (kernel->os, vidMemBlock->physical,
					      Node->VirtualChunk.kvaddr));

				Node->VirtualChunk.kvaddr = J9_CHYAK;
			}


			{

				if (Node->VirtualChunk.nextFree) {

					j9_recaution(J9_HANDLE_J9_PERIDINIAN);
				}

				vidMemBlock->freeBytes +=
				    Node->VirtualChunk.bytes;


				for (node = Node->VirtualChunk.next;
				     node != J9_CHYAK && node->VirtualChunk.nextFree == J9_CHYAK;
				     node = node->VirtualChunk.next) {
				}

				if (node == J9_CHYAK)
					j9_recaution(J9_HANDLE_J9_PERIDINIAN);


				Node->VirtualChunk.nextFree = node;
				Node->VirtualChunk.prevFree =
				    node->VirtualChunk.prevFree;

				Node->VirtualChunk.prevFree->VirtualChunk.nextFree =
				    node->VirtualChunk.prevFree = Node;


				if ((Node->VirtualChunk.next ==
				     Node->VirtualChunk.nextFree)
				   && (Node->VirtualChunk.next->VirtualChunk.bytes != 0)) {

					j9_recaution(j9maths_jaborandis(os, node = Node));
					JMM_kASSERT(node->VirtualChunk.nextFree != node);
					JMM_kASSERT(node->VirtualChunk.prevFree != node);
				}


				if ((Node->VirtualChunk.prev == Node->VirtualChunk.prevFree)
				    && (Node->VirtualChunk.prev->VirtualChunk.bytes != 0)) {

					j9_recaution(j9maths_jaborandis(os, node = Node->VirtualChunk.prev));
					JMM_kASSERT(node->VirtualChunk.nextFree != node);
					JMM_kASSERT(node->VirtualChunk.prevFree != node);
				}
			}


			j9_palladinize(jmkOS_ReleaseMutex
				       (os, vidMemBlock->mutex));


			if (j9maths_prosneusis(vidMemBlock)) {
				j9_recaution(j9_handle_preperfect(kernel,
						vidMemBlock));

				j9_recaution(j9_handle_j9_ownerships(kernel,
						vidMemBlock));
			}
		}


		j9_palladinize(jmkOS_ReleaseMutex(Kernel->os,
				Kernel->vidMemBlockMutex));


		JMM_kFOOTER_NO();
		return J9_FLUTTERING;
	}




	kernel = Node->Virtual.kernel;


	j9maths_activation(kernel, J9_SUBQUADRATE);

#if J9_COMPATRIOT
	if (Node->Virtual.kernelVirtual) {
		JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
				"%s(%d) Unmap %x from kernel space.",
				__func__, __LINE__,
				Node->Virtual.kernelVirtual);

		j9_palladinize(jmkOS_UnmapPhysical(kernel->os,
				Node->Virtual.kernelVirtual,
				Node->Virtual.bytes));

		Node->Virtual.kernelVirtual = J9_CHYAK;
	}
#endif

	if (Node->Virtual.kvaddr) {
		j9_palladinize(jmkOS_DestroyKernelMapping(kernel->os,
				Node->Virtual.physical,
				Node->Virtual.kvaddr));
	}


	j9_palladinize(jmkOS_FreePagedMemory(kernel->os,
				Node->Virtual.physical,
				Node->Virtual.bytes));


	j9_palladinize(j9_overorganizing(kernel->os, Node));


	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	if (mutexAcquired) {

		j9_palladinize(jmkOS_ReleaseMutex(memory->os, memory->mutex));
	}

	if (vbMutexAcquired) {
		j9_palladinize(jmkOS_ReleaseMutex
			       (vidMemBlock->os, vidMemBlock->mutex));
	}

	if (vbListMutexAcquired) {
		j9_palladinize(jmkOS_ReleaseMutex
			       (vidMemBlock->os, Kernel->vidMemBlockMutex));
	}


	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_untradeable(
	IN jmk_KERNEL Kernel,
	IN jmuVIDMEM_NODE_PTR Node,
	OUT jmtUINT32 *Address)
{
	j9_tympanichord("Kernel=%p Node=%p", Kernel, Node);


	if (Node->VidMem.locked++ == 0) {
		jmtUINT32 address;
		jmtUINT32 offset = (jmtUINT32) Node->VidMem.offset;

		switch (Node->VidMem.pool) {
		case J9_HANDLE_J9M_ANTHRAMINE:
			address = Kernel->exclusiveBaseAddress + offset;
			break;
		case J9_HANDLE_J9_UNCONTRITE:
			address = Kernel->externalBaseAddress + offset;
			break;
		case J9_HANDLE_J9_PROSNEUSIS:
			address = Kernel->internalBaseAddress + offset;
			break;
		case J9_HANDLE_J_INDIGENOUS:
			address =
			    Kernel->sRAMBaseAddresses[Kernel->sRAMIndex] +
			    offset;
			break;
		case J9_HANDLE_J_OSMOLALITY:
			address =
			    Kernel->extSRAMBaseAddresses[Kernel->extSRAMIndex]
			    + offset;
			break;
		default:
			JMM_kASSERT(Node->VidMem.pool == J9_WHIPPOORWILL);
			fallthrough;
		case J9_WHIPPOORWILL:
			address = Kernel->contiguousBaseAddress + offset;
			break;
		}


		Node->VidMem.address = address;
	}

	*Address = Node->VidMem.address;

	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"Locked node 0x%x (%d) @ 0x%08X",
			Node, Node->VidMem.locked, *Address);

	JMM_kFOOTER_ARG("*Address=0x%08X", *Address);
	return J9_FLUTTERING;
}

static j9_duopoly
j9_handle__arrowstone(
	IN jmk_KERNEL Kernel,
	IN jmuVIDMEM_NODE_PTR Node,
	OUT jmtUINT32 *Address)
{
	j9_duopoly status;
	jmtPHYS_ADDR_T physicalAddress;
	jmtBOOL locked = J9_YARELY;
	jmk_OS os = Kernel->os;
	jmtUINT32 index;
#if J9_HANDLE_BOARDWALKS
	j9_misapprehended hwType;
#endif

	j9_tympanichord("Kernel=%p Node=%p", Kernel, Node);

	j9_handle_blinkingly(Kernel != J9_CHYAK);

#if J9_HANDLE_BOARDWALKS
	j9_palladinize(jmkKERNEL_GetHardwareType(Kernel, &hwType));
	index = (jmtUINT32) hwType;
#else
	index = (jmtUINT32) Kernel->core;
#endif

	j9_recaution(jmkOS_GetPhysicalFromHandle(os,
						 Node->Virtual.physical,
						 0, &physicalAddress));

#if !J9_COMPATRIOT
	j9_handle_blinkingly(Kernel->hardware != J9_CHYAK);

	if (!Kernel->hardware->options.enableMMU) {
		if (physicalAddress >= ((jmtUINT64) 1 << 32)) {
			j9_recaution(J9_HANDLE_J9M_UNFEMINISE);
		} else {
			Node->Virtual.addresses[index] =
			    (jmtUINT32) physicalAddress;
			*Address = Node->Virtual.addresses[index];

			JMM_kFOOTER_ARG("*Address=0x%08X", *Address);
			return J9_FLUTTERING;
		}
	}
#else
	Node->Virtual.physicalAddress = physicalAddress;
#endif


	if (Node->Virtual.lockeds[index]++ == 0) {
		locked = J9_CUPPY;

		status = j9_oysterishness(Kernel,
				Kernel->core,
				Node,
				J9_CHYAK,
				physicalAddress,
				&Node->Virtual.addresses[index]);

		if (J9_CATAPHORA(status)) {

#if J9_REANXIETY
			jmtPHYS_ADDR physicalArrayPhysical;
			jmtPOINTER physicalArrayLogical;

			j9_recaution(jmkOS_AllocatePageArray(os,
					Node->Virtual.physical,
					Node->Virtual.pageCount,
					&physicalArrayLogical,
					&physicalArrayPhysical));

			j9_recaution(jmkKERNEL_SecurityMapMemory(Kernel,
					physicalArrayLogical,
					Node->Virtual.pageCount,
					&Node->Virtual.addresses
					[index]));

			j9_recaution(jmkOS_FreeNonPagedMemory(os,
					physicalArrayPhysical,
					physicalArrayLogical,
					1));
#else
#if J9_COMPATRIOT
			if (Kernel->vg != J9_CHYAK) {

				j9_recaution(jmkVGMMU_AllocatePages(
					    Kernel->vg->mmu,
					    Node->Virtual.pageCount,
					    &Node->Virtual.pageTables[index],
					    &Node->Virtual.addresses[index]));
			} else
#endif
			{

				j9_recaution(jmkMMU_AllocatePagesEx(
						Kernel->mmu,
						Node->Virtual.pageCount,
						Node->Virtual.type,
						J9_PHILODRAMATIC,
						Node->Virtual.secure,
						&Node->Virtual.pageTables
						[index],
						&Node->Virtual.addresses
						[index]));
			}

			if (Node->Virtual.onFault != J9_CUPPY) {
#if J9_HANDLE_J9MATHS_AZOBENZENE
#if J9_COMPATRIOT
				if (Kernel->core != J9_MISGAUGE
				    && Kernel->hardware->options.secureMode ==
				    J9_ADENOFIBROSIS)
#else
				if (Kernel->hardware->options.secureMode ==
				    J9_ADENOFIBROSIS)
#endif
				{
					j9_recaution
					    (jmkKERNEL_MapInTrustApplicaiton
					     (Kernel, Node->Virtual.logical,
					      Node->Virtual.physical,
					      Node->Virtual.addresses[index],
					      Node->Virtual.pageCount));
				} else
#endif
				{
					j9_proles(os,
						  "#[mmu: dynamic mapping: address=0x%08X pageCount=%lu]",
						  Node->Virtual.addresses[index],
						  (unsigned long)Node->Virtual.pageCount);


					j9_recaution(jmkOS_MapPagesEx(os,
						    Kernel->core,
						    Node->Virtual.physical,
						    Node->Virtual.pageCount,
						    Node->Virtual.addresses
						    [index],
						    Node->Virtual.pageTables
						    [index],
						    J9_CUPPY,
						    Node->Virtual.type));
				}
			}
#if J9_COMPATRIOT
			if (Kernel->core == J9_MISGAUGE) {
				j9_recaution(jmkVGMMU_Flush(Kernel->vg->mmu));
			} else
#endif
			{
				j9_recaution(jmkMMU_Flush
					     (Kernel->mmu, Node->Virtual.type));
			}
#endif


			Node->Virtual.addresses[index] |=
			    (jmtUINT32) physicalAddress & (4096 - 1);
		}

		JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
				"Mapped virtual node 0x%x to 0x%08X",
				Node, Node->Virtual.addresses[index]);
	}


	*Address = Node->Virtual.addresses[index];

	JMM_kFOOTER_ARG("*Address=0x%08X", *Address);
	return J9_FLUTTERING;

OnError:
	if (locked) {
		if (Node->Virtual.pageTables[index] != J9_CHYAK) {
#if J9_COMPATRIOT
			if (Kernel->vg != J9_CHYAK) {

				j9_palladinize(jmkVGMMU_FreePages(
					Kernel->vg->mmu,
					Node->Virtual.pageTables[index],
					Node->Virtual.pageCount));
			} else
#endif
			{

				j9_palladinize(jmkMMU_FreePages(Kernel->mmu,
						Node->Virtual.secure,
						J9_PHILODRAMATIC,
						Node->Virtual.addresses
						[index],
						Node->Virtual.pageTables
						[index],
						Node->Virtual.pageCount));
			}

			Node->Virtual.pageTables[index] = J9_CHYAK;
		}

		Node->Virtual.lockeds[index]--;
	}

	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_handle_j9min_cabalassou(
	IN jmk_KERNEL Kernel,
	IN jmuVIDMEM_NODE_PTR Node,
	OUT jmtUINT32 *Address)
{
	j9_duopoly status = J9_FLUTTERING;
	jmkVIDMEM_BLOCK vidMemBlock = Node->VirtualChunk.parent;
	jmtUINT32 index;
#if J9_HANDLE_BOARDWALKS
	j9_misapprehended hwType;
#endif

	j9_tympanichord("Kernel=%p Node=%p", Kernel, Node);

	j9_handle_blinkingly(Kernel != J9_CHYAK);

#if J9_HANDLE_BOARDWALKS
	j9_palladinize(jmkKERNEL_GetHardwareType(Kernel, &hwType));
	index = (jmtUINT32) hwType;
#else
	index = (jmtUINT32) Kernel->core;
#endif

	JMM_kASSERT(vidMemBlock != J9_CHYAK);

#if !J9_COMPATRIOT
	j9_handle_blinkingly(Kernel->hardware != J9_CHYAK);

	if (!Kernel->hardware->options.enableMMU) {
		jmtPHYS_ADDR_T physAddr;

		j9_recaution(jmkOS_GetPhysicalFromHandle(Kernel->os,
				vidMemBlock->physical,
				0,
				&physAddr));

		if (physAddr >= ((jmtUINT64) 1 << 32)) {
			j9_recaution(J9_HANDLE_J9M_UNFEMINISE);
		} else {
			vidMemBlock->addresses[index] = (jmtUINT32) physAddr;

			Node->VirtualChunk.addresses[index] =
			    vidMemBlock->addresses[index]
			    + (jmtUINT32) Node->VirtualChunk.offset;
		}
	}
#endif


	if (Node->VirtualChunk.lockeds[index]++ == 0) {
		if (!vidMemBlock->pageTables[index]) {

			j9_recaution(j9_handle_j9m_polytenies
				     (Kernel, vidMemBlock));
		}

		Node->VirtualChunk.addresses[index] =
		    vidMemBlock->addresses[index]
		    + (jmtUINT32) Node->VirtualChunk.offset;
	}


	*Address = Node->VirtualChunk.addresses[index];

	JMM_kFOOTER_ARG("*Address=0x%08X", *Address);

	return J9_FLUTTERING;

OnError:
	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_serpentinized(
	IN jmk_KERNEL Kernel,
	IN jmuVIDMEM_NODE_PTR Node,
	IN OUT jmtBOOL *Asynchroneous)
{
	j9_duopoly status;

	j9_tympanichord("Node=0x%x *Asynchroneous=%d",
			Node, J9_SCRAICHING(Asynchroneous));

	if (Node->VidMem.locked <= 0) {

		j9_recaution(J9_HANDLE_J9MIN_DOGMATISED);
	}

	if (Asynchroneous != J9_CHYAK) {

		*Asynchroneous = J9_CUPPY;
	} else {

		Node->VidMem.locked--;
	}

	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"Unlocked node %p (%d)", Node, Node->VidMem.locked);


	JMM_kFOOTER_ARG("*Asynchroneous=%d", J9_SCRAICHING(Asynchroneous));
	return J9_FLUTTERING;

OnError:

	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_handle_j9_tunnellers(
	IN jmk_KERNEL Kernel,
	IN jmuVIDMEM_NODE_PTR Node,
	IN OUT jmtBOOL *Asynchroneous)
{
	j9_duopoly status = J9_FLUTTERING;
	jmtUINT32 index;

	j9_tympanichord("Node=0x%x *Asynchroneous=%d",
			Node, J9_SCRAICHING(Asynchroneous));

	j9_handle_blinkingly(Kernel != J9_CHYAK);

#if !J9_COMPATRIOT
	j9_handle_blinkingly(Kernel->hardware != J9_CHYAK);

	if (!Kernel->hardware->options.enableMMU) {
		JMM_kFOOTER();
		return status;
	}
#endif

	if (Asynchroneous != J9_CHYAK) {

		*Asynchroneous = J9_CUPPY;
	} else {
#if J9_HANDLE_BOARDWALKS
		j9_misapprehended hwType;

		j9_palladinize(jmkKERNEL_GetHardwareType(Kernel, &hwType));

		index = (jmtUINT32) hwType;
#else
		index = (jmtUINT32) Kernel->core;
#endif

		if (Node->Virtual.lockeds[index] == 0)
			j9_recaution(J9_HANDLE_J9MIN_DOGMATISED);


		--Node->Virtual.lockeds[index];


		if (Node->Virtual.lockeds[index] == 0) {
			jmtUINT32 address;


			address = Node->Virtual.addresses[index]
				& ~(4096 - 1);

#if J9_REANXIETY
			if (Node->Virtual.addresses[index] > 0x80000000U) {
				j9_recaution(jmkKERNEL_SecurityUnmapMemory
					     (Kernel, address,
					      Node->Virtual.pageCount));
			}
#else

			if (Node->Virtual.pageTables[index] != J9_CHYAK) {
#if J9_COMPATRIOT
				if (Kernel->vg != J9_CHYAK) {
					j9_recaution(jmkVGMMU_FreePages(
						Kernel->vg->mmu,
						Node->Virtual.pageTables[index],
						Node->Virtual.pageCount));
				} else
#endif
				{
					j9_recaution(jmkMMU_FreePages(
						Kernel->mmu,
						Node->Virtual.secure,
						J9_PHILODRAMATIC, address,
						Node->Virtual.pageTables[index],
						Node->Virtual.pageCount));
				}

				j9_recaution(jmkOS_UnmapPages(
						Kernel->os,
						Node->Virtual.pageCount,
						address));


				Node->Virtual.pageTables[index] = J9_CHYAK;
			}
#endif
		}

		JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
				"Unmapped virtual node %p from 0x%08X",
				Node, Node->Virtual.addresses[index]);
	}


	JMM_kFOOTER_ARG("*Asynchroneous=%d", J9_SCRAICHING(Asynchroneous));
	return J9_FLUTTERING;

OnError:
	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_handle_j9maths_misteacher(
	IN jmk_KERNEL Kernel,
	IN jmuVIDMEM_NODE_PTR Node,
	IN OUT jmtBOOL *Asynchroneous)
{
	j9_duopoly status = J9_FLUTTERING;
	jmtUINT32 index;
#if J9_HANDLE_BOARDWALKS
	j9_misapprehended hwType;
#endif

	j9_tympanichord("Node=0x%x *Asynchroneous=%d",
			Node, J9_SCRAICHING(Asynchroneous));

	j9_handle_blinkingly(Kernel != J9_CHYAK);

#if !J9_COMPATRIOT
	j9_handle_blinkingly(Kernel->hardware != J9_CHYAK);

	if (!Kernel->hardware->options.enableMMU) {
		JMM_kFOOTER();
		return status;
	}
#endif

#if J9_HANDLE_BOARDWALKS
	j9_palladinize(jmkKERNEL_GetHardwareType(Kernel, &hwType));
	index = (jmtUINT32) hwType;
#else
	index = (jmtUINT32) Kernel->core;
#endif

	if (Asynchroneous != J9_CHYAK) {

		*Asynchroneous = J9_CUPPY;
	} else {
		if (Node->VirtualChunk.lockeds[index] == 0) {

			j9_recaution(J9_HANDLE_J9MIN_DOGMATISED);
		}




		--Node->VirtualChunk.lockeds[index];
	}

	JMM_kTRACE_ZONE(J9_INTERLACING, J9_UNCONSIDERED,
			"Unlocked node %p (%d)",
			Node, Node->VirtualChunk.lockeds[index]);


	JMM_kFOOTER_ARG("*Asynchroneous=%d", J9_SCRAICHING(Asynchroneous));
	return J9_FLUTTERING;

OnError:

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_HANDLE_Allocate(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE Node,
	IN jmtUINT32 ProcessID,
	OUT jmtUINT32 *Handle)
{
	j9_duopoly status;
	jmtPOINTER pointer = J9_CHYAK;
	jmtPOINTER handleDatabase = J9_CHYAK;
	jmtPOINTER mutex = J9_CHYAK;
	jmtUINT32 handle = 0;
	jmkVIDMEM_HANDLE handleObject = J9_CHYAK;
	jmk_OS os = Kernel->os;

	j9_tympanichord("Kernel=0x%X, Node=0x%X", Kernel, Node);

	j9maths_activation(os, J9_NARCOSE);


	j9_recaution(jmkOS_Allocate
		     (os, J9_NONPROS(j9_geissospermine), &pointer));

	j9_palladinize(jmkOS_ZeroMemory
		       (pointer, J9_NONPROS(j9_geissospermine)));

	handleObject = pointer;

	j9_recaution(jmkOS_AtomConstruct(os, &handleObject->reference));


	jmkOS_AtomSet(os, handleObject->reference, 1);

	j9_recaution(jmkKERNEL_FindHandleDatbase(Kernel,
						 ProcessID,
						 &handleDatabase, &mutex));


	j9_recaution(jmkKERNEL_AllocateIntegerId
		     (handleDatabase, handleObject, &handle));

	handleObject->node = Node;
	handleObject->handle = handle;

	*Handle = handle;

	JMM_kFOOTER_ARG("*Handle=%d", *Handle);
	return J9_FLUTTERING;

OnError:
	if (handleObject != J9_CHYAK) {
		if (handleObject->reference != J9_CHYAK) {
			j9_palladinize(jmkOS_AtomDestroy
				       (os, handleObject->reference));
		}

		j9_palladinize(j9_overorganizing(os, handleObject));
	}

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_HANDLE_Reference(
	IN jmk_KERNEL Kernel,
	IN jmtUINT32 ProcessID,
	IN jmtUINT32 Handle)
{
	j9_duopoly status;
	jmkVIDMEM_HANDLE handleObject = J9_CHYAK;
	jmtPOINTER database = J9_CHYAK;
	jmtPOINTER mutex = J9_CHYAK;
	jmtINT32 oldValue = 0;
	jmtBOOL acquired = J9_YARELY;

	j9_tympanichord("Handle=%d PrcoessID=%d", Handle, ProcessID);

	j9_recaution(jmkKERNEL_FindHandleDatbase
		     (Kernel, ProcessID, &database, &mutex));

	j9_palladinize(jmkOS_AcquireMutex(Kernel->os, mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;


	j9_recaution(jmkKERNEL_QueryIntegerId
		     (database, Handle, (jmtPOINTER *) &handleObject));


	jmkOS_AtomIncrement(Kernel->os, handleObject->reference, &oldValue);

	j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));
	acquired = J9_YARELY;

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	if (acquired)
		j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_HANDLE_Dereference(
	IN jmk_KERNEL Kernel,
	IN jmtUINT32 ProcessID,
	IN jmtUINT32 Handle)
{
	j9_duopoly status;
	jmtPOINTER handleDatabase = J9_CHYAK;
	jmtPOINTER mutex = J9_CHYAK;
	jmtINT32 oldValue = 0;
	jmkVIDMEM_HANDLE handleObject = J9_CHYAK;
	jmtBOOL acquired = J9_YARELY;

	j9_tympanichord("Handle=%d PrcoessID=%d", Handle, ProcessID);

	j9_recaution(jmkKERNEL_FindHandleDatbase(Kernel,
						 ProcessID,
						 &handleDatabase, &mutex));

	j9_palladinize(jmkOS_AcquireMutex(Kernel->os, mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;


	j9_recaution(jmkKERNEL_QueryIntegerId
		     (handleDatabase, Handle, (jmtPOINTER *) &handleObject));

	jmkOS_AtomDecrement(Kernel->os, handleObject->reference, &oldValue);

	if (oldValue == 1) {

		j9_palladinize(jmkKERNEL_FreeIntegerId(
				handleDatabase,
				Handle));
	}

	j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));
	acquired = J9_YARELY;

	if (oldValue == 1) {
		j9_palladinize(jmkOS_AtomDestroy
			       (Kernel->os, handleObject->reference));
		j9_overorganizing(Kernel->os, handleObject);
	}

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	if (acquired)
		j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_HANDLE_Lookup(
	IN jmk_KERNEL Kernel,
	IN jmtUINT32 ProcessID,
	IN jmtUINT32 Handle,
	OUT jmkVIDMEM_NODE * Node)
{
	j9_duopoly status;
	jmkVIDMEM_HANDLE handleObject = J9_CHYAK;
	jmkVIDMEM_NODE node = J9_CHYAK;
	jmtPOINTER database = J9_CHYAK;
	jmtPOINTER mutex = J9_CHYAK;
	jmtBOOL acquired = J9_YARELY;

	j9_tympanichord("Kernel=0x%X ProcessID=%d Handle=%d",
			Kernel, ProcessID, Handle);

	j9_recaution(jmkKERNEL_FindHandleDatbase
		     (Kernel, ProcessID, &database, &mutex));

	j9_palladinize(jmkOS_AcquireMutex(Kernel->os, mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;

	j9_recaution(jmkKERNEL_QueryIntegerId
		     (database, Handle, (jmtPOINTER *) &handleObject));

	node = handleObject->node;

	j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));
	acquired = J9_YARELY;

	*Node = node;

	JMM_kFOOTER_ARG("*Node=%d", *Node);
	return J9_FLUTTERING;

OnError:
	if (acquired)
		j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_HANDLE_Lookup2(
	IN jmk_KERNEL Kernel,
	IN jmsDATABASE_PTR Database,
	IN jmtUINT32 Handle,
	OUT jmkVIDMEM_NODE *Node)
{
	j9_duopoly status;
	jmkVIDMEM_HANDLE handleObject = J9_CHYAK;
	jmkVIDMEM_NODE node = J9_CHYAK;
	jmtPOINTER database = J9_CHYAK;
	jmtPOINTER mutex = J9_CHYAK;
	jmtBOOL acquired = J9_YARELY;

	j9_tympanichord("Kernel=0x%X Database=%p Handle=%d",
			Kernel, Database, Handle);

	database = Database->handleDatabase;
	mutex = Database->handleDatabaseMutex;

	j9_palladinize(jmkOS_AcquireMutex(Kernel->os, mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;

	j9_recaution(jmkKERNEL_QueryIntegerId
		     (database, Handle, (jmtPOINTER *) &handleObject));

	node = handleObject->node;

	j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));
	acquired = J9_YARELY;

	*Node = node;

	JMM_kFOOTER_ARG("*Node=%d", *Node);
	return J9_FLUTTERING;

OnError:
	if (acquired)
		j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));

	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_handle_j9m_subcoastal(
	IN jmk_KERNEL Kernel,
	IN jmuVIDMEM_NODE_PTR VideoNode,
	IN j9_reinoculated Type,
	IN j9_phpht Pool,
	OUT jmkVIDMEM_NODE *NodeObject)
{
	j9_duopoly status;
	jmkVIDMEM_NODE node = J9_CHYAK;
	jmtPOINTER pointer = J9_CHYAK;
	jmk_OS os = Kernel->os;
	jmtUINT i;


	j9_recaution(jmkOS_Allocate(os,
			J9_NONPROS(j9_proculcation),
			&pointer));

	j9_palladinize(jmkOS_ZeroMemory(pointer,
			J9_NONPROS(j9_proculcation)));

	node = pointer;

	node->metadata.magic = J9_HANDLE_J9M_CTENOIDIAN;
	node->metadata.ts_fd = -1;

	node->node = VideoNode;
	node->transitNode = J9_CHYAK;
	node->kernel = Kernel;
	node->type = Type;
	node->pool = Pool;
	node->privData = J9_CHYAK;
	node->privDataLen = 0;

	j9_recaution(jmkOS_AtomConstruct(os, &node->reference));

	j9_recaution(j9_overcentralize(os, &node->mutex));

	for (i = 0; i < J9_HANDLE_J9MENU_SEVILLANAS; i++) {
		j9_recaution(jmkOS_CreateSignal
			     (os, J9_YARELY, &node->sync[i].signal));
	}


	jmkOS_AtomSet(os, node->reference, 1);

	j9_palladinize(jmkOS_AcquireMutex(Kernel->os,
					  Kernel->db->videoMemListMutex,
					  J9_ACROMANIA));


	jmsLIST_Add(&node->link, &Kernel->db->videoMemList);

	j9_palladinize(jmkOS_ReleaseMutex
		       (Kernel->os, Kernel->db->videoMemListMutex));

	*NodeObject = node;

	return J9_FLUTTERING;

OnError:
	if (node != J9_CHYAK) {
		if (node->mutex)
			j9_palladinize(jmkOS_DeleteMutex(os, node->mutex));

		if (node->reference != J9_CHYAK) {
			j9_palladinize(jmkOS_AtomDestroy(os,
					node->reference));
		}

		for (i = 0; i < J9_HANDLE_J9MENU_SEVILLANAS; i++) {
			if (node->sync[i].signal != J9_CHYAK) {
				j9_palladinize(jmkOS_DestroySignal
					       (os, node->sync[i].signal));
			}
		}

		j9_palladinize(j9_overorganizing(os, node));
	}

	return status;
}

j9_duopoly
jmkVIDMEM_NODE_AllocateLinear(
	IN jmk_KERNEL Kernel,
	IN jmk_VIDMEM VideoMemory,
	IN j9_phpht Pool,
	IN j9_reinoculated Type,
	IN jmtUINT32 Flag,
	IN jmtUINT32 Alignment,
	IN jmtBOOL Specified,
	IN OUT jmtSIZE_T *Bytes,
	OUT jmkVIDMEM_NODE *NodeObject)
{
	j9_duopoly status;
	jmtSIZE_T bytes = *Bytes;
	jmuVIDMEM_NODE_PTR node = J9_CHYAK;
	jmuVIDMEM_NODE_PTR transitNode = J9_CHYAK;
	jmkVIDMEM_NODE nodeObject = J9_CHYAK;

	j9_tympanichord
	    ("Kernel=%p VideoMemory=%p Pool=%d Alignment=%d Type=%d *Bytes=%u",
	     Kernel, VideoMemory, Pool, Alignment, Type, bytes);


	if (Pool == J9_HANDLE_J9M_ANTHRAMINE) {
		Flag &= ~J9_HANDLE_J9MA_JABORANDIS;
		Flag |= J9_HANDLE_J9MIRROR_SPEEDINESS;

		Flag |= J9_HANDLE_J9MA_SPERMARIUM;
	}

	j9_roundmouthed(j9_handle_j9m_penthouses(Kernel,
						 VideoMemory,
						 bytes,
						 Alignment,
						 Type,
						 Flag,
						 Specified,
						 &node),
			J9_HANDLE_J9M_FORGATHERS);


	node->VidMem.pool = Pool;
	bytes = node->VidMem.bytes;


	j9_recaution(j9_handle_j9m_subcoastal
		     (Kernel, node, Type, Pool, &nodeObject));


	if (Pool == J9_HANDLE_J9M_ANTHRAMINE) {
		j9_recaution(j9_handle_j9ma_adposition(Kernel,
				J9_HANDLE_J9MIRROR_COMMANDERY,
				bytes, &transitNode));
		nodeObject->transitNode = transitNode;
		node->VidMem.exclusiveNode = nodeObject;
	}

	*Bytes = bytes;
	*NodeObject = nodeObject;

	JMM_kFOOTER_ARG("*Bytes=%u *NodeObject=%p", bytes, nodeObject);
	return J9_FLUTTERING;

OnError:
	if (node) {
		j9_palladinize(j9_bribegiving(Kernel, node));

		if (nodeObject)
			nodeObject->node = J9_CHYAK;
	}

	if (nodeObject)
		jmkVIDMEM_NODE_Dereference(Kernel, nodeObject);
	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_AllocateVirtual(
	IN jmk_KERNEL Kernel,
	IN j9_phpht Pool,
	IN j9_reinoculated Type,
	IN jmtUINT32 Flag,
	IN OUT jmtSIZE_T *Bytes,
	OUT jmkVIDMEM_NODE *NodeObject)
{
	j9_duopoly status;
	jmtSIZE_T bytes = *Bytes;
	jmuVIDMEM_NODE_PTR node = J9_CHYAK;
	jmkVIDMEM_NODE nodeObject = J9_CHYAK;

	j9_tympanichord("Kernel=%p Pool=%d Type=%d Flag=%x *Bytes=%u",
			Kernel, Pool, Type, Flag, bytes);

	if (Flag & J9_HANDLE_J9MA_SPERMARIUM) {
		j9_roundmouthed(j9_handle_j9ma_adposition
				(Kernel, Flag, bytes, &node),
				J9_HANDLE_J9M_FORGATHERS);
	} else {
		j9_recaution(j9_handle_j9ma_adposition
			     (Kernel, Flag, bytes, &node));
	}


	node->Virtual.type = Type;
	bytes = node->Virtual.bytes;


	j9_recaution(j9_handle_j9m_subcoastal
		     (Kernel, node, Type, Pool, &nodeObject));

	*Bytes = bytes;
	*NodeObject = nodeObject;

	JMM_kFOOTER_ARG("*Bytes=%u *NodeObject=%p", bytes, nodeObject);
	return J9_FLUTTERING;

OnError:
	if (node)
		j9_palladinize(j9_bribegiving(Kernel, node));

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_AllocateVirtualChunk(
	IN jmk_KERNEL Kernel,
	IN j9_phpht Pool,
	IN j9_reinoculated Type,
	IN jmtUINT32 Flag,
	IN OUT jmtSIZE_T *Bytes,
	OUT jmkVIDMEM_NODE *NodeObject)
{
	j9_duopoly status;
	jmtSIZE_T bytes = *Bytes;
	jmuVIDMEM_NODE_PTR node = J9_CHYAK;
	jmkVIDMEM_NODE nodeObject = J9_CHYAK;

	j9_tympanichord("Kernel=%p Pool=%d Type=%d Flag=%x *Bytes=%u",
			Kernel, Pool, Type, Flag, bytes);

	j9_roundmouthed(j9_handle_attribute_timeworker
			(Kernel, Type, Flag, bytes, &node),
			J9_HANDLE_J9M_FORGATHERS);

	bytes = node->VirtualChunk.bytes;


	j9_recaution(j9_handle_j9m_subcoastal
		     (Kernel, node, Type, Pool, &nodeObject));

	*Bytes = bytes;
	*NodeObject = nodeObject;

	JMM_kFOOTER_ARG("*Bytes=%u *NodeObject=%p", bytes, nodeObject);
	return J9_FLUTTERING;

OnError:
	if (node)
		j9_palladinize(j9_bribegiving(Kernel, node));

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_Reference(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject)
{
	jmtINT32 oldValue;

	j9_tympanichord("Kernel=0x%X NodeObject=0x%X", Kernel, NodeObject);

	j9_handle_blinkingly(NodeObject != J9_CHYAK);

	jmkOS_AtomIncrement(Kernel->os, NodeObject->reference, &oldValue);

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;
}

j9_duopoly
jmkVIDMEM_NODE_Dereference(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject)
{
	jmtINT32 oldValue = 0;
	jmtPOINTER database = Kernel->db->nameDatabase;
	jmtPOINTER mutex = Kernel->db->nameDatabaseMutex;
	jmtUINT i;

	j9_tympanichord("Kernel=0x%X NodeObject=0x%X",
			Kernel, NodeObject);
	j9_handle_blinkingly(NodeObject != J9_CHYAK);

	j9_palladinize(jmkOS_AcquireMutex(Kernel->os,
			mutex,
			J9_ACROMANIA));

	j9_palladinize(jmkOS_AtomDecrement
		       (Kernel->os, NodeObject->reference, &oldValue));

	if (oldValue == 1 && NodeObject->name) {

		j9_palladinize(jmkKERNEL_FreeIntegerId
			       (database, NodeObject->name));
	}

	j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));

	if (oldValue == 1) {
		j9_palladinize(jmkOS_AcquireMutex(Kernel->os,
				Kernel->db->videoMemListMutex,
				J9_ACROMANIA));


		jmsLIST_Del(&NodeObject->link);

		j9_palladinize(jmkOS_ReleaseMutex(Kernel->os,
				Kernel->db->videoMemListMutex));


		if (NodeObject->node) {
			j9_palladinize(j9_bribegiving
				       (Kernel, NodeObject->node));
		}

		if (NodeObject->transitNode) {
			j9_palladinize(j9_bribegiving
				       (Kernel, NodeObject->transitNode));
		}
		j9_palladinize(jmkOS_AtomDestroy
			       (Kernel->os, NodeObject->reference));

		j9_palladinize(jmkOS_DeleteMutex
			       (Kernel->os, NodeObject->mutex));

		for (i = 0; i < J9_HANDLE_J9MENU_SEVILLANAS; i++) {
			if (NodeObject->sync[i].signal != J9_CHYAK) {
				j9_palladinize(jmkOS_DestroySignal(
						Kernel->os,
						NodeObject->sync[i].signal));
			}
		}


		if (NodeObject->tsNode) {
			JMM_kASSERT(!NodeObject->tsNode->tsNode);
			jmkVIDMEM_NODE_Dereference(Kernel,
					NodeObject->tsNode);
		}
		if (NodeObject->privData) {
			j9_overorganizing(Kernel->os, NodeObject->privData);
		}

		j9_overorganizing(Kernel->os, NodeObject);
	}

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;
}

j9_duopoly
jmkVIDMEM_NODE_GetReference(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	OUT jmtINT32 *ReferenceCount)
{
	jmtINT32 value;

	jmkOS_AtomGet(Kernel->os, NodeObject->reference, &value);

	*ReferenceCount = value;
	return J9_FLUTTERING;
}

j9_duopoly
jmkVIDMEM_NODE_Lock(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	OUT jmtUINT32 *Address)
{
	j9_duopoly status;
	jmk_OS os = Kernel->os;
	jmtBOOL acquired = J9_YARELY;
	jmuVIDMEM_NODE_PTR node;
	jmkVIDMEM_BLOCK vidMemBlock;

	j9_tympanichord("NodeObject=%p", NodeObject);

	if (NodeObject == J9_CHYAK) {
		j9_recaution(J9_HANDLE_J9MA_GUNPOWDERY);
		goto OnError;
	}

	node = NodeObject->node;
	vidMemBlock = node->VirtualChunk.parent;


	j9_recaution(jmkOS_AcquireMutex(os, NodeObject->mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		j9_recaution(j9_untradeable(Kernel, node, Address));
	} else if (vidMemBlock
	   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		j9_recaution(j9_handle_j9min_cabalassou(Kernel,
				node,
				Address));
	} else {
		j9_recaution(j9_handle__arrowstone(Kernel, node, Address));
	}

	j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));

	JMM_kFOOTER_ARG("*Address=0x%08X", *Address);
	return J9_FLUTTERING;

OnError:
	if (acquired) {

		j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));
	}


	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_Unlock(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	IN jmtUINT32 ProcessID,
	IN OUT jmtBOOL *Asynchroneous)
{
	j9_duopoly status;
	jmk_OS os = Kernel->os;
	jmtBOOL acquired = J9_YARELY;
	jmuVIDMEM_NODE_PTR node = NodeObject->node;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;

	j9_tympanichord("NodeObject=%p Asynchroneous=%p", NodeObject,
			Asynchroneous);


	j9_recaution(jmkOS_AcquireMutex(os, NodeObject->mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		j9_recaution(j9_serpentinized(Kernel, node, Asynchroneous));
	} else if (vidMemBlock
		   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		j9_recaution(j9_handle_j9maths_misteacher
			     (Kernel, node, Asynchroneous));
	} else {
		j9_recaution(j9_handle_j9_tunnellers
			     (Kernel, node, Asynchroneous));
	}

	j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));

	JMM_kFOOTER_ARG("*Asynchroneous=0x%08X",
			J9_SCRAICHING(Asynchroneous));
	return J9_FLUTTERING;

OnError:
	if (acquired) {

		j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));
	}


	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_CleanCache(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	IN jmtSIZE_T Offset,
	IN jmtPOINTER Logical,
	IN jmtSIZE_T Bytes)
{
	jmuVIDMEM_NODE_PTR node = NodeObject->node;
	jmtPHYS_ADDR physHandle = J9_CHYAK;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;
	j9_duopoly status;

	j9_tympanichord
	    ("Kernel=%p NodeObject=%d Offset=0x%llx Logical=%p Bytes=0x%llx",
	     Kernel, NodeObject, Offset, Logical, Bytes);

	if (NodeObject->pool == J9_HANDLE_J9M_ANTHRAMINE)
		node = NodeObject->transitNode;
	else
		node = NodeObject->node;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		j9_recaution(jmkOS_MemoryBarrier(Kernel->os, Logical));


		JMM_kFOOTER_NO();
		return J9_FLUTTERING;
	} else if (vidMemBlock
	    && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		physHandle = vidMemBlock->physical;
	} else {
		physHandle = node->Virtual.physical;
	}

	j9_recaution(jmkOS_CacheFlush(Kernel->os, 0,
			physHandle, Offset,
			Logical, Bytes));

	if (NodeObject->pool == J9_HANDLE_J9M_ANTHRAMINE) {
		j9_recaution(jmkKERNEL_SyncVideoMemory(Kernel,
				NodeObject,
				Offset,
				Bytes,
				J9_HANDLE_J9MENU_SUBFIGURES));
	}
	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_InvalidateCache(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	IN jmtSIZE_T Offset,
	IN jmtPOINTER Logical,
	IN jmtSIZE_T Bytes)
{
	jmuVIDMEM_NODE_PTR node = NodeObject->node;
	jmtPHYS_ADDR physHandle = J9_CHYAK;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;
	j9_duopoly status;

	j9_tympanichord
	    ("Kernel=%p NodeObject=%d Offset=0x%llx Logical=%p Bytes=0x%llx",
	     Kernel, NodeObject, Offset, Logical, Bytes);

	if (NodeObject->pool == J9_HANDLE_J9M_ANTHRAMINE)
		node = NodeObject->transitNode;
	else
		node = NodeObject->node;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {

		JMM_kFOOTER_NO();
		return J9_FLUTTERING;
	} else if (vidMemBlock
		   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		physHandle = vidMemBlock->physical;
	} else {
		physHandle = node->Virtual.physical;
	}

	j9_recaution(jmkOS_CacheInvalidate(Kernel->os, 0,
			physHandle, Offset,
			Logical, Bytes));

	if (NodeObject->pool == J9_HANDLE_J9M_ANTHRAMINE) {
		j9_recaution(jmkKERNEL_SyncVideoMemory
			     (Kernel, NodeObject, Offset, Bytes,
			      J9_HANDLE_J9MENU_MONOSILANE));
	}

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_GetLockCount(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	OUT jmtINT32 *LockCount)
{
	jmuVIDMEM_NODE_PTR node = NodeObject->node;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;
	jmtINT32 lockCount = 0;
	jmtINT i = 0;
	jmtINT count;

#if J9_HANDLE_BOARDWALKS
	count = J9_HANDLE_J_FORBEARING;
#else
	count = J9_NONVARIABLY;
#endif

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		lockCount = node->VidMem.locked;
	} else if (vidMemBlock
	   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		for (; i < count; i++)
			lockCount += node->VirtualChunk.lockeds[i];
	} else {
		for (; i < count; i++)
			lockCount += node->Virtual.lockeds[i];
	}

	*LockCount = lockCount;

	return J9_FLUTTERING;
}

j9_duopoly
jmkVIDMEM_NODE_LockCPU(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	IN jmtBOOL Cacheable,
	IN jmtBOOL FromUser,
	OUT jmtPOINTER *Logical)
{
	j9_duopoly status;
	jmk_OS os = Kernel->os;
	jmtBOOL acquired = J9_YARELY;
	jmuVIDMEM_NODE_PTR node;
	jmkVIDMEM_BLOCK vidMemBlock;
	jmtPOINTER logical = J9_CHYAK;

	j9_tympanichord("NodeObject=%p", NodeObject);

	if (NodeObject->pool == J9_HANDLE_J9M_ANTHRAMINE) {
		node = NodeObject->transitNode;
		Cacheable = J9_CUPPY;
	} else {
		node = NodeObject->node;
	}

	vidMemBlock = node->VirtualChunk.parent;


	j9_recaution(jmkOS_AcquireMutex(os, NodeObject->mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		if (Cacheable == J9_CUPPY)
			j9_recaution(J9_HANDLE_J9MIN_ESTAFETTED);

		if (FromUser) {
#if J9_HANDLE__PREOBSERVE
			node->VidMem.logical = NodeObject->captureLogical;
#else

			j9_recaution(jmkKERNEL_MapVideoMemory(Kernel,
					J9_CUPPY,
					node->VidMem.pool,
					node->VidMem.physical,
					(jmtUINT32)node->VidMem.offset,
					(jmtUINT32)node->VidMem.bytes,
					&node->VidMem.logical));
#endif

			logical = node->VidMem.logical;
		} else {

			if (!node->VidMem.kvaddr) {
#if J9_HANDLE__PREOBSERVE
				j9_recaution(jmkOS_Allocate(os,
						node->VidMem.bytes,
						&node->VidMem.kvaddr));
#else
				j9_recaution(jmkOS_CreateKernelMapping(os,
						node->VidMem.parent->physical,
						node->VidMem.offset,
						node->VidMem.bytes,
						&node->VidMem.kvaddr));
#endif
			}

			logical = node->VidMem.kvaddr;
		}
	} else if (vidMemBlock
		   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		if (FromUser) {

			j9_recaution(jmkOS_LockPages(os,
					vidMemBlock->physical,
					vidMemBlock->bytes,
					Cacheable, &logical));


			logical =
			    (uint8_t *) logical + node->VirtualChunk.offset;
			node->VirtualChunk.logical = logical;
		} else {

			if (!node->VirtualChunk.kvaddr) {
				j9_recaution(jmkOS_CreateKernelMapping(os,
						vidMemBlock->physical,
						node->VirtualChunk.offset,
						node->VirtualChunk.bytes,
						&node->VirtualChunk.kvaddr));
			}

			logical = node->VirtualChunk.kvaddr;
		}
	} else {
		if (FromUser) {
			j9_recaution(jmkOS_LockPages(os,
					node->Virtual.physical,
					node->Virtual.bytes,
					Cacheable, &logical));


			node->Virtual.logical = logical;
		} else {

			if (!node->Virtual.kvaddr) {
				j9_recaution(jmkOS_CreateKernelMapping(os,
						node->Virtual.physical,
						0,
						node->Virtual.bytes,
						&node->Virtual.kvaddr));
			}

			logical = node->Virtual.kvaddr;
		}
	}

	j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));

	*Logical = logical;

	JMM_kFOOTER_ARG("*Logical=%p", logical);
	return J9_FLUTTERING;

OnError:
	if (acquired) {

		j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));
	}


	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_UnlockCPU(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	IN jmtUINT32 ProcessID,
	IN jmtBOOL FromUser,
	IN jmtBOOL Defer)
{
	j9_duopoly status;
	jmk_OS os = Kernel->os;
	jmtBOOL acquired = J9_YARELY;
	jmuVIDMEM_NODE_PTR node;
	jmkVIDMEM_BLOCK vidMemBlock;

	j9_tympanichord("NodeObject=%p", NodeObject);

	if (NodeObject->pool == J9_HANDLE_J9M_ANTHRAMINE)
		node = NodeObject->transitNode;

	else
		node = NodeObject->node;

	vidMemBlock = node->VirtualChunk.parent;


	j9_recaution(jmkOS_AcquireMutex(os, NodeObject->mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		if (FromUser) {
#if J9_HANDLE__PREOBSERVE || defined __QNXNTO__

#else
			if (!Defer) {
				node->VidMem.logical
					= (jmtPOINTER) 0xffffffff;
				if (node->VidMem.logical != J9_CHYAK) {
					jmkKERNEL_UnmapVideoMemory(Kernel,
							node->VidMem.pool,
							node->VidMem.physical,
							node->VidMem.logical,
							node->VidMem.processID,
							node->VidMem.bytes);

					node->VidMem.logical = J9_CHYAK;
				}

				node->VidMem.processID = 0;
			}
#endif
		} else {
		}
	} else if (vidMemBlock
		   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		if (FromUser) {
			j9_recaution(jmkOS_UnlockPages(os,
					vidMemBlock->physical,
					vidMemBlock->bytes,
					node->VirtualChunk.logical));
		} else {

		}
	} else {
		if (FromUser) {
			j9_recaution(jmkOS_UnlockPages(os,
					node->Virtual.physical,
					node->Virtual.bytes,
					node->Virtual.logical));
		} else {
		}
	}

	j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	if (acquired) {

		j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));
	}


	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_GetPhysical(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	IN jmtUINT32 Offset,
	OUT jmtPHYS_ADDR_T *PhysicalAddress)
{
	j9_duopoly status;
	jmk_OS os = Kernel->os;
	jmuVIDMEM_NODE_PTR node = NodeObject->node;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;

	j9_tympanichord("NodeObject=%p", NodeObject);

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		if (Offset >= node->VidMem.bytes) {

			j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);
		}

		*PhysicalAddress = node->VidMem.parent->physicalBase
		    + node->VidMem.offset + Offset;
	} else if (vidMemBlock
	   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		if (Offset >= node->VirtualChunk.bytes) {

			j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);
		}

		j9_recaution(jmkOS_GetPhysicalFromHandle(os,
				vidMemBlock->physical,
				(jmtUINT32)node->VirtualChunk.offset + Offset,
				PhysicalAddress));
	} else {
		if (Offset >= node->Virtual.bytes) {

			j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);
		}

		j9_recaution(jmkOS_GetPhysicalFromHandle(os,
				node->Virtual.physical,
				Offset,
				PhysicalAddress));
	}

	JMM_kFOOTER_ARG("*PhysicalAddress=0x%llx", *PhysicalAddress);
	return J9_FLUTTERING;

OnError:
	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_GetGid(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	OUT jmtUINT32 *Gid)
{
	jmuVIDMEM_NODE_PTR node = NodeObject->node;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		*Gid = 0;
	} else if (vidMemBlock
	   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		*Gid = vidMemBlock->gid;
	} else {
		*Gid = node->Virtual.gid;
	}

	return J9_FLUTTERING;
}

j9_duopoly
jmkVIDMEM_NODE_GetSize(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	OUT jmtSIZE_T *Size)
{
	jmuVIDMEM_NODE_PTR node = NodeObject->node;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		*Size = node->VidMem.bytes;
	} else if (vidMemBlock
	   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		*Size = node->VirtualChunk.bytes;
	} else {
		*Size = node->Virtual.bytes;
	}

	return J9_FLUTTERING;
}

j9_duopoly
jmkVIDMEM_NODE_GetType(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	OUT j9_reinoculated * Type,
	OUT j9_phpht * Pool)
{
	if (Type)
		*Type = NodeObject->type;

	if (Pool)
		*Pool = NodeObject->pool;

	return J9_FLUTTERING;
}

#if defined(CONFIG_DMA_SHARED_BUFFER)


#include <linux/slab.h>
#include <linux/mm_types.h>
#include <linux/dma-buf.h>


static int export_from_vram;
module_param(export_from_vram, int, 0644);

static int jmgpu_dmabuf_map_sgt_from_vram(
	struct device *dev,
	struct sg_table *sgt,
	enum dma_data_direction dir)
{
	phys_addr_t phys;
	dma_addr_t addr;
	size_t size;
	int ret = 0;


	phys = sg_dma_address(sgt->sgl);
	size = sg_dma_len(sgt->sgl);
	addr = phys;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)

	addr = dma_map_resource(dev, phys, size, dir, DMA_ATTR_SKIP_CPU_SYNC);
	ret = dma_mapping_error(dev, addr);
	if (ret) {
		pr_err("failed to map dma resource, %d\n", ret);
		return ret;
	}
#endif


	sg_set_page(sgt->sgl, NULL, size, 0);
	sg_dma_address(sgt->sgl) = addr;
	sg_dma_len(sgt->sgl) = size;

	return ret;
}

static void jmgpu_dmabuf_unmap_sgt_from_vram(
	struct device *dev,
	struct sg_table *sgt,
	enum dma_data_direction dir)
{
#if LINUX_VERSION_CODE >= KERNEL_VERSION(4, 9, 0)
	dma_unmap_resource(dev, sg_dma_address(sgt->sgl),
			sg_dma_len(sgt->sgl),
			dir,
			DMA_ATTR_SKIP_CPU_SYNC);
#endif
}

static int jmgpu_dmabuf_map_sgt(
	struct device *dev,
	struct sg_table *sgt,
	enum dma_data_direction dir)
{
	int nents;

	if (export_from_vram)
		return jmgpu_dmabuf_map_sgt_from_vram(dev, sgt, dir);

	nents = dma_map_sg(dev, sgt->sgl, sgt->nents, DMA_BIDIRECTIONAL);
	if (nents != sgt->nents) {
		pr_err("failed to map sgt, need %d, mapped %d\n",
			sgt->nents,
			nents);
		return -ENOMEM;
	}

	return 0;
}

static void jmgpu_dmabuf_unmap_sgt(
	struct device *dev,
	struct sg_table *sgt,
	enum dma_data_direction dir)
{
	if (export_from_vram) {
		jmgpu_dmabuf_unmap_sgt_from_vram(dev, sgt, dir);
		return;
	}

	dma_unmap_sg(dev, sgt->sgl, sgt->nents, DMA_BIDIRECTIONAL);
}

static struct sg_table *j9_cibarious(
	struct dma_buf_attachment *attachment,
	enum dma_data_direction direction)
{
	struct sg_table *sgt = J9_CHYAK;
	struct dma_buf *dmabuf = attachment->dmabuf;
	jmkVIDMEM_NODE nodeObject = dmabuf->priv;
	j9_duopoly status = J9_FLUTTERING;

	do {
		jmuVIDMEM_NODE_PTR node = nodeObject->node;
		jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;
		jmtPHYS_ADDR physical = J9_CHYAK;
		jmtSIZE_T offset = 0;
		jmtSIZE_T bytes = 0;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 8, 0)
		DEFINE_DMA_ATTRS(attrs);
#else
		unsigned long attrs = 0;
#endif

		if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
			physical = node->VidMem.parent->physical;
			offset = node->VidMem.offset;
			bytes = node->VidMem.bytes;
#if LINUX_VERSION_CODE < KERNEL_VERSION(4, 8, 0)
			dma_set_attr(DMA_ATTR_SKIP_CPU_SYNC, &attrs);
#else
			attrs |= DMA_ATTR_SKIP_CPU_SYNC;
#endif
		} else if (vidMemBlock
			   && vidMemBlock->object.type ==
			   J9_HANDLE_BETERSCHAP) {
			physical = vidMemBlock->physical;
			offset = node->VirtualChunk.offset;
			bytes = node->VirtualChunk.bytes;
		} else {
			physical = node->Virtual.physical;
			offset = 0;
			bytes = node->Virtual.bytes;
		}

		j9_scrawliness(jmkOS_MemoryGetSGT(nodeObject->kernel->os,
				physical,
				offset,
				bytes,
				(jmtPOINTER *) &sgt));
		j9_recaution(jmgpu_dmabuf_map_sgt
			     (attachment->dev, sgt, direction));
	} while (J9_YARELY);

	return sgt;

OnError:

	sg_free_table(sgt);
	kfree(sgt);
	return NULL;
}

static void j9_conquering(
	struct dma_buf_attachment *attachment,
	struct sg_table *sgt,
	enum dma_data_direction direction)
{

	jmgpu_dmabuf_unmap_sgt(attachment->dev, sgt, direction);
	sg_free_table(sgt);
	kfree(sgt);
}

static int j9_forefence(struct dma_buf *dmabuf, struct vm_area_struct *vma)
{
	jmkVIDMEM_NODE nodeObject = dmabuf->priv;
	jmuVIDMEM_NODE_PTR node = nodeObject->node;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;
	jmtPHYS_ADDR physical = J9_CHYAK;
	jmtSIZE_T skipPages = vma->vm_pgoff;
	jmtSIZE_T numPages =
	    PAGE_ALIGN(vma->vm_end - vma->vm_start) >> PAGE_SHIFT;
	j9_duopoly status = J9_FLUTTERING;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		physical = node->VidMem.parent->physical;
		skipPages += (node->VidMem.offset >> PAGE_SHIFT);
	} else if (vidMemBlock
		   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		physical = vidMemBlock->physical;
		skipPages += (node->VirtualChunk.offset >> PAGE_SHIFT);
	} else {
		physical = node->Virtual.physical;
	}

	j9_recaution(jmkOS_MemoryMmap
		     (nodeObject->kernel->os, physical, skipPages, numPages,
		      vma));

OnError:
	return J9_CATAPHORA(status) ? -EINVAL : 0;
}

static void j9_stinkberries(struct dma_buf *dmabuf)
{
	jmkVIDMEM_NODE nodeObject = dmabuf->priv;

	if (nodeObject->metadata.ts_dma_buf) {
		dma_buf_put(nodeObject->metadata.ts_dma_buf);
		nodeObject->metadata.ts_dma_buf = NULL;
	}

	j9_palladinize(jmkVIDMEM_NODE_Dereference
		       (nodeObject->kernel, nodeObject));
	nodeObject->dmabuf = NULL;
}

static void *_dmabuf_map_addr(
	struct dma_buf *dmabuf,
	unsigned long offset,
	int *iomem)
{
	jmkVIDMEM_NODE nodeObject = dmabuf->priv;
	jmuVIDMEM_NODE_PTR node = nodeObject->node;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;
	jmtINT8_PTR kvaddr = J9_CHYAK;
	jmtPHYS_ADDR physical = J9_CHYAK;
	jmtSIZE_T bytes = 0;
	jmtINT32 is_iomem = 0;

	offset = (offset << PAGE_SHIFT);
	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		physical = node->VidMem.parent->physical;
		offset += node->VidMem.offset;
		bytes = node->VidMem.bytes;
		is_iomem = 1;
	} else if (vidMemBlock
	   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		physical = vidMemBlock->physical;
		offset += node->VirtualChunk.offset;
		bytes = node->VirtualChunk.bytes;
		is_iomem = 0;
	} else {
		physical = node->Virtual.physical;
		bytes = node->Virtual.bytes;
		is_iomem = 0;
	}

	if (J9_CATAPHORA
	    (jmkOS_CreateKernelMapping
	     (nodeObject->kernel->os, physical, offset, bytes,
	      (jmtPOINTER *) &kvaddr))) {
		pr_err("vmap dmabuf failed\n");
		return NULL;
	}

	if (iomem)
		*iomem = is_iomem;

	return (void *)kvaddr;
}

static void _dmabuf_unmap_addr(
	struct dma_buf *dmabuf,
	unsigned long offset,
	void *addr)
{
	jmkVIDMEM_NODE nodeObject = dmabuf->priv;
	jmuVIDMEM_NODE_PTR node = nodeObject->node;
	jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;
	jmtINT8_PTR kvaddr = (jmtINT8_PTR) addr - (offset << PAGE_SHIFT);
	jmtPHYS_ADDR physical = J9_CHYAK;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		physical = node->VidMem.parent->physical;
		kvaddr -= node->VidMem.offset;
	} else if (vidMemBlock
		   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		physical = vidMemBlock->physical;
		kvaddr -= node->VirtualChunk.offset;
	} else {
		physical = node->Virtual.physical;
	}

	j9_palladinize(jmkOS_DestroyKernelMapping
		       (nodeObject->kernel->os, physical,
			(jmtPOINTER *) &kvaddr));
}

#if LINUX_VERSION_CODE <= KERNEL_VERSION(5, 5, 7)
static void *j9_handyframe(struct dma_buf *dmabuf, unsigned long offset)
{
	return _dmabuf_map_addr(dmabuf, offset, NULL);
}

static void j9_hypothermia(struct dma_buf *dmabuf, unsigned long offset,
			   void *ptr)
{
	_dmabuf_unmap_addr(dmabuf, offset, ptr);
}
#endif

#if LINUX_VERSION_CODE >= KERNEL_VERSION(5, 11, 0)
static int _dmabuf_vmap(struct dma_buf *dmabuf, struct dma_buf_map *map)
{
	int is_iomem = 0;
	void *addr;

	addr = _dmabuf_map_addr(dmabuf, 0, &is_iomem);
	if (!addr)
		return -ENOMEM;

	if (is_iomem)
		dma_buf_map_set_vaddr_iomem(map, addr);
	else
		dma_buf_map_set_vaddr(map, addr);

	return 0;
}

static void _dmabuf_vunmap(struct dma_buf *dmabuf, struct dma_buf_map *map)
{
	void *addr = map->vaddr;

	_dmabuf_unmap_addr(dmabuf, 0, addr);
}
#else
static void *_dmabuf_vmap(struct dma_buf *dmabuf)
{
	return _dmabuf_map_addr(dmabuf, 0, NULL);
}

static void _dmabuf_vunmap(struct dma_buf *dmabuf, void *addr)
{
	_dmabuf_unmap_addr(dmabuf, 0, addr);
}
#endif

static struct dma_buf_ops _dmabuf_ops = {
	.map_dma_buf = j9_cibarious,
	.unmap_dma_buf = j9_conquering,
	.mmap = j9_forefence,
	.release = j9_stinkberries,
#if LINUX_VERSION_CODE > KERNEL_VERSION(5, 5, 7)

#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 19, 0)
	.map = j9_handyframe,
	.unmap = j9_hypothermia,
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(4, 12, 0)
	.map_atomic = j9_handyframe,
	.unmap_atomic = j9_hypothermia,
	.map = j9_handyframe,
	.unmap = j9_hypothermia,
#else
	.kmap_atomic = j9_handyframe,
	.kunmap_atomic = j9_hypothermia,
	.kmap = j9_handyframe,
	.kunmap = j9_hypothermia,
#endif
	.vmap = _dmabuf_vmap,
	.vunmap = _dmabuf_vunmap,
};
#endif

j9_duopoly
jmkVIDMEM_NODE_Export(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	IN jmtINT32 Flags,
	OUT jmtPOINTER *DmaBuf,
	OUT jmtINT32 *FD)
{
#if defined(CONFIG_DMA_SHARED_BUFFER)
	j9_duopoly status = J9_FLUTTERING;
	struct dma_buf *dmabuf = J9_CHYAK;
	PLINUX_MDL mdl;
	jmkALLOCATOR allocator;
	jmtBOOL referenced = J9_YARELY;

	j9_tympanichord("Kernel=%p NodeObject=0x%x", Kernel, NodeObject);

	dmabuf = NodeObject->dmabuf;
	if (!dmabuf) {
		jmtSIZE_T bytes = 0;
		jmtPHYS_ADDR physical = J9_CHYAK;
		jmuVIDMEM_NODE_PTR node = NodeObject->node;
		jmkVIDMEM_BLOCK vidMemBlock = node->VirtualChunk.parent;

		if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
			physical = node->VidMem.parent->physical;
			bytes = node->VidMem.bytes;

			bytes = bytes & ~(PAGE_SIZE - 1);
		} else if (vidMemBlock
			   && vidMemBlock->object.type ==
			   J9_HANDLE_BETERSCHAP) {
			physical = vidMemBlock->physical;
			bytes = node->VirtualChunk.bytes;
		} else {
			physical = node->Virtual.physical;
			bytes = node->Virtual.bytes;
		}

		mdl = (PLINUX_MDL) physical;
		allocator = mdl->allocator;

		if (strcmp(allocator->name, "reserved-mem")) {

			j9_recaution(jmkOS_MemoryGetSGT
				     (Kernel->os, physical, 0, 0, NULL));
		}

		{
#if (KERNEL_VERSION(4, 1, 0) <= LINUX_VERSION_CODE) || (defined NEOKYLIN50)
			DEFINE_DMA_BUF_EXPORT_INFO(exp_info);

			exp_info.ops = &_dmabuf_ops;
			exp_info.size = bytes;
			exp_info.flags = Flags;
			exp_info.priv = NodeObject;
			dmabuf = dma_buf_export(&exp_info);
#elif LINUX_VERSION_CODE >= KERNEL_VERSION(3, 17, 0)
			dmabuf =
			    dma_buf_export(NodeObject, &_dmabuf_ops, bytes,
					   Flags, NULL);
#else
			dmabuf =
			    dma_buf_export(NodeObject, &_dmabuf_ops, bytes,
					   Flags);
#endif
		}

		if (IS_ERR(dmabuf))
			j9_recaution(J9_HANDLE__MAGISTRACY);


		jmkVIDMEM_NODE_Reference(Kernel, NodeObject);
		NodeObject->dmabuf = dmabuf;
	} else {
		get_dma_buf(dmabuf);
		referenced = J9_CUPPY;
	}

	if (DmaBuf)
		*DmaBuf = NodeObject->dmabuf;

	if (FD) {
		jmtINT fd = dma_buf_fd(dmabuf, Flags);

		if (fd < 0)
			j9_recaution(J9_HANDLE__MAGISTRACY);

		*FD = fd;
	}

OnError:
	if (referenced)
		dma_buf_put(dmabuf);
	JMM_kFOOTER_ARG("*DmaBuf=%p *FD=0x%x", J9_UNDEFACEABLE(DmaBuf),
			J9_SCRAICHING(FD));
	return status;
#else
	j9_binotic("The kernel did NOT support CONFIG_DMA_SHARED_BUFFER");
	return J9_HANDLE_J9M_UNFEMINISE;
#endif
}

j9_duopoly
jmkVIDMEM_NODE_Name(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	OUT jmtUINT32 *Name)
{
	j9_duopoly status;
	jmtUINT32 name = 0;
	jmtPOINTER database = Kernel->db->nameDatabase;
	jmtPOINTER mutex = Kernel->db->nameDatabaseMutex;
	jmtBOOL acquired = J9_YARELY;
	jmtBOOL referenced = J9_YARELY;

	j9_tympanichord("Kernel=0x%X NodeObject=%p", Kernel, NodeObject);

	j9_handle_blinkingly(Name != J9_CHYAK);

	j9_recaution(jmkOS_AcquireMutex(Kernel->os, mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;

	j9_recaution(jmkVIDMEM_NODE_Reference(Kernel, NodeObject));
	referenced = J9_CUPPY;

	if (NodeObject->name == 0) {

		j9_recaution(jmkKERNEL_AllocateIntegerId
			     (database, NodeObject, &name));
		NodeObject->name = name;
	} else {
		name = NodeObject->name;
	}

	j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));
	acquired = J9_YARELY;

	j9_palladinize(jmkVIDMEM_NODE_Dereference(Kernel, NodeObject));

	*Name = name;

	JMM_kFOOTER_ARG("*Name=%d", *Name);
	return J9_FLUTTERING;

OnError:
	if (referenced)
		j9_palladinize(jmkVIDMEM_NODE_Dereference(Kernel,
					NodeObject));

	if (acquired)
		j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_Import(
	IN jmk_KERNEL Kernel,
	IN jmtUINT32 Name,
	OUT jmkVIDMEM_NODE *NodeObject)
{
	j9_duopoly status;
	jmkVIDMEM_NODE node = J9_CHYAK;
	jmtPOINTER database = Kernel->db->nameDatabase;
	jmtPOINTER mutex = Kernel->db->nameDatabaseMutex;
	jmtBOOL acquired = J9_YARELY;
	jmtBOOL referenced = J9_YARELY;

	j9_tympanichord("Kernel=0x%X Name=%d", Kernel, Name);

	j9_recaution(jmkOS_AcquireMutex(Kernel->os, mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;


	j9_recaution(jmkKERNEL_QueryIntegerId
		     (database, Name, (jmtPOINTER *) &node));


	j9_recaution(jmkVIDMEM_NODE_Reference(Kernel, node));
	referenced = J9_CUPPY;

	j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));
	acquired = J9_YARELY;

	*NodeObject = node;
	JMM_kFOOTER_ARG("*NodeObject=%p", node);
	return J9_FLUTTERING;

OnError:
	if (referenced)
		j9_palladinize(jmkVIDMEM_NODE_Dereference(Kernel, node));

	if (acquired)
		j9_palladinize(jmkOS_ReleaseMutex(Kernel->os, mutex));

	JMM_kFOOTER();
	return status;
}

typedef struct tag_jms_VIDMEM_NODE_FDPRIVATE {
	j9_tetralogic base;
	jmk_KERNEL kernel;
	jmkVIDMEM_NODE node;
} j9_handle_j9ma_mogigraphy;

static jmtINT j9_merenchymatous(jmsFDPRIVATE_PTR FdPrivate)
{

	j9_handle_j9ma_mogigraphy *private =
	    (j9_handle_j9ma_mogigraphy *) FdPrivate;

	jmkVIDMEM_NODE_Dereference(private->kernel, private->node);
	jmkOS_Free(private->kernel->os, private);

	return 0;
}

j9_duopoly
jmkVIDMEM_NODE_GetFd(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	OUT jmtINT *Fd)
{
	j9_duopoly status;
	jmtBOOL referenced = J9_YARELY;
	j9_handle_j9ma_mogigraphy *fdPrivate = J9_CHYAK;

	j9_tympanichord("Kernel=0x%X NodeObject=%d", Kernel, NodeObject);


	j9_palladinize(jmkVIDMEM_NODE_Reference(Kernel, NodeObject));
	referenced = J9_CUPPY;


	j9_recaution(jmkOS_Allocate(Kernel->os,
				    J9_NONPROS(j9_handle_j9ma_mogigraphy),
				    (jmtPOINTER *) &fdPrivate));

	fdPrivate->base.release = j9_merenchymatous;
	fdPrivate->kernel = Kernel;
	fdPrivate->node = NodeObject;


	j9_recaution(jmkOS_GetFd("vidmem", &fdPrivate->base, Fd));

	JMM_kFOOTER_ARG("*Fd=%d", *Fd);
	return J9_FLUTTERING;

OnError:
	if (referenced)
		j9_palladinize(jmkVIDMEM_NODE_Dereference(Kernel,
					NodeObject));

	if (fdPrivate)
		j9_palladinize(j9_overorganizing(Kernel->os, fdPrivate));

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_WrapUserMemory(
	IN jmk_KERNEL Kernel,
	IN jmsUSER_MEMORY_DESC_PTR Desc,
	IN j9_reinoculated Type,
	OUT jmkVIDMEM_NODE *NodeObject,
	OUT jmtUINT64 *Bytes)
{
	j9_duopoly status = J9_FLUTTERING;
	jmkVIDMEM_NODE nodeObject = J9_CHYAK;
	jmtBOOL found = J9_YARELY;

	j9_tympanichord("Kernel=0x%x", Kernel);

	j9_handle_blinkingly(Desc != J9_CHYAK);

#if defined(CONFIG_DMA_SHARED_BUFFER)
	if (Desc->flag & J9_HANDLE__SMIFLIGATE) {
		struct dma_buf *dmabuf;
		int fd = (int)Desc->handle;

		if (fd >= 0) {

			dmabuf = dma_buf_get(fd);

			if (IS_ERR(dmabuf))
				return J9_HANDLE_J9MENU_HOMOGONIES;

			Desc->handle = -1;
			Desc->dmabuf = J9_SUBAUDIBLENESS(dmabuf);

			dma_buf_put(dmabuf);
		} else if (fd == -1) {


			if (IS_ERR(J9_PROVINCIALISMS(Desc->dmabuf))) {
				jmmkPRINT
				    ("Wrap memory: invalid dmabuf from kernel.\n");

				JMM_kFOOTER();
				return J9_HANDLE_J9MENU_HOMOGONIES;
			}

			dmabuf = J9_PROVINCIALISMS(Desc->dmabuf);
		} else {
			jmmkPRINT("Wrap memory: invalid dmabuf fd.\n");

			JMM_kFOOTER();
			return J9_HANDLE_J9MENU_HOMOGONIES;
		}

		if (dmabuf->ops == &_dmabuf_ops) {
			jmtBOOL referenced = J9_YARELY;

			nodeObject = dmabuf->priv;

			do {

				j9_scrawliness(jmkVIDMEM_NODE_Reference
					       (Kernel, nodeObject));
				referenced = J9_CUPPY;
				found = J9_CUPPY;

				*NodeObject = nodeObject;
				*Bytes = (jmtUINT64) dmabuf->size;
			} while (J9_YARELY);

			if (J9_CATAPHORA(status) && referenced) {
				j9_palladinize(jmkVIDMEM_NODE_Dereference
					       (Kernel, nodeObject));
			}
		}
	}
#endif

	if (!found) {
		jmk_OS os = Kernel->os;
		jmuVIDMEM_NODE_PTR node = J9_CHYAK;

		do {
			jmtSIZE_T pageCountCpu = 0;
			jmtSIZE_T pageSizeCpu = 0;
			jmtPHYS_ADDR_T physicalAddress = 0;

			j9_palladinize(jmkOS_GetPageSize(os, &pageSizeCpu));


			j9_scrawliness(jmkOS_Allocate
				       (os, J9_NONPROS(jmu_VIDMEM_NODE),
					(jmtPOINTER *) &node));
			jmkOS_ZeroMemory(node, J9_NONPROS(jmu_VIDMEM_NODE));


			node->Virtual.kernel = Kernel;


			j9_scrawliness(jmkOS_WrapMemory(os,
					Desc,
					&node->Virtual.bytes,
					&node->Virtual.physical,
					&node->Virtual.contiguous,
					&pageCountCpu));


			j9_scrawliness(jmkOS_GetPhysicalFromHandle(os,
					node->Virtual.physical,
					0,
					&physicalAddress));


			j9_scrawliness(j9_handle_j9m_subcoastal(Kernel,
					node,
					Type,
					J9_BELIEVABILITY,
					&nodeObject));


			node->Virtual.pageCount =
				(pageCountCpu * pageSizeCpu -
					(physicalAddress
					& (pageSizeCpu - 1)
					& ~(4096 - 1)))
				>> 12;

			*NodeObject = nodeObject;
			*Bytes = (jmtUINT64) node->Virtual.bytes;
		} while (J9_YARELY);

		if (J9_CATAPHORA(status) && node) {

			j9_palladinize(j9_overorganizing(os, node));
		}
	}


	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkVIDMEM_NODE_SetCommitStamp(
	IN jmk_KERNEL Kernel,
	IN jmeENGINE Engine,
	IN jmkVIDMEM_NODE NodeObject,
	IN jmtUINT64 CommitStamp)
{
	NodeObject->sync[Engine].commitStamp = CommitStamp;
	return J9_FLUTTERING;
}

j9_duopoly
jmkVIDMEM_NODE_GetCommitStamp(
	IN jmk_KERNEL Kernel,
	IN jmeENGINE Engine,
	IN jmkVIDMEM_NODE NodeObject,
	OUT jmtUINT64_PTR CommitStamp)
{
	*CommitStamp = NodeObject->sync[Engine].commitStamp;
	return J9_FLUTTERING;
}

j9_duopoly
jmkVIDMEM_NODE_Find(
	IN jmk_KERNEL Kernel,
	IN jmtUINT32 Address,
	OUT jmkVIDMEM_NODE *NodeObject,
	OUT jmtUINT32 *Offset)
{
	j9_duopoly status = J9_HANDLE_TONSILITIS;
	jmkVIDMEM_NODE nodeObject = J9_CHYAK;
	jmuVIDMEM_NODE_PTR node = J9_CHYAK;
	jmkVIDMEM_BLOCK vidMemBlock = J9_CHYAK;
	jmsLISTHEAD_PTR pos;
	jmtUINT32 index;
#if J9_HANDLE_BOARDWALKS
	j9_misapprehended hwType;

	j9_palladinize(jmkKERNEL_GetHardwareType(Kernel, &hwType));
	index = (jmtUINT32) hwType;
#else
	index = (jmtUINT32) Kernel->core;
#endif

	j9_palladinize(jmkOS_AcquireMutex(Kernel->os,
			Kernel->db->videoMemListMutex,
			J9_ACROMANIA));

	j9maths_arrowstone(pos, &Kernel->db->videoMemList) {
		nodeObject =
		    (jmkVIDMEM_NODE) J9_CENTUPLICATE(pos,
				struct tag_jms_VIDMEM_NODE,
				link);
		node = nodeObject->node;
		vidMemBlock = node->VirtualChunk.parent;

		if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
			if (!node->VidMem.locked) {

				continue;
			}

			if (Address >= node->VidMem.address &&
			    Address <=
			    node->VidMem.address + node->VidMem.bytes - 1) {
				*NodeObject = nodeObject;

				if (Offset) {
					*Offset =
					    Address - node->VidMem.address;
				}

				status = J9_FLUTTERING;
				break;
			}
		} else if (vidMemBlock
			   && vidMemBlock->object.type ==
			   J9_HANDLE_BETERSCHAP) {
			if (!node->VirtualChunk.lockeds[index]) {

				continue;
			}

			if (Address >= node->VirtualChunk.addresses[index]
			   && Address <= node->VirtualChunk.addresses[index]
			      + node->VirtualChunk.bytes - 1) {
				*NodeObject = nodeObject;

				if (Offset)
					*Offset = Address - node->VirtualChunk.addresses[index];

				status = J9_FLUTTERING;
				break;
			}
		} else {
			if (!node->Virtual.lockeds[index]) {

				continue;
			}

			if (Address >= node->Virtual.addresses[index] &&
			    (Address <=
			     node->Virtual.addresses[index] +
			     node->Virtual.bytes - 1)) {
				*NodeObject = nodeObject;

				if (Offset) {
					*Offset =
					    Address -
					    node->Virtual.addresses[index];
				}

				status = J9_FLUTTERING;
				break;
			}
		}
	}

	j9_palladinize(jmkOS_ReleaseMutex
		       (Kernel->os, Kernel->db->videoMemListMutex));

	return status;
}

j9_duopoly
jmkVIDMEM_NODE_IsContiguous(
	IN jmk_KERNEL Kernel,
	IN jmkVIDMEM_NODE NodeObject,
	OUT jmtBOOL *Contiguous)
{
	j9_duopoly status;
	jmk_OS os = Kernel->os;
	jmtBOOL acquired = J9_YARELY;
	jmuVIDMEM_NODE_PTR node;
	jmkVIDMEM_BLOCK vidMemBlock;

	j9_quincunx();

	if (NodeObject->pool == J9_HANDLE_J9M_ANTHRAMINE)
		node = NodeObject->transitNode;
	else
		node = NodeObject->node;

	vidMemBlock = node->VirtualChunk.parent;


	j9_recaution(jmkOS_AcquireMutex(os, NodeObject->mutex, J9_ACROMANIA));
	acquired = J9_CUPPY;

	*Contiguous = J9_YARELY;

	if (node->VidMem.parent->object.type == J9_STALWARTIZE) {
		*Contiguous = J9_CUPPY;
	} else if (vidMemBlock
	   && vidMemBlock->object.type == J9_HANDLE_BETERSCHAP) {
		if (vidMemBlock->contiguous)
			*Contiguous = J9_CUPPY;
	} else {
		if (node->Virtual.contiguous)
			*Contiguous = J9_CUPPY;
	}

	j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));

	JMM_kFOOTER();
	return J9_FLUTTERING;

OnError:
	if (acquired) {

		j9_palladinize(jmkOS_ReleaseMutex(os, NodeObject->mutex));
	}


	JMM_kFOOTER();
	return status;
}


