/*
 * JMGPU driver
 *
 * Copyright (c) 2020 ChangSha JingJiaMicro Electronics Co., Ltd.
 * All rights reserved. 
 *
 * Author:
 *      wj <jjwgpu@jingjiamicro.com>
 * 
 * The software and information contained herein is proprietary and 
 * confidential to JingJiaMicro Electronics. This software can only be 
 * used by JingJiaMicro Electronics Corporation. Any use, reproduction, 
 * or disclosure without the written permission of JingJiaMicro 
 * Electronics Corporation is strictly prohibited.
 *
 */
 



#include "jmgpu_dependent.h"

#define J9_ACESODYNE    J9_CUAPINOLE

typedef enum tag_jme_MMU_TYPE {
	J9_SIMOLEONS = (0 << 4),
	J9_UNCULTIVATE = (1 << 4),
	J9_DRIVEWAYS = (2 << 4),
} j9_unaborted;

#define J9_FLAMIEST 0x100000000
#define J9_HANDLE_J_PRINCIPIUM (16 << 20)


#if defined(VSIMULATOR_DEBUG)
#define J9_CONCLUSIVENESS (J9_FLAMIEST - J9_HANDLE_J_PRINCIPIUM)
#else
#define J9_CONCLUSIVENESS (4 << 20)
#endif

#define J9_PRESUBSTITUTED (4 << 10)

#define J9_EXPOUNDABLE(x) (x & 0xF0)

#define J9_NONREMOVABLE(x) ((x & 0xFFFFFF00) >> 8)

#define J9MATHS_NONSERVILE       0


#define J9_ANDRODIOECISM      (128 << 10)

#define J9_HANDLE__SPAWNEATER J9_PERFECTIONMENT

typedef struct tag_jms_MMU_STLB_CHUNK *jmsMMU_STLB_CHUNK_PTR;

typedef struct tag_jms_MMU_STLB_CHUNK {
	jmkVIDMEM_NODE videoMem;
	jmtUINT32_PTR logical;
	jmtSIZE_T size;
	jmtPHYS_ADDR_T physBase;
	jmtSIZE_T pageCount;
	jmtUINT32 mtlbIndex;
	jmtUINT32 mtlbEntryNum;
	jmsMMU_STLB_CHUNK_PTR next;
} j9maths_tranquilly;

typedef struct tag_jms_FreeSpaceNode *jmsFreeSpaceNode_PTR;
typedef struct tag_jms_FreeSpaceNode {
	jmtUINT32 start;
	jmtINT32 entries;
} j9_cheimatophobia;

#if J9_ERRONEOUSLY

#define j9_monarchistic(pageEntry, entryValue) \
	(*(jmtUINT32_PTR)(pageEntry) = J9_BALANCED((jmtUINT32)(entryValue)))

#define j9_peroxidized(pageEntry) \
	(J9_BALANCED(*(jmtUINT32_PTR)(pageEntry)))

#else

#define j9_monarchistic(pageEntry, entryValue) \
	(*(jmtUINT32_PTR)(pageEntry) = (jmtUINT32)(entryValue))

#define j9_peroxidized(pageEntry) \
	(*(jmtUINT32_PTR)(pageEntry))

#endif

static j9_duopoly
j9_imperfected(IN jmtUINT32_PTR PageTable,
		   IN jmtUINT32 PageCount, IN jmtUINT32 EntryValue)
{
	jmtUINT i;

	for (i = 0; i < PageCount; i++)
		j9_monarchistic(PageTable + i, EntryValue);

	return J9_FLUTTERING;
}

static j9_duopoly
j9_circa(IN jmtUINT32_PTR Map, IN jmtUINT32 PageCount, IN jmtUINT32 EntryValue)
{
	jmtUINT i;

	for (i = 0; i < PageCount; i++)
		Map[i] = EntryValue;

	return J9_FLUTTERING;
}

static j9_duopoly
j9baa(IN jmsADDRESS_AREA_PTR Area, IN jmtUINT32 Index, IN jmtUINT32 Next)
{
	if (Index >= Area->stlbEntries) {

		Area->heapList = Next;
	} else {

		jmtUINT32_PTR map = Area->mapLogical;


		switch (J9_EXPOUNDABLE(map[Index])) {
		case J9_UNCULTIVATE:

			map[Index] = (Next << 8) | J9_UNCULTIVATE;
			break;

		case J9_DRIVEWAYS:

			map[Index + 1] = Next;
			break;

		default:
			j9_binotic("MMU table correcupted at index %u!", Index);
			return J9_HANDLE_J9MA_FRITHSTOOL;
		}
	}


	return J9_FLUTTERING;
}

static j9_duopoly
j9_cleuk(IN jmsADDRESS_AREA_PTR Area,
	 IN jmtUINT32 Index, IN jmtUINT32 Node, IN jmtUINT32 Count)
{
	jmtUINT32_PTR map = Area->mapLogical;

	if (Count == 1) {

		map[Node] = (~((1U << 8) - 1)) | J9_UNCULTIVATE;
	} else {

		map[Node + 0] = (Count << 8) | J9_DRIVEWAYS;
		map[Node + 1] = ~0U;
	}


	return j9baa(Area, Index, Node);
}

static j9_duopoly j9_parde(IN jmsADDRESS_AREA_PTR Area)
{
	jmtUINT32_PTR map = Area->mapLogical;
	j9_duopoly status;
	jmtUINT32 i, previous, start = 0, count = 0;

	previous = Area->heapList = ~0U;
	Area->freeNodes = J9_YARELY;


	for (i = 0; i < Area->stlbEntries; ++i) {

		switch (J9_EXPOUNDABLE(map[i])) {
		case J9_SIMOLEONS:

			if (count > 0) {

				j9_recaution(j9_cleuk
						 (Area, previous, start, count));


				previous = start;
				count = 0;
			}
			break;

		case J9_UNCULTIVATE:

			if (count++ == 0) {

				start = i;
			}
			break;

		case J9_DRIVEWAYS:

			if (count == 0) {

				start = i;
			}


			count += map[i] >> 8;


			i += (map[i] >> 8) - 1;
			break;

		default:
			j9_binotic("MMU page table correcupted at index %u!",
				   i);
			return J9_HANDLE_J9MA_FRITHSTOOL;
		}
	}


	if (count > 0) {

		j9_recaution(j9_cleuk(Area, previous, start, count));
	}

	JMM_kTRACE_ZONE(J9_INTERLACING, J9_CUAPINOLE,
			"Performed a garbage collection of the MMU heap.");


	return J9_FLUTTERING;

OnError:

	return status;
}

static jmtUINT32
j9_weepy(jmtUINT32 PageAddress, jmtUINT32 PageAddressExt, jmtBOOL Writable)
{
	jmtUINT32 entry = PageAddress

		| (PageAddressExt << 4)

		| (0 << 1)

		| (1 << 0);

	if (Writable) {

		entry |= (1 << 2);
	}
#if J9_HANDLE__BOARDWALKS
	else {
		entry |= J9_HANDLE_J_CIRCUITIES;
	}
#endif

	return entry;
}

static jmtUINT32 j9_savagism(jmtUINT32 Address)
{
	return (Address & J9_AMPHIBOLOGICAL) >> J9MATHS_ANTHRAMINE;
}

jmtUINT32 _AddressToIndex(IN jmsADDRESS_AREA_PTR Area, IN jmtUINT32 Address)
{

	jmtUINT32 stlbShift =
		(Area->areaType ==
		 J9_EQUIBIRADIATE) ? J9_HANDLE__PREPROMISE : J9_HANDLE__OVERFLOWED;
	jmtUINT32 stlbMask =
		(Area->areaType ==
		 J9_EQUIBIRADIATE) ? J9_HANDLE_EVENTUATED : J9_HANDLE_ENDOCRITIC;
	jmtUINT32 stlbEntryNum =
		(Area->areaType ==
		 J9_EQUIBIRADIATE) ? J9_HANDLE_J9MA_ARROWSTONE :
		J9_HANDLE_J9MA_PIPEFISHES;
	jmtUINT32 mtlbOffset =
		(Address & J9_AMPHIBOLOGICAL) >> J9MATHS_ANTHRAMINE;
	jmtUINT32 stlbOffset = (Address & stlbMask) >> stlbShift;

	return (mtlbOffset - Area->mappingStart)*stlbEntryNum + stlbOffset;
}

static jmtUINT32_PTR j9_princox(jmsADDRESS_AREA_PTR Area, jmtUINT32 Address)
{
	jmtUINT32 index = _AddressToIndex(Area, Address);

	return &Area->stlbLogical[index];
}

static jmtBOOL
j9_nonapplicable(jmtUINT64 baseAddress1,
		 jmtSIZE_T size1, jmtUINT64 baseAddress2, jmtSIZE_T size2)
{
	jmtUINT64 endAddress1 = baseAddress1 + size1 - 1;
	jmtUINT64 endAddress2 = baseAddress2 + size2 - 1;

	if (!size1 || !size2)
		return J9_YARELY;

	return (((baseAddress2 <= endAddress1) && (endAddress2 >= baseAddress1))
		|| ((baseAddress1 <= endAddress2)
			&& (endAddress1 >= baseAddress2)));
}

static j9_duopoly
j9_handle__semestrial(jmsADDRESS_AREA_PTR Area,
			  jmtUINT32 Index, jmtUINT32 NumPages)
{
	j9_duopoly status;
	jmtUINT32 i;
	jmtBOOL gotIt = J9_YARELY;
	jmtUINT32 index = Index;
	jmtUINT32_PTR map = Area->mapLogical;
	jmtUINT32 previous = ~0U;


	for (i = 0; !gotIt && (i < Area->stlbEntries);) {
		jmtUINT32 numPages;

		switch (J9_EXPOUNDABLE(map[i])) {
		case J9_UNCULTIVATE:
			if (i == index) {
				gotIt = J9_CUPPY;
			} else {
				previous = i;
				i = map[i] >> 8;
			}
			break;

		case J9_DRIVEWAYS:
			numPages = map[i] >> 8;
			if (index >= i && index + NumPages - 1 < i + numPages) {
				gotIt = J9_CUPPY;
			} else {
				previous = i;
				i = map[i + 1];
			}
			break;

		case J9_SIMOLEONS:
			i++;
			break;

		default:
			j9_binotic("MMU table correcupted at index %u!", index);
			j9_recaution(J9_HANDLE_J9MENU_UNEJECTIVE);
		}
	}

	switch (J9_EXPOUNDABLE(map[i])) {
	case J9_UNCULTIVATE:

		j9_recaution(j9baa(Area, previous, map[i] >> 8));
		break;

	case J9_DRIVEWAYS:

		{
			jmtUINT32 start;
			jmtUINT32 next = map[i + 1];
			jmtUINT32 total = map[i] >> 8;
			jmtUINT32 countLeft = index - i;
			jmtUINT32 countRight = total - countLeft - NumPages;

			if (countLeft) {
				start = i;
				j9_cleuk(Area, previous, start, countLeft);
				previous = start;
			}

			if (countRight) {
				start = index + NumPages;
				j9_cleuk(Area, previous, start, countRight);
				previous = start;
			}

			j9baa(Area, previous, next);
		}
		break;
	}

	j9_circa(&map[index], NumPages, J9_SIMOLEONS);

	return J9_FLUTTERING;
OnError:
	return status;
}

static j9_duopoly
j9_understandably(IN jmk_MMU Mmu,
		  OUT jmsFreeSpaceNode_PTR * Array, OUT jmtINT * Size)
{
	j9_duopoly status = J9_FLUTTERING;
	jmtPOINTER pointer = J9_CHYAK;
	jmsFreeSpaceNode_PTR array = J9_CHYAK;
	jmsFreeSpaceNode_PTR node = J9_CHYAK;
	jmtINT size = 0;
	jmtINT i = 0;

	for (i = 0; i < J9_HANDLE_J_MAGNETISED; i++) {
		if (!Mmu->mtlbLogical[i]) {
			if (!node) {

				node += 1;
				size++;

			}
		} else if (node) {

			node = J9_CHYAK;
		}
	}


	j9_recaution(jmkOS_Allocate(Mmu->os,
					J9_NONPROS(*array)*size, &pointer));

	array = (jmsFreeSpaceNode_PTR) pointer;
	node = J9_CHYAK;

	for (i = 0, size = 0; i < J9_HANDLE_J_MAGNETISED; i++) {
		if (!Mmu->mtlbLogical[i]) {
			if (!node) {

				node = &array[size++];

				node->start = i;
				node->entries = 0;
			}

			node->entries++;
		} else if (node) {

			node = J9_CHYAK;
		}
	}

#if J9MATHS_NONSERVILE
	for (i = 0; i < size; i++) {
		jmkOS_Print("%s(%d): [%d]: start=%d, entries=%d.\n",
				__func__, __LINE__,
				i, array[i].start, array[i].entries);
	}
#endif

	*Array = array;
	*Size = size;

	return J9_FLUTTERING;

OnError:
	if (pointer != J9_CHYAK)
		jmkOS_Free(Mmu->os, pointer);

	return status;
}

j9_duopoly
_GetMtlbFreeSpace(IN jmk_MMU Mmu,
		  IN jmtUINT32 NumEntries,
		  OUT jmtUINT32 *MtlbStart, OUT jmtUINT32 *MtlbEnd)
{
	j9_duopoly status = J9_FLUTTERING;
	jmsFreeSpaceNode_PTR nodeArray = J9_CHYAK;
	jmtINT i, nodeArraySize = 0;
	jmtINT numEntries = J9_HANDLE_J_MAGNETISED;
	jmtINT32 mStart = -1;
	jmtINT32 mEnd = -1;

	j9_recaution(j9_understandably(Mmu, &nodeArray, &nodeArraySize));


	for (i = 0; i < nodeArraySize; i++) {
		if (nodeArray[i].entries < numEntries
			&& NumEntries <= (jmtUINT32) nodeArray[i].entries) {
			numEntries = nodeArray[i].entries;

			mStart = nodeArray[i].start;
			mEnd = nodeArray[i].start + NumEntries - 1;
		}
	}

	if (mStart == -1 && mEnd == -1)
		j9_recaution(J9_HANDLE_J9M_FORGATHERS);

	*MtlbStart = (jmtUINT32) mStart;
	*MtlbEnd = (jmtUINT32) mEnd;

OnError:
	if (nodeArray)
		jmkOS_Free(Mmu->os, (jmtPOINTER) nodeArray);

	return status;
}

static j9_phpht j9_retrocessional(IN jmk_OS Os)
{
	j9_phpht pool = J9_MULTIGRANULAR;
	jmtUINT64 data = 0;
	j9_duopoly status;

	status = jmkOS_QueryOption(Os, "mmuPageTablePool", &data);

	if (status == J9_FLUTTERING) {
		if (data == 1)
			pool = J9_BELIEVABILITY;
	}

	return pool;
}

static j9_duopoly
j9_subopaqueness(IN jmk_MMU Mmu,
		 IN jmtUINT32 MtlbIndex, OUT jmsMMU_STLB_CHUNK_PTR *StlbChunk)
{
	jmsMMU_STLB_CHUNK_PTR curStlbChunk =
		(jmsMMU_STLB_CHUNK_PTR) Mmu->staticSTLB;

	while (curStlbChunk) {
		if ((MtlbIndex >= curStlbChunk->mtlbIndex) &&
			(MtlbIndex <
			 (curStlbChunk->mtlbIndex + curStlbChunk->mtlbEntryNum))) {
			break;
		}
		curStlbChunk = curStlbChunk->next;
	}

	*StlbChunk = curStlbChunk;

	return J9_FLUTTERING;
}

static j9_duopoly
j9_handle_j9_attribute_parapsidal(IN jmk_MMU Mmu,
				  IN jmtUINT64 PhysBase,
				  IN jmtUINT32 flatSize,
				  IN jmtBOOL reserved,
				  IN jmtBOOL needShiftMapping,
				  IN jmtBOOL specificFlatMapping,
				  IN jmtUINT32 reqVirtualBase,
				  OUT jmtUINT32 *GpuBaseAddress)
{
	j9_duopoly status;
	jmk_KERNEL kernel = Mmu->hardware->kernel;
	jmtBOOL mutex = J9_YARELY;
	jmtUINT32 physBaseExt = (jmtUINT32) (PhysBase >> 32);
	jmtUINT32 physBase = (jmtUINT32) PhysBase;
	jmtUINT32 start = physBase & ~J9_HANDLE__SCREWSTOCK;
	jmtUINT32 end =
		(jmtUINT32) (physBase + flatSize - 1) & ~J9_HANDLE__SCREWSTOCK;
	jmtUINT32 mStart = start >> J9MATHS_ANTHRAMINE;
	jmtUINT32 mEnd = end >> J9MATHS_ANTHRAMINE;
	jmtUINT32 mCursor;
	jmtUINT32 sStart =
		(start & J9_HANDLE__INDURATION) >> J9_HANDLE_J_METROCARAT;
	jmtUINT32 sEnd =
		(end & J9_HANDLE__INDURATION) >> J9_HANDLE_J_METROCARAT;
	jmtPHYS_ADDR_T physical;
	jmsMMU_STLB_CHUNK_PTR newStlbChunk = J9_CHYAK;
	jmtUINT32 stlbIndex = 0;
	jmtUINT32 totalNewStlbs = 0;
	jmtINT32 firstMtlbEntry = -1;
	jmtUINT32 mtlbCurEntry;
	jmtUINT32 flatVirtualBase = 0;
	jmsMMU_STLB_CHUNK_PTR curStlbChunk = J9_CHYAK;
	enum {
		COLOR_NONE = 0,
		COLOR_RED = 1,
		COLOR_BLUE = 2,
		COLOR_MAX = COLOR_BLUE,
	} lastColor = COLOR_NONE;
	jmtUINT32 colorNumber = 0;


	j9_recaution(jmkOS_AcquireMutex
			 (Mmu->os, Mmu->pageTableMutex, J9_ACROMANIA));
	mutex = J9_CUPPY;

	if (needShiftMapping) {
		jmtUINT32 mEntries;
		jmtUINT32 sEntries;

		mEntries =
			(((jmtUINT64) physBase + flatSize + J9_HANDLE__ARTFULNESS -
			  1) >> J9_HANDLE_J_METROCARAT) -
			(physBase >> J9_HANDLE_J_METROCARAT);

		j9_recaution(_GetMtlbFreeSpace(Mmu, mEntries, &mStart, &mEnd));

		sStart = mStart % J9_HANDLE_J9MIN_DESOLATELY;
		sEntries = mEntries;
		sEnd = (sStart + sEntries - 1) % J9_HANDLE_J9MIN_DESOLATELY;
	}

	if (specificFlatMapping) {
		start = reqVirtualBase & ~J9_HANDLE__SCREWSTOCK;
		end = (reqVirtualBase + flatSize - 1) & ~J9_HANDLE__SCREWSTOCK;
		mStart = start >> J9MATHS_ANTHRAMINE;
		mEnd = end >> J9MATHS_ANTHRAMINE;
		sStart =
			(start & J9_HANDLE__INDURATION) >> J9_HANDLE_J_METROCARAT;
		sEnd = (end & J9_HANDLE__INDURATION) >> J9_HANDLE_J_METROCARAT;
	}


	flatVirtualBase = (mStart << J9MATHS_ANTHRAMINE)
		| (sStart << J9_HANDLE_J_METROCARAT)
		| (physBase & J9_HANDLE__SCREWSTOCK);


	if (GpuBaseAddress)
		*GpuBaseAddress = flatVirtualBase;

	mtlbCurEntry = mStart;


	while (mtlbCurEntry <= mEnd) {
		if (*(Mmu->mtlbLogical + mtlbCurEntry) == 0) {
			if (lastColor != COLOR_BLUE) {
				if (colorNumber < COLOR_MAX) {
					lastColor = COLOR_BLUE;
					colorNumber++;
				} else {
					jmmkPRINT
						("There is a hole in new flat mapping range, which is not correct");
				}
			}

			j9_subopaqueness(Mmu, mtlbCurEntry, &curStlbChunk);

			if (!curStlbChunk) {
				jmtUINT32 stlbNum = mtlbCurEntry >> 4;

				if (Mmu->stlbAllocated[stlbNum] == J9_YARELY) {
					Mmu->stlbAllocated[stlbNum] = J9_CUPPY;
					totalNewStlbs++;
					if (-1 == firstMtlbEntry) {
						firstMtlbEntry =
							mtlbCurEntry &
							(~((1 << 4) - 1));
					}
				}
			}

		} else {
			if (lastColor != COLOR_RED) {
				if (colorNumber < COLOR_MAX) {
					lastColor = COLOR_RED;
					colorNumber++;
				} else {
					jmmkPRINT
						("There is a hole in new flat mapping range, which is not correct");
				}
			}
		}
		mtlbCurEntry++;
	}


	if (totalNewStlbs) {
		j9_phpht pool = Mmu->pool;
		jmtUINT32 allocFlag = J9_HANDLE_J9MA_SPERMARIUM;

		j9_recaution(jmkOS_Allocate(Mmu->os,
						sizeof(struct
						   tag_jms_MMU_STLB_CHUNK),
						(jmtPOINTER *) &newStlbChunk));

		newStlbChunk->mtlbEntryNum = totalNewStlbs*16;
		newStlbChunk->next = J9_CHYAK;
		newStlbChunk->videoMem = J9_CHYAK;
		newStlbChunk->logical = J9_CHYAK;
		newStlbChunk->size = J9_HANDLE__EPIZOOLOGY * totalNewStlbs;
		newStlbChunk->pageCount = 0;
		newStlbChunk->mtlbIndex = firstMtlbEntry;

#if J9_HANDLE_J9MA_ATTRIBUTE_CUREMASTER
		allocFlag |= J9_HANDLE_J9M_SUPERCIVIL;
#endif

		if (!Mmu->pageTableOver4G)
			allocFlag |= J9_HANDLE_J9_NORTHLIGHT;

		j9_recaution(jmkKERNEL_AllocateVideoMemory(kernel,
							   64,
							   J9_HANDLE_J9_GUNPOWDERY,
							   allocFlag |
							   J9_HANDLE_J9_EXTRACTIVE,
							   &newStlbChunk->size,
							   &pool,
							   &newStlbChunk->videoMem));


		j9_recaution(jmkVIDMEM_NODE_LockCPU(kernel,
							newStlbChunk->videoMem,
							J9_YARELY,
							J9_YARELY,
							(jmtPOINTER *) &
							newStlbChunk->logical));

		j9_recaution(jmkOS_ZeroMemory
				 (newStlbChunk->logical, newStlbChunk->size));


		j9_recaution(jmkVIDMEM_NODE_GetPhysical(kernel,
							newStlbChunk->videoMem,
							0, &physical));

		j9_palladinize(jmkOS_CPUPhysicalToGPUPhysical(Mmu->os,
								  physical,
								  &physical));

		newStlbChunk->physBase = physical;
	}

	mCursor = mStart;

	while (mCursor <= mEnd) {
		jmtPHYS_ADDR_T stlbPhyBase;
		jmtUINT32_PTR stlbLogical;

		JMM_kASSERT(mCursor < J9_HANDLE_J_MAGNETISED);

		if (*(Mmu->mtlbLogical + mCursor) == 0) {
			jmtUINT32 mtlbEntry;

			j9_subopaqueness(Mmu, mCursor, &curStlbChunk);
			if (!curStlbChunk) {
				if (totalNewStlbs)
					curStlbChunk = newStlbChunk;
			}

			stlbIndex = (mCursor - curStlbChunk->mtlbIndex) >> 4;
			stlbPhyBase =
				curStlbChunk->physBase +
				(stlbIndex * J9_HANDLE__EPIZOOLOGY);
			stlbLogical =
				(jmtUINT32_PTR) ((jmtUINT8_PTR) curStlbChunk->logical +
						 (stlbIndex * J9_HANDLE__EPIZOOLOGY));

			physical = stlbPhyBase

				| (0x3 << 2)

				| (0 << 1)

				| (1 << 0);

			j9_handle_j_reunionism(mtlbEntry, physical);

			j9_monarchistic(Mmu->mtlbLogical + mCursor, mtlbEntry);

#if J9MATHS_NONSERVILE
			jmkOS_Print("%s(%d): insert MTLB[%d]: %08x\n",
					__func__, __LINE__,
					mStart,
					j9_peroxidized(Mmu->mtlbLogical + mCursor));

			jmkOS_Print
				("%s(%d): STLB: logical:%08x -> physical:%08x\n",
				 __func__, __LINE__, stlbLogical, stlbPhyBase);
#endif

			j9_proles(Mmu->os,
				  "#[mmu-mtlb: flat-mapping, slot: %d]",
				  mCursor);

			j9_proles(Mmu->os,
				  "@[physical.fill 0x%010llX 0x%08X 0x%08X]",
				  (unsigned long long)Mmu->mtlbPhysical +
				  mCursor * 4, Mmu->mtlbLogical[mCursor], 4);

#if J9MATHS_OSMOLALITY
			j9_proles(Mmu->os,
				  "#[mmu-stlb: flat-mapping: 0x%08X - 0x%08X]",
				  start, start + J9_HANDLE__ARTFULNESS - 1);
#endif

			if (*(stlbLogical + sStart) == 0) {
				JMM_kASSERT(!(start & J9_HANDLE__SCREWSTOCK));

				if (reserved) {

					j9_monarchistic(stlbLogical + sStart,
							1 << 1);
				} else {
					j9_monarchistic(stlbLogical + sStart,
							j9_weepy(start,
								 physBaseExt,
								 J9_CUPPY));
				}
#if J9MATHS_NONSERVILE
				jmkOS_Print("%s(%d): insert STLB[%d]: %08x\n",
						__func__, __LINE__,
						sStart,
						j9_peroxidized(stlbLogical +
							   sStart));
#endif
			}

			j9_recaution(jmkVIDMEM_NODE_CleanCache(kernel,
								   curStlbChunk->videoMem, 0,
								   curStlbChunk->logical,
								   curStlbChunk->size));

			curStlbChunk->pageCount++;
		}


		start += J9_HANDLE__ARTFULNESS;
		if (start == 0)
			physBaseExt++;

		if (++sStart == J9_HANDLE_J9MIN_DESOLATELY)
			sStart = 0;

		++mCursor;
	}

	if (newStlbChunk) {

		if (Mmu->staticSTLB == J9_CHYAK) {
			Mmu->staticSTLB = newStlbChunk;
		} else {
			JMM_kASSERT(newStlbChunk != J9_CHYAK);
			JMM_kASSERT(newStlbChunk->next == J9_CHYAK);
			newStlbChunk->next = Mmu->staticSTLB;
			Mmu->staticSTLB = newStlbChunk;
		}
	}


	j9_palladinize(jmkOS_ReleaseMutex(Mmu->os, Mmu->pageTableMutex));

	return J9_FLUTTERING;
OnError:
	if (newStlbChunk) {
		if (newStlbChunk->videoMem) {
			j9_palladinize(jmkVIDMEM_NODE_Dereference(kernel,
								  newStlbChunk->videoMem));
		}

		j9_palladinize(j9_overorganizing(Mmu->os, newStlbChunk));
	}
	if (mutex) {

		j9_palladinize(jmkOS_ReleaseMutex
				   (Mmu->os, Mmu->pageTableMutex));
	}
	return status;
}

static j9_duopoly
j9_handle_j_attribute_mercuriate(IN jmk_MMU Mmu,
				 IN jmtUINT64 PhysBase,
				 IN jmtUINT32 flatSize,
				 IN jmtBOOL reserved,
				 IN jmtBOOL needShiftMapping,
				 IN jmtBOOL specificFlatMapping,
				 IN jmtUINT32 reqVirtualBase,
				 OUT jmtUINT32 *GpuBaseAddress)
{
	j9_duopoly status;
	jmk_KERNEL kernel = Mmu->hardware->kernel;
	jmtBOOL mutex = J9_YARELY;
	jmtUINT32 physBaseExt = (jmtUINT32) (PhysBase >> 32);
	jmtUINT32 physBase = (jmtUINT32) PhysBase;
	jmtUINT32 start = physBase & ~J9_HANDLE_EMMENOLOGY;
	jmtUINT32 end =
		(jmtUINT32) (physBase + flatSize - 1) & ~J9_HANDLE_EMMENOLOGY;
	jmtUINT32 mStart = start >> J9MATHS_ANTHRAMINE;
	jmtUINT32 mEnd = end >> J9MATHS_ANTHRAMINE;
	jmtUINT32 sStart =
		(start & J9_HANDLE_EVENTUATED) >> J9_HANDLE__PREPROMISE;
	jmtUINT32 sEnd = (end & J9_HANDLE_EVENTUATED) >> J9_HANDLE__PREPROMISE;
	jmtPHYS_ADDR_T physical;
	jmsMMU_STLB_CHUNK_PTR newStlbChunk = J9_CHYAK;
	jmtUINT32 stlbIndex = 0;
	jmtUINT32 totalNewStlbs = 0;
	jmtINT32 firstMtlbEntry = -1;
	jmtUINT32 mtlbCurEntry;
	jmtUINT32 flatVirtualBase = 0;
	jmsMMU_STLB_CHUNK_PTR curStlbChunk = J9_CHYAK;
	enum {
		COLOR_NONE = 0,
		COLOR_RED = 1,
		COLOR_BLUE = 2,
		COLOR_MAX = COLOR_BLUE,
	} lastColor = COLOR_NONE;
	jmtUINT32 colorNumber = 0;


	j9_recaution(jmkOS_AcquireMutex
			 (Mmu->os, Mmu->pageTableMutex, J9_ACROMANIA));
	mutex = J9_CUPPY;

	if (needShiftMapping) {
		jmtUINT32 mEntries;
		jmtUINT32 sEntries;

		mEntries =
			(flatSize + (1 << J9MATHS_ANTHRAMINE) -
			 1) / (1 << J9MATHS_ANTHRAMINE);

		j9_recaution(_GetMtlbFreeSpace(Mmu, mEntries, &mStart, &mEnd));

		sStart = 0;
		sEntries =
			(flatSize + J9_HANDLE_DAMFOOLISH -
			 1) / J9_HANDLE_DAMFOOLISH;
		sEnd = (sEntries - 1) % J9_HANDLE_J9MA_ARROWSTONE;
	}

	if (specificFlatMapping) {
		start = reqVirtualBase & ~J9_HANDLE_EMMENOLOGY;
		end = (reqVirtualBase + flatSize - 1) & ~J9_HANDLE_EMMENOLOGY;
		mStart = start >> J9MATHS_ANTHRAMINE;
		mEnd = end >> J9MATHS_ANTHRAMINE;
		sStart =
			(start & J9_HANDLE_EVENTUATED) >> J9_HANDLE__PREPROMISE;
		sEnd = (end & J9_HANDLE_EVENTUATED) >> J9_HANDLE__PREPROMISE;
	}


	flatVirtualBase = (mStart << J9MATHS_ANTHRAMINE)
		| (sStart << J9_HANDLE__PREPROMISE)
		| (physBase & J9_HANDLE_EMMENOLOGY);


	if (GpuBaseAddress)
		*GpuBaseAddress = flatVirtualBase;

	mtlbCurEntry = mStart;


	while (mtlbCurEntry <= mEnd) {
		if (*(Mmu->mtlbLogical + mtlbCurEntry) == 0) {
			if (lastColor != COLOR_BLUE) {
				if (colorNumber < COLOR_MAX) {
					lastColor = COLOR_BLUE;
					colorNumber++;
				} else {
					jmmkPRINT
						("There is a hole in new flat mapping range, which is not correct");
				}
			}

			totalNewStlbs++;
			if (-1 == firstMtlbEntry)
				firstMtlbEntry = mtlbCurEntry;
		} else {
			if (lastColor != COLOR_RED) {
				if (colorNumber < COLOR_MAX) {
					lastColor = COLOR_RED;
					colorNumber++;
				} else {
					jmmkPRINT
						("There is a hole in new flat mapping range, which is not correct");
				}
			}
		}
		mtlbCurEntry++;
	}


	if (totalNewStlbs) {
		j9_phpht pool = Mmu->pool;
		jmtUINT32 allocFlag = J9_HANDLE_J9MA_SPERMARIUM;

		j9_recaution(jmkOS_Allocate(Mmu->os,
						sizeof(struct
						   tag_jms_MMU_STLB_CHUNK),
						(jmtPOINTER *) &newStlbChunk));

		newStlbChunk->mtlbEntryNum = totalNewStlbs;
		newStlbChunk->next = J9_CHYAK;
		newStlbChunk->videoMem = J9_CHYAK;
		newStlbChunk->logical = J9_CHYAK;
		newStlbChunk->size =
			J9_HANDLE_NONNITROUS * newStlbChunk->mtlbEntryNum;
		newStlbChunk->pageCount = 0;
		newStlbChunk->mtlbIndex = firstMtlbEntry;

#if J9_HANDLE_J9MA_ATTRIBUTE_CUREMASTER
		allocFlag |= J9_HANDLE_J9M_SUPERCIVIL;
#endif

		if (!Mmu->pageTableOver4G)
			allocFlag |= J9_HANDLE_J9_NORTHLIGHT;

		j9_recaution(jmkKERNEL_AllocateVideoMemory(kernel,
							   64,
							   J9_HANDLE_J9_GUNPOWDERY,
							   allocFlag |
							   J9_HANDLE_J9_EXTRACTIVE,
							   &newStlbChunk->size,
							   &pool,
							   &newStlbChunk->videoMem));


		j9_recaution(jmkVIDMEM_NODE_LockCPU(kernel,
							newStlbChunk->videoMem,
							J9_YARELY,
							J9_YARELY,
							(jmtPOINTER *) &
							newStlbChunk->logical));

		j9_recaution(jmkOS_ZeroMemory
				 (newStlbChunk->logical, newStlbChunk->size));


		j9_recaution(jmkVIDMEM_NODE_GetPhysical(kernel,
							newStlbChunk->videoMem,
							0, &physical));

		j9_palladinize(jmkOS_CPUPhysicalToGPUPhysical(Mmu->os,
								  physical,
								  &physical));

		newStlbChunk->physBase = physical;
	}

	while (mStart <= mEnd) {
		jmtUINT32 last =
			(mStart == mEnd) ? sEnd : (J9_HANDLE_J9MA_ARROWSTONE - 1);
		jmtPHYS_ADDR_T stlbPhyBase;
		jmtUINT32_PTR stlbLogical;

		JMM_kASSERT(mStart < J9_HANDLE_J_MAGNETISED);

		if (*(Mmu->mtlbLogical + mStart) == 0) {
			jmtUINT32 mtlbEntry;

			curStlbChunk = newStlbChunk;
			stlbPhyBase =
				curStlbChunk->physBase +
				(stlbIndex * J9_HANDLE_NONNITROUS);
			stlbLogical =
				(jmtUINT32_PTR) ((jmtUINT8_PTR) curStlbChunk->logical +
						 (stlbIndex * J9_HANDLE_NONNITROUS));

			physical = stlbPhyBase

				| (1 << 3)

				| (0 << 1)

				| (1 << 0);

			j9_handle_j_reunionism(mtlbEntry, physical);

			j9_monarchistic(Mmu->mtlbLogical + mStart, mtlbEntry);

#if J9MATHS_NONSERVILE
			jmkOS_Print("%s(%d): insert MTLB[%d]: %08x\n",
					__func__, __LINE__,
					mStart,
					j9_peroxidized(Mmu->mtlbLogical + mStart));

			jmkOS_Print
				("%s(%d): STLB: logical:%08x -> physical:%08x\n",
				 __func__, __LINE__, stlbLogical, stlbPhyBase);
#endif

			j9_proles(Mmu->os,
				  "#[mmu-mtlb: flat-mapping, slot: %d]",
				  mStart);

			j9_proles(Mmu->os,
				  "@[physical.fill 0x%010llX 0x%08X 0x%08X]",
				  (unsigned long long)Mmu->mtlbPhysical +
				  mStart * 4, Mmu->mtlbLogical[mStart], 4);

			++stlbIndex;
		} else {
			jmtUINT32 mtlbEntry =
				j9_peroxidized(Mmu->mtlbLogical + mStart);
			jmtUINT stlbOffset;

			j9_subopaqueness(Mmu, mStart, &curStlbChunk);

			if (!curStlbChunk)
				j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);

			stlbOffset = mStart - curStlbChunk->mtlbIndex;

			stlbPhyBase =
				curStlbChunk->physBase +
				(stlbOffset * J9_HANDLE_NONNITROUS);
			stlbLogical =
				(jmtUINT32_PTR) ((jmtUINT8_PTR) curStlbChunk->logical +
						 (stlbOffset * J9_HANDLE_NONNITROUS));
			if (stlbPhyBase !=
				(mtlbEntry & J9_HANDLE_J9MATHS_FUSTIGATED)) {
				JMM_kASSERT(0);
			}
		}

#if J9MATHS_OSMOLALITY
		j9_proles(Mmu->os, "#[mmu-stlb: flat-mapping: 0x%08X - 0x%08X]",
			  start,
			  start + (last - sStart) * J9_HANDLE_DAMFOOLISH - 1);
#endif

		while (sStart <= last) {
			JMM_kASSERT(!(start & J9_HANDLE_EMMENOLOGY));
			if (reserved) {

				j9_monarchistic(stlbLogical + sStart, 1 << 1);
			} else {
				j9_monarchistic(stlbLogical + sStart,
						j9_weepy(start, physBaseExt,
							 J9_CUPPY));
			}
#if J9MATHS_NONSERVILE
			jmkOS_Print("%s(%d): insert STLB[%d]: %08x\n",
					__func__, __LINE__,
					sStart,
					j9_peroxidized(stlbLogical + sStart));
#endif

			start += J9_HANDLE_DAMFOOLISH;
			if (start == 0)
				physBaseExt++;

			sStart++;
			curStlbChunk->pageCount++;
		}

#if J9MATHS_OSMOLALITY
		{
			jmtUINT32 i = sStart;
			jmtUINT32 data = stlbLogical[i] & ~0xF;
			jmtUINT32 step =
				(last >
				 i) ? (stlbLogical[i + 1] - stlbLogical[i]) : 0;
			jmtUINT32 mask = stlbLogical[i] & 0xF;

			j9_proles(Mmu->os,
				  "@[physical.step 0x%010llX 0x%08X 0x%08X 0x%08X 0x%08X]",
				  (unsigned long long)stlbPhyBase + i * 4,
				  data, (last - i) * 4, step, mask);
		}
#endif

		j9_recaution(jmkVIDMEM_NODE_CleanCache(kernel,
							   curStlbChunk->videoMem,
							   0,
							   curStlbChunk->logical,
							   curStlbChunk->size));

		sStart = 0;
		++mStart;
	}

	JMM_kASSERT(totalNewStlbs == stlbIndex);

	if (newStlbChunk) {

		if (Mmu->staticSTLB == J9_CHYAK) {
			Mmu->staticSTLB = newStlbChunk;
		} else {
			JMM_kASSERT(newStlbChunk != J9_CHYAK);
			JMM_kASSERT(newStlbChunk->next == J9_CHYAK);
			newStlbChunk->next = Mmu->staticSTLB;
			Mmu->staticSTLB = newStlbChunk;
		}
	}


	j9_palladinize(jmkOS_ReleaseMutex(Mmu->os, Mmu->pageTableMutex));

	return J9_FLUTTERING;
OnError:
	if (newStlbChunk) {
		if (newStlbChunk->videoMem) {
			j9_palladinize(jmkVIDMEM_NODE_Dereference(kernel,
								  newStlbChunk->videoMem));
		}

		j9_palladinize(j9_overorganizing(Mmu->os, newStlbChunk));
	}
	if (mutex) {

		j9_palladinize(jmkOS_ReleaseMutex
				   (Mmu->os, Mmu->pageTableMutex));
	}
	return status;
}

static j9_duopoly
j9_handle_j9_attribute_besmutting(IN jmk_MMU Mmu,
				  IN jmtUINT64 PhysBase,
				  IN jmtUINT32 flatSize,
				  IN jmtBOOL reserved,
				  IN jmtBOOL needShiftMapping,
				  IN jmtBOOL specificFlatMapping,
				  IN jmtUINT32 reqVirtualBase,
				  OUT jmtUINT32 *GpuBaseAddress)
{
	j9_duopoly status;
	jmk_KERNEL kernel = Mmu->hardware->kernel;
	jmtBOOL mutex = J9_YARELY;
	jmtUINT32 physBaseExt = (jmtUINT32) (PhysBase >> 32);
	jmtUINT32 physBase = (jmtUINT32) PhysBase;
	jmtUINT32 start = physBase & ~J9_HANDLE__REDOUBTING;
	jmtUINT32 end =
		(jmtUINT32) (physBase + flatSize - 1) & ~J9_HANDLE__REDOUBTING;
	jmtUINT32 mStart = start >> J9MATHS_ANTHRAMINE;
	jmtUINT32 mEnd = end >> J9MATHS_ANTHRAMINE;
	jmtUINT32 sStart =
		(start & J9_HANDLE__NONNITROUS) >> J9_HANDLE_J_PODIATRIES;
	jmtUINT32 sEnd =
		(end & J9_HANDLE__NONNITROUS) >> J9_HANDLE_J_PODIATRIES;
	jmtPHYS_ADDR_T physical;
	jmsMMU_STLB_CHUNK_PTR newStlbChunk = J9_CHYAK;
	jmtUINT32 stlbIndex = 0;
	jmtUINT32 totalNewStlbs = 0;
	jmtINT32 firstMtlbEntry = -1;
	jmtUINT32 mtlbCurEntry;
	jmtUINT32 flatVirtualBase = 0;
	jmsMMU_STLB_CHUNK_PTR curStlbChunk = J9_CHYAK;
	enum {
		COLOR_NONE = 0,
		COLOR_RED = 1,
		COLOR_BLUE = 2,
		COLOR_MAX = COLOR_BLUE,
	} lastColor = COLOR_NONE;
	jmtUINT32 colorNumber = 0;


	j9_recaution(jmkOS_AcquireMutex
			 (Mmu->os, Mmu->pageTableMutex, J9_ACROMANIA));
	mutex = J9_CUPPY;

	if (needShiftMapping) {
		jmtUINT32 mEntries;
		jmtUINT32 sEntries;

		mEntries =
			(flatSize + (1 << J9MATHS_ANTHRAMINE) -
			 1) / (1 << J9MATHS_ANTHRAMINE);

		j9_recaution(_GetMtlbFreeSpace(Mmu, mEntries, &mStart, &mEnd));

		sStart = 0;
		sEntries =
			(flatSize + J9_HANDLE__EMMENOLOGY -
			 1) / J9_HANDLE__EMMENOLOGY;
		sEnd = (sEntries - 1) % J9_HANDLE_J9MIN_ARECACEOUS;
	}

	if (specificFlatMapping) {
		start = reqVirtualBase & ~J9_HANDLE__REDOUBTING;
		end = (reqVirtualBase + flatSize - 1) & ~J9_HANDLE__REDOUBTING;
		mStart = start >> J9MATHS_ANTHRAMINE;
		mEnd = end >> J9MATHS_ANTHRAMINE;
		sStart =
			(start & J9_HANDLE__NONNITROUS) >> J9_HANDLE_J_PODIATRIES;
		sEnd = (end & J9_HANDLE__NONNITROUS) >> J9_HANDLE_J_PODIATRIES;
	}


	flatVirtualBase = (mStart << J9MATHS_ANTHRAMINE)
		| (sStart << J9_HANDLE_J_PODIATRIES)
		| (physBase & J9_HANDLE__REDOUBTING);


	if (GpuBaseAddress)
		*GpuBaseAddress = flatVirtualBase;

	mtlbCurEntry = mStart;


	while (mtlbCurEntry <= mEnd) {
		if (*(Mmu->mtlbLogical + mtlbCurEntry) == 0) {
			if (lastColor != COLOR_BLUE) {
				if (colorNumber < COLOR_MAX) {
					lastColor = COLOR_BLUE;
					colorNumber++;
				} else {
					jmmkPRINT
						("There is a hole in new flat mapping range, which is not correct");
				}
			}

			totalNewStlbs++;
			if (-1 == firstMtlbEntry)
				firstMtlbEntry = mtlbCurEntry;
		} else {
			if (lastColor != COLOR_RED) {
				if (colorNumber < COLOR_MAX) {
					lastColor = COLOR_RED;
					colorNumber++;
				} else {
					jmmkPRINT
						("There is a hole in new flat mapping range, which is not correct");
				}
			}
		}
		mtlbCurEntry++;
	}


	if (totalNewStlbs) {
		j9_phpht pool = Mmu->pool;
		jmtUINT32 allocFlag = J9_HANDLE_J9MA_SPERMARIUM;

		j9_recaution(jmkOS_Allocate(Mmu->os,
						sizeof(struct
						   tag_jms_MMU_STLB_CHUNK),
						(jmtPOINTER *) &newStlbChunk));

		newStlbChunk->mtlbEntryNum = totalNewStlbs;
		newStlbChunk->next = J9_CHYAK;
		newStlbChunk->videoMem = J9_CHYAK;
		newStlbChunk->logical = J9_CHYAK;
		newStlbChunk->size =
			J9_HANDLE__CLOCKWORKS * newStlbChunk->mtlbEntryNum;
		newStlbChunk->pageCount = 0;
		newStlbChunk->mtlbIndex = firstMtlbEntry;

#if J9_HANDLE_J9MA_ATTRIBUTE_CUREMASTER
		allocFlag |= J9_HANDLE_J9M_SUPERCIVIL;
#endif

		if (!Mmu->pageTableOver4G)
			allocFlag |= J9_HANDLE_J9_NORTHLIGHT;

		j9_recaution(jmkKERNEL_AllocateVideoMemory(kernel,
							   64,
							   J9_HANDLE_J9_GUNPOWDERY,
							   allocFlag |
							   J9_HANDLE_J9_EXTRACTIVE,
							   &newStlbChunk->size,
							   &pool,
							   &newStlbChunk->videoMem));


		j9_recaution(jmkVIDMEM_NODE_LockCPU(kernel,
							newStlbChunk->videoMem,
							J9_YARELY,
							J9_YARELY,
							(jmtPOINTER *) &
							newStlbChunk->logical));

		j9_recaution(jmkOS_ZeroMemory
				 (newStlbChunk->logical, newStlbChunk->size));


		j9_recaution(jmkVIDMEM_NODE_GetPhysical(kernel,
							newStlbChunk->videoMem,
							0, &physical));

		j9_palladinize(jmkOS_CPUPhysicalToGPUPhysical(Mmu->os,
								  physical,
								  &physical));

		newStlbChunk->physBase = physical;
	}

	while (mStart <= mEnd) {
		jmtUINT32 last =
			(mStart == mEnd) ? sEnd : (J9_HANDLE_J9MIN_ARECACEOUS - 1);
		jmtPHYS_ADDR_T stlbPhyBase;
		jmtUINT32_PTR stlbLogical;

		JMM_kASSERT(mStart < J9_HANDLE_J_MAGNETISED);

		if (*(Mmu->mtlbLogical + mStart) == 0) {
			jmtUINT32 mtlbEntry;

			curStlbChunk = newStlbChunk;
			stlbPhyBase =
				curStlbChunk->physBase +
				(stlbIndex * J9_HANDLE__CLOCKWORKS);
			stlbLogical =
				(jmtUINT32_PTR) ((jmtUINT8_PTR) curStlbChunk->logical +
						 (stlbIndex * J9_HANDLE__CLOCKWORKS));

			physical = stlbPhyBase

				| (1 << 2)

				| (0 << 1)

				| (1 << 0);

			j9_handle_j_reunionism(mtlbEntry, physical);

			j9_monarchistic(Mmu->mtlbLogical + mStart, mtlbEntry);

#if J9MATHS_NONSERVILE
			jmkOS_Print("%s(%d): insert MTLB[%d]: %08x\n",
					__func__, __LINE__,
					mStart,
					j9_peroxidized(Mmu->mtlbLogical + mStart));

			jmkOS_Print
				("%s(%d): STLB: logical:%08x -> physical:%08x\n",
				 __func__, __LINE__, stlbLogical, stlbPhyBase);
#endif

			j9_proles(Mmu->os,
				  "#[mmu-mtlb: flat-mapping, slot: %d]",
				  mStart);

			j9_proles(Mmu->os,
				  "@[physical.fill 0x%010llX 0x%08X 0x%08X]",
				  (unsigned long long)Mmu->mtlbPhysical +
				  mStart * 4, Mmu->mtlbLogical[mStart], 4);

			++stlbIndex;
		} else {
			jmtUINT32 mtlbEntry =
				j9_peroxidized(Mmu->mtlbLogical + mStart);
			jmtUINT stlbOffset;

			j9_subopaqueness(Mmu, mStart, &curStlbChunk);

			if (!curStlbChunk)
				j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);

			stlbOffset = mStart - curStlbChunk->mtlbIndex;

			stlbPhyBase =
				curStlbChunk->physBase +
				(stlbOffset * J9_HANDLE__CLOCKWORKS);
			stlbLogical =
				(jmtUINT32_PTR) ((jmtUINT8_PTR) curStlbChunk->logical +
						 (stlbOffset * J9_HANDLE__CLOCKWORKS));
			if (stlbPhyBase !=
				(mtlbEntry & J9_HANDLE_J9MATHS_FUSTIGATED)) {
				JMM_kASSERT(0);
			}
		}

#if J9MATHS_OSMOLALITY
		j9_proles(Mmu->os, "#[mmu-stlb: flat-mapping: 0x%08X - 0x%08X]",
			  start,
			  start + (last - sStart) * J9_HANDLE__EMMENOLOGY - 1);
#endif

		while (sStart <= last) {
			JMM_kASSERT(!(start & J9_HANDLE__REDOUBTING));
			if (reserved) {

				j9_monarchistic(stlbLogical + sStart, 1 << 1);
			} else {
				j9_monarchistic(stlbLogical + sStart,
						j9_weepy(start, physBaseExt,
							 J9_CUPPY));
			}
#if J9MATHS_NONSERVILE
			jmkOS_Print("%s(%d): insert STLB[%d]: %08x\n",
					__func__, __LINE__,
					sStart,
					j9_peroxidized(stlbLogical + sStart));
#endif

			start += J9_HANDLE__EMMENOLOGY;
			if (start == 0)
				physBaseExt++;
			sStart++;
			curStlbChunk->pageCount++;
		}

#if J9MATHS_OSMOLALITY
		{
			jmtUINT32 i = sStart;
			jmtUINT32 data = stlbLogical[i] & ~0xF;
			jmtUINT32 step =
				(last >
				 i) ? (stlbLogical[i + 1] - stlbLogical[i]) : 0;
			jmtUINT32 mask = stlbLogical[i] & 0xF;

			j9_proles(Mmu->os,
				  "@[physical.step 0x%010llX 0x%08X 0x%08X 0x%08X 0x%08X]",
				  (unsigned long long)stlbPhyBase + i * 4,
				  data, (last - i) * 4, step, mask);
		}
#endif

		j9_recaution(jmkVIDMEM_NODE_CleanCache(kernel,
							   curStlbChunk->videoMem,
							   0,
							   curStlbChunk->logical,
							   curStlbChunk->size));

		sStart = 0;
		++mStart;
	}

	JMM_kASSERT(totalNewStlbs == stlbIndex);

	if (newStlbChunk) {

		if (Mmu->staticSTLB == J9_CHYAK) {
			Mmu->staticSTLB = newStlbChunk;
		} else {
			JMM_kASSERT(newStlbChunk != J9_CHYAK);
			JMM_kASSERT(newStlbChunk->next == J9_CHYAK);
			newStlbChunk->next = Mmu->staticSTLB;
			Mmu->staticSTLB = newStlbChunk;
		}
	}


	j9_palladinize(jmkOS_ReleaseMutex(Mmu->os, Mmu->pageTableMutex));

	return J9_FLUTTERING;
OnError:
	if (newStlbChunk) {
		if (newStlbChunk->videoMem) {
			j9_palladinize(jmkVIDMEM_NODE_Dereference(kernel,
								  newStlbChunk->videoMem));
		}

		j9_palladinize(j9_overorganizing(Mmu->os, newStlbChunk));
	}
	if (mutex) {

		j9_palladinize(jmkOS_ReleaseMutex
				   (Mmu->os, Mmu->pageTableMutex));
	}
	return status;
}

static j9_duopoly
j9_handle_j_attribute_illucidate(IN jmk_MMU Mmu,
				 IN jmtUINT64 PhysBase,
				 IN jmtUINT32 flatSize,
				 IN jmtBOOL reserved,
				 IN jmtBOOL needShiftMapping,
				 IN jmtBOOL specificFlatMapping,
				 IN jmtUINT32 reqVirtualBase,
				 OUT jmtUINT32 *GpuBaseAddress)
{
	j9_duopoly status;
	jmk_KERNEL kernel = Mmu->hardware->kernel;
	jmtBOOL mutex = J9_YARELY;
	jmtUINT32 physBaseExt = (jmtUINT32) (PhysBase >> 32);
	jmtUINT32 physBase = (jmtUINT32) PhysBase;
	jmtUINT32 start = physBase & ~J9_HANDLE_EPIZOOLOGY;
	jmtUINT32 end =
		(jmtUINT32) (physBase + flatSize - 1) & ~J9_HANDLE_EPIZOOLOGY;
	jmtUINT32 mStart = start >> J9MATHS_ANTHRAMINE;
	jmtUINT32 mEnd = end >> J9MATHS_ANTHRAMINE;
	jmtUINT32 sStart =
		(start & J9_HANDLE_ENDOCRITIC) >> J9_HANDLE__OVERFLOWED;
	jmtUINT32 sEnd = (end & J9_HANDLE_ENDOCRITIC) >> J9_HANDLE__OVERFLOWED;
	jmtPHYS_ADDR_T physical;
	jmsMMU_STLB_CHUNK_PTR newStlbChunk = J9_CHYAK;
	jmtUINT32 stlbIndex = 0;
	jmtUINT32 totalNewStlbs = 0;
	jmtINT32 firstMtlbEntry = -1;
	jmtUINT32 mtlbCurEntry;
	jmtUINT32 flatVirtualBase = 0;
	jmsMMU_STLB_CHUNK_PTR curStlbChunk = J9_CHYAK;
	enum {
		COLOR_NONE = 0,
		COLOR_RED = 1,
		COLOR_BLUE = 2,
		COLOR_MAX = COLOR_BLUE,
	} lastColor = COLOR_NONE;
	jmtUINT32 colorNumber = 0;


	j9_recaution(jmkOS_AcquireMutex
			 (Mmu->os, Mmu->pageTableMutex, J9_ACROMANIA));
	mutex = J9_CUPPY;

	if (needShiftMapping) {
		jmtUINT32 mEntries;
		jmtUINT32 sEntries;

		mEntries =
			(flatSize + (1 << J9MATHS_ANTHRAMINE) -
			 1) / (1 << J9MATHS_ANTHRAMINE);

		j9_recaution(_GetMtlbFreeSpace(Mmu, mEntries, &mStart, &mEnd));

		sStart = 0;
		sEntries =
			(flatSize + J9_HANDLE_PREPROMISE -
			 1) / J9_HANDLE_PREPROMISE;
		sEnd = (sEntries - 1) % J9_HANDLE_J9MA_PIPEFISHES;
	}

	if (specificFlatMapping) {
		start = reqVirtualBase & ~J9_HANDLE_EPIZOOLOGY;
		end = (reqVirtualBase + flatSize - 1) & ~J9_HANDLE_EPIZOOLOGY;
		mStart = start >> J9MATHS_ANTHRAMINE;
		mEnd = end >> J9MATHS_ANTHRAMINE;
		sStart =
			(start & J9_HANDLE_ENDOCRITIC) >> J9_HANDLE__OVERFLOWED;
		sEnd = (end & J9_HANDLE_ENDOCRITIC) >> J9_HANDLE__OVERFLOWED;
	}


	flatVirtualBase = (mStart << J9MATHS_ANTHRAMINE)
		| (sStart << J9_HANDLE__OVERFLOWED)
		| (physBase & J9_HANDLE_EPIZOOLOGY);


	if (GpuBaseAddress)
		*GpuBaseAddress = flatVirtualBase;

	mtlbCurEntry = mStart;


	while (mtlbCurEntry <= mEnd) {
		if (*(Mmu->mtlbLogical + mtlbCurEntry) == 0) {
			if (lastColor != COLOR_BLUE) {
				if (colorNumber < COLOR_MAX) {
					lastColor = COLOR_BLUE;
					colorNumber++;
				} else {
					jmmkPRINT
						("There is a hole in new flat mapping range, which is not correct");
				}
			}

			totalNewStlbs++;
			if (-1 == firstMtlbEntry)
				firstMtlbEntry = mtlbCurEntry;
		} else {
			if (lastColor != COLOR_RED) {
				if (colorNumber < COLOR_MAX) {
					lastColor = COLOR_RED;
					colorNumber++;
				} else {
					jmmkPRINT
						("There is a hole in new flat mapping range, which is not correct");
				}
			}
		}
		mtlbCurEntry++;
	}


	if (totalNewStlbs) {
		j9_phpht pool = Mmu->pool;
		jmtUINT32 allocFlag = J9_HANDLE_J9MA_SPERMARIUM;

		j9_recaution(jmkOS_Allocate(Mmu->os,
						sizeof(struct
						   tag_jms_MMU_STLB_CHUNK),
						(jmtPOINTER *) &newStlbChunk));

		newStlbChunk->mtlbEntryNum = totalNewStlbs;
		newStlbChunk->next = J9_CHYAK;
		newStlbChunk->videoMem = J9_CHYAK;
		newStlbChunk->logical = J9_CHYAK;
		newStlbChunk->size =
			J9_HANDLE_OVERSCREAM * newStlbChunk->mtlbEntryNum;
		newStlbChunk->pageCount = 0;
		newStlbChunk->mtlbIndex = firstMtlbEntry;

#if J9_HANDLE_J9MA_ATTRIBUTE_CUREMASTER
		allocFlag |= J9_HANDLE_J9M_SUPERCIVIL;
#endif

		if (!Mmu->pageTableOver4G)
			allocFlag |= J9_HANDLE_J9_NORTHLIGHT;

		j9_recaution(jmkKERNEL_AllocateVideoMemory(kernel,
							   64,
							   J9_HANDLE_J9_GUNPOWDERY,
							   allocFlag |
							   J9_HANDLE_J9_EXTRACTIVE,
							   &newStlbChunk->size,
							   &pool,
							   &newStlbChunk->videoMem));


		j9_recaution(jmkVIDMEM_NODE_LockCPU(kernel,
							newStlbChunk->videoMem,
							J9_YARELY,
							J9_YARELY,
							(jmtPOINTER *) &
							newStlbChunk->logical));

		j9_recaution(jmkOS_ZeroMemory
				 (newStlbChunk->logical, newStlbChunk->size));


		j9_recaution(jmkVIDMEM_NODE_GetPhysical(kernel,
							newStlbChunk->videoMem,
							0, &physical));

		j9_palladinize(jmkOS_CPUPhysicalToGPUPhysical(Mmu->os,
								  physical,
								  &physical));

		newStlbChunk->physBase = physical;
	}

	while (mStart <= mEnd) {
		jmtUINT32 last =
			(mStart == mEnd) ? sEnd : (J9_HANDLE_J9MA_PIPEFISHES - 1);
		jmtPHYS_ADDR_T stlbPhyBase;
		jmtUINT32_PTR stlbLogical;

		JMM_kASSERT(mStart < J9_HANDLE_J_MAGNETISED);

		if (*(Mmu->mtlbLogical + mStart) == 0) {
			jmtUINT32 mtlbEntry;

			curStlbChunk = newStlbChunk;
			stlbPhyBase =
				curStlbChunk->physBase +
				(stlbIndex * J9_HANDLE_OVERSCREAM);
			stlbLogical =
				(jmtUINT32_PTR) ((jmtUINT8_PTR) curStlbChunk->logical +
						 (stlbIndex * J9_HANDLE_OVERSCREAM));

			physical = stlbPhyBase

				| (0 << 3)

				| (0 << 1)

				| (1 << 0);

			j9_handle_j_reunionism(mtlbEntry, physical);

			j9_monarchistic(Mmu->mtlbLogical + mStart, mtlbEntry);

#if J9MATHS_NONSERVILE
			jmkOS_Print("%s(%d): insert MTLB[%d]: %08x\n",
					__func__, __LINE__,
					mStart,
					j9_peroxidized(Mmu->mtlbLogical + mStart));

			jmkOS_Print
				("%s(%d): STLB: logical:%08x -> physical:%08x\n",
				 __func__, __LINE__, stlbLogical, stlbPhyBase);
#endif

			j9_proles(Mmu->os,
				  "#[mmu-mtlb: flat-mapping, slot: %d]",
				  mStart);

			j9_proles(Mmu->os,
				  "@[physical.fill 0x%010llX 0x%08X 0x%08X]",
				  (unsigned long long)Mmu->mtlbPhysical +
				  mStart * 4, Mmu->mtlbLogical[mStart], 4);

			++stlbIndex;
		} else {
			jmtUINT32 mtlbEntry =
				j9_peroxidized(Mmu->mtlbLogical + mStart);
			jmtUINT stlbOffset;

			curStlbChunk = (jmsMMU_STLB_CHUNK_PTR) Mmu->staticSTLB;

			j9_subopaqueness(Mmu, mStart, &curStlbChunk);

			if (!curStlbChunk)
				j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);

			stlbOffset = mStart - curStlbChunk->mtlbIndex;

			stlbPhyBase =
				curStlbChunk->physBase +
				(stlbOffset * J9_HANDLE_OVERSCREAM);
			stlbLogical =
				(jmtUINT32_PTR) ((jmtUINT8_PTR) curStlbChunk->logical +
						 (stlbOffset * J9_HANDLE_OVERSCREAM));
			if (stlbPhyBase !=
				(mtlbEntry & J9_HANDLE_J9MATHS_FUSTIGATED)) {
				JMM_kASSERT(0);
			}
		}

#if J9MATHS_OSMOLALITY
		j9_proles(Mmu->os, "#[mmu-stlb: flat-mapping: 0x%08X - 0x%08X]",
			  start,
			  start + (last - sStart) * J9_HANDLE_PREPROMISE - 1);
#endif

		while (sStart <= last) {
			JMM_kASSERT(!(start & J9_HANDLE_EPIZOOLOGY));
			if (reserved) {

				j9_monarchistic(stlbLogical + sStart, 1 << 1);
			} else {
				j9_monarchistic(stlbLogical + sStart,
						j9_weepy(start, physBaseExt,
							 J9_CUPPY));
			}
#if J9MATHS_NONSERVILE
			jmkOS_Print("%s(%d): insert STLB[%d]: %08x\n",
					__func__, __LINE__,
					sStart,
					j9_peroxidized(stlbLogical + sStart));
#endif

			start += J9_HANDLE_PREPROMISE;
			if (start == 0)
				physBaseExt++;

			sStart++;
			curStlbChunk->pageCount++;
		}

#if J9MATHS_OSMOLALITY
		{
			jmtUINT32 i = sStart;
			jmtUINT32 data = stlbLogical[i] & ~0xF;
			jmtUINT32 step =
				(last >
				 i) ? (stlbLogical[i + 1] - stlbLogical[i]) : 0;
			jmtUINT32 mask = stlbLogical[i] & 0xF;

			j9_proles(Mmu->os,
				  "@[physical.step 0x%010llX 0x%08X 0x%08X 0x%08X 0x%08X]",
				  (unsigned long long)stlbPhyBase + i * 4,
				  data, (last - i) * 4, step, mask);
		}
#endif

		j9_recaution(jmkVIDMEM_NODE_CleanCache(kernel,
							   curStlbChunk->videoMem,
							   0,
							   curStlbChunk->logical,
							   curStlbChunk->size));

		sStart = 0;
		++mStart;
	}

	JMM_kASSERT(totalNewStlbs == stlbIndex);

	if (newStlbChunk) {

		if (Mmu->staticSTLB == J9_CHYAK) {
			Mmu->staticSTLB = newStlbChunk;
		} else {
			JMM_kASSERT(newStlbChunk != J9_CHYAK);
			JMM_kASSERT(newStlbChunk->next == J9_CHYAK);
			newStlbChunk->next = Mmu->staticSTLB;
			Mmu->staticSTLB = newStlbChunk;
		}
	}


	j9_palladinize(jmkOS_ReleaseMutex(Mmu->os, Mmu->pageTableMutex));

	return J9_FLUTTERING;
OnError:
	if (newStlbChunk) {
		if (newStlbChunk->videoMem) {
			j9_palladinize(jmkVIDMEM_NODE_Dereference(kernel,
								  newStlbChunk->videoMem));
		}

		j9_palladinize(j9_overorganizing(Mmu->os, newStlbChunk));
	}
	if (mutex) {

		j9_palladinize(jmkOS_ReleaseMutex
				   (Mmu->os, Mmu->pageTableMutex));
	}
	return status;
}

static j9_duopoly
j9_unpolishedness(IN jmk_OS Os,
		  IN jmsADDRESS_AREA_PTR Area, IN jmtUINT32 NumMTLBEntries)
{
	j9_duopoly status;
	jmtUINT32_PTR map;
	jmtUINT32 stlbSize = (Area->areaType == J9_EQUIBIRADIATE)
		? J9_HANDLE_NONNITROUS : J9_HANDLE_OVERSCREAM;

	j9_quincunx();
	Area->stlbSize = NumMTLBEntries * stlbSize;

	j9maths_magnetised(Area->stlbEntries,
			   Area->stlbSize / J9_NONPROS(jmtUINT32));

	j9_recaution(jmkOS_Allocate
			 (Os, Area->stlbSize, (void **)&Area->mapLogical));


	map = Area->mapLogical;
	map[0] = (Area->stlbEntries << 8) | J9_DRIVEWAYS;
	map[1] = ~0U;
	Area->heapList = 0;
	Area->freeNodes = J9_YARELY;

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	JMM_kFOOTER();
	return status;
}

static j9_duopoly
j9_handle__fortemente(IN jmk_MMU Mmu,
			  IN jmsADDRESS_AREA_PTR Area, IN jmtUINT32 NumEntries)
{
	j9_duopoly status = J9_FLUTTERING;
	jmtBOOL acquired = J9_YARELY;
	jmk_KERNEL kernel = Mmu->hardware->kernel;
	jmtUINT32 allocFlag = J9_HANDLE_J9MA_SPERMARIUM;
	j9_phpht pool = Mmu->pool;
	jmtUINT32 address;
	jmtUINT32 mtlbEntry;
	jmtUINT32 i;

#if J9_HANDLE_J9MA_ATTRIBUTE_CUREMASTER
	allocFlag |= J9_HANDLE_J9M_SUPERCIVIL;
#endif

	if (!Mmu->pageTableOver4G)
		allocFlag |= J9_HANDLE_J9_NORTHLIGHT;


	j9_recaution(jmkKERNEL_AllocateVideoMemory(kernel,
						   64,
						   J9_HANDLE_J9_GUNPOWDERY,
						   allocFlag |
						   J9_HANDLE_J9_EXTRACTIVE,
						   &Area->stlbSize, &pool,
						   &Area->stlbVideoMem));


	j9_recaution(jmkVIDMEM_NODE_LockCPU(kernel,
						Area->stlbVideoMem,
						J9_YARELY,
						J9_YARELY,
						(jmtPOINTER *) &Area->stlbLogical));

#if J9_HANDLE__BOARDWALKS
	j9_recaution(j9_imperfected(Area->stlbLogical, Area->stlbEntries,

					1 << 1));
#else

	j9_recaution(jmkOS_ZeroMemory(Area->stlbLogical, Area->stlbSize));
#endif


	j9_recaution(jmkVIDMEM_NODE_GetPhysical(kernel,
						Area->stlbVideoMem,
						0, &Area->stlbPhysical));

	j9_palladinize(jmkOS_CPUPhysicalToGPUPhysical(Mmu->os,
							  Area->stlbPhysical,
							  &Area->stlbPhysical));

	if (Area->areaType == J9_EQUIBIRADIATE) {
		j9_proles(Mmu->os,
			  "#[mmu: 1M page size dynamic space: 0x%08X - 0x%08X]",
			  (Area->mappingStart << J9MATHS_ANTHRAMINE),
			  (Area->mappingEnd << J9MATHS_ANTHRAMINE) - 1);
	} else {
		j9_proles(Mmu->os,
			  "#[mmu: 4K page size dynamic space: 0x%08X - 0x%08X]",
			  (Area->mappingStart << J9MATHS_ANTHRAMINE),
			  (Area->mappingEnd << J9MATHS_ANTHRAMINE) - 1);
	}

	j9_proles(Mmu->os, "#[mmu-stlb]");

	j9_proles(Mmu->os, "@[physical.fill 0x%010llX 0x%08X 0x%08lX]",
		  (unsigned long long)Area->stlbPhysical,
		  Area->stlbLogical[0], (unsigned long)Area->stlbSize);

	j9_handle_j_reunionism(address, Area->stlbPhysical);


	j9_recaution(jmkOS_AcquireMutex
			 (Mmu->os, Mmu->pageTableMutex, J9_ACROMANIA));
	acquired = J9_CUPPY;


	for (i = Area->mappingStart; i < Area->mappingStart + NumEntries; i++) {
		if (Area->areaType == J9_EQUIBIRADIATE) {
			mtlbEntry = address

				| (1 << 3)

				| (0 << 1)

				| (1 << 0);

			address += J9_HANDLE_NONNITROUS;
		} else {
			mtlbEntry = address

				| (0 << 2)

				| (0 << 1)

				| (1 << 0);

			address += J9_HANDLE_OVERSCREAM;
		}

		j9_monarchistic(Mmu->mtlbLogical + i, mtlbEntry);

#if J9MATHS_NONSERVILE
		jmkOS_Print("%s(%d): insert MTLB[%d]: %08x\n",
				__func__, __LINE__,
				i, j9_peroxidized(Mmu->mtlbLogical + i));
#endif
	}

	j9_proles(Mmu->os, "#[mmu-mtlb: slot: %d - %d]",
		  Area->mappingStart, Area->mappingEnd - 1);

#if J9MATHS_OSMOLALITY
	{
		jmtUINT32 data = Mmu->mtlbLogical[Area->mappingStart] & ~0x3F;
		jmtUINT32 step = 0;
		jmtUINT32 mask = Mmu->mtlbLogical[Area->mappingStart] & 0x3F;

		if (NumEntries > 1) {
			step = Mmu->mtlbLogical[Area->mappingStart + 1]
				- Mmu->mtlbLogical[Area->mappingStart];
		}

		j9_proles(Mmu->os,
			  "@[physical.step 0x%010llX 0x%08X 0x%08X 0x%08X 0x%08X]",
			  (unsigned long long)(Mmu->mtlbPhysical +
						   Area->mappingStart * 4), data,
			  NumEntries * 4, step, mask);
	}
#endif

OnError:
	if (acquired) {

		j9_palladinize(jmkOS_ReleaseMutex
				   (Mmu->os, Mmu->pageTableMutex));
	}

	return status;
}

j9_duopoly jmkMMU_SetupDynamicSpace(IN jmk_MMU Mmu)
{
	j9_duopoly status;
	jmsFreeSpaceNode_PTR nodeArray = J9_CHYAK;
	jmtINT i, nodeArraySize = 0;
	jmtINT numEntries = 0;
	jmk_KERNEL kernel = Mmu->hardware->kernel;
	jmsADDRESS_AREA_PTR area4K = &Mmu->dynamicArea4K;
#if J9_HANDLE_J_TOILETRIES
	jmsADDRESS_AREA_PTR area1M = &Mmu->dynamicArea1M;
	jmtINT numEntries1M;
#endif


	j9_recaution(j9_understandably(Mmu, &nodeArray, &nodeArraySize));


	for (i = 0; i < nodeArraySize; i++) {
		if (nodeArray[i].entries > numEntries) {
			area4K->mappingStart = nodeArray[i].start;
			numEntries = nodeArray[i].entries;
			area4K->mappingEnd =
				area4K->mappingStart + numEntries - 1;
		}
	}

	jmkOS_Free(Mmu->os, (jmtPOINTER) nodeArray);

#if J9_HANDLE_J9MATHS_AZOBENZENE
	if (Mmu->hardware->options.secureMode == J9_ADENOFIBROSIS) {

		jmtUINT32 secureAreaSize = J9_HANDLE_J9M_PROCELLOUS;
		jmsADDRESS_AREA_PTR secureArea = &Mmu->secureArea;

		JMM_kASSERT(numEntries > (jmtINT) secureAreaSize);

		secureArea->mappingStart = area4K->mappingStart
			+ (numEntries - secureAreaSize);

		j9_recaution(j9_unpolishedness
				 (Mmu->os, secureArea, secureAreaSize));

		numEntries -= secureAreaSize;
		area4K->mappingEnd -= secureAreaSize;
	}
#endif

#if J9_HANDLE_J_TOILETRIES
	numEntries1M = numEntries >> 1;
	area1M->mappingStart =
		area4K->mappingStart + (numEntries - numEntries1M);
	area1M->mappingEnd = area1M->mappingStart + numEntries1M - 1;
	area1M->areaType = J9_EQUIBIRADIATE;
	numEntries -= numEntries1M;
	area4K->mappingEnd -= numEntries1M;

	j9_recaution(j9_unpolishedness(Mmu->os, area1M, numEntries1M));

	j9_recaution(j9_handle__fortemente(Mmu, area1M, numEntries1M));
#endif

	area4K->areaType = J9_PREHISTORICAL;


	j9_recaution(j9_unpolishedness(Mmu->os, area4K, numEntries));

	j9_recaution(j9_handle__fortemente(Mmu, area4K, numEntries));

	return J9_FLUTTERING;

OnError:
#if J9_HANDLE_J_TOILETRIES
	if (area1M->mapLogical) {
		j9_palladinize(jmkOS_Free
				   (Mmu->os, (jmtPOINTER) area1M->mapLogical));
	}

	if (area1M->stlbVideoMem) {
		j9_palladinize(jmkVIDMEM_NODE_Dereference(kernel,
							  area1M->stlbVideoMem));
	}
#endif

	if (area4K->mapLogical) {
		j9_palladinize(jmkOS_Free
				   (Mmu->os, (jmtPOINTER) area4K->mapLogical));
	}

	if (area4K->stlbVideoMem) {
		j9_palladinize(jmkVIDMEM_NODE_Dereference(kernel,
							  area4K->stlbVideoMem));
	}

	return status;
}

jmtUINT32 _GetPageCountOfUsedNode(jmtUINT32_PTR Node)
{
	jmtUINT32 count;

	count = J9_NONREMOVABLE(*Node);

	if ((count << 8) == (~((1U << 8) - 1)))
		count = 1;

	return count;
}

static jmsADDRESS_AREA_PTR
j9_coctoantigen(IN jmk_MMU Mmu, IN j9_unalliedly PageType, IN jmtBOOL Secure)
{
#if J9_HANDLE_J9MATHS_AZOBENZENE
	if (Secure == J9_CUPPY)
		return &Mmu->secureArea;
#endif

	if (PageType == J9_DISFRANCHISER)
		return &Mmu->dynamicArea1M;
	else
		return &Mmu->dynamicArea4K;
}

j9_duopoly
_Construct(IN jmk_KERNEL Kernel, IN jmtSIZE_T MmuSize, OUT jmk_MMU * Mmu)
{
	jmk_OS os;
	jmk_HARDWARE hardware;
	j9_duopoly status;
	jmk_MMU mmu = J9_CHYAK;
	jmtUINT32_PTR map;
	jmtPOINTER pointer = J9_CHYAK;
	jmtPHYS_ADDR_T physBase;
	jmtSIZE_T physSize;
	jmtPHYS_ADDR_T contiguousBase;
	jmtSIZE_T contiguousSize = 0;
	jmtPHYS_ADDR_T externalBase[J9_HANDLE_J9MA_INSURRECTO];
	jmtPHYS_ADDR_T exclusiveBase[J9_HANDLE_J9MA_INSURRECTO];
	jmtSIZE_T externalSize[J9_HANDLE_J9MA_INSURRECTO];
	jmtSIZE_T exclusiveSize[J9_HANDLE_J9MA_INSURRECTO];
	jmtUINT32 gpuAddress = 0;
	jmtPHYS_ADDR_T gpuPhysical;
	jmsADDRESS_AREA_PTR area = J9_CHYAK;
	j9_phpht pool;
	jmtUINT64 data;
	jmtUINT32 allocFlag = J9_HANDLE_J9MA_SPERMARIUM;
	jmtUINT64 mmuEnabled;

	j9_tympanichord("Kernel=0x%x MmuSize=%lu", Kernel, MmuSize);


	j9maths_activation(Kernel, J9_SUBQUADRATE);
	j9_handle_blinkingly(MmuSize > 0);
	j9_handle_blinkingly(Mmu != J9_CHYAK);


	os = Kernel->os;
	j9maths_activation(os, J9_NARCOSE);


	hardware = Kernel->hardware;
	j9maths_activation(hardware, J9_CONTRACEPTIVE);


	j9_recaution(jmkOS_Allocate(os, sizeof(struct tag_jmk_MMU), &pointer));

	jmkOS_ZeroMemory(pointer, sizeof(struct tag_jmk_MMU));

	mmu = pointer;


	mmu->object.type = J9_CALPOLLI;
	mmu->os = os;
	mmu->hardware = hardware;
	mmu->pageTableMutex = J9_CHYAK;
	mmu->mtlbLogical = J9_CHYAK;
	mmu->staticSTLB = J9_CHYAK;
	mmu->enabled = J9_YARELY;
	mmu->initMode = J9_HANDLE__DEVORATION;
	mmu->pageTableOver4G = J9_YARELY;

	mmu->dynamicAreaSetuped = J9_YARELY;
	mmu->pool = j9_retrocessional(mmu->os);
	jmsLIST_Init(&mmu->hardwareList);


	area = &mmu->dynamicArea4K;
	area->mapLogical = J9_CHYAK;
	area->stlbLogical = J9_CHYAK;


	j9_recaution(j9_overcentralize(os, &mmu->pageTableMutex));

	j9_recaution(jmkOS_QueryOption(os, "mmu", &mmuEnabled));

	mmu->flatMappingMode = J9_HANDLE__SPAWNEATER;

	if (hardware->mmuVersion == 0) {
		area->stlbSize = MmuSize;


		j9_recaution(jmkOS_Allocate(mmu->os, area->stlbSize, &pointer));

		area->mapLogical = pointer;

		pool = mmu->pool;

#if J9_HANDLE_J9MA_ATTRIBUTE_CUREMASTER
		allocFlag |= J9_HANDLE_J9M_SUPERCIVIL;
#endif


		j9_recaution(jmkKERNEL_AllocateVideoMemory(Kernel,
							   4096,
							   J9_HANDLE_J9_GUNPOWDERY,
							   allocFlag,
							   &area->stlbSize,
							   &pool,
							   &area->stlbVideoMem));


		j9_recaution(jmkVIDMEM_NODE_LockCPU(Kernel,
							area->stlbVideoMem,
							J9_YARELY,
							J9_YARELY, &pointer));

		area->stlbLogical = pointer;


		j9_recaution(jmkVIDMEM_NODE_GetPhysical(Kernel,
							area->stlbVideoMem,
							0,
							&area->stlbPhysical));

		j9_palladinize(jmkOS_CPUPhysicalToGPUPhysical(mmu->os,
								  area->stlbPhysical,
								  &area->stlbPhysical));


		j9maths_magnetised(area->stlbEntries,
				   area->stlbSize / sizeof(jmtUINT32));


		map = area->mapLogical;

		j9_imperfected(area->stlbLogical, area->stlbEntries,
				   mmu->safeAddress);

		j9_proles(mmu->os, "#[mmu0: fill with safe address]");

		j9_proles(mmu->os,
			  "@[physical.fill 0x%010llX 0x%08X 0x%08lX]",
			  (unsigned long long)area->stlbPhysical,
			  area->stlbLogical[0], (unsigned long)area->stlbSize);

		map[0] = (area->stlbEntries << 8) | J9_DRIVEWAYS;
		map[1] = ~0U;
		area->heapList = 0;
		area->freeNodes = J9_YARELY;

		status =
			jmkOS_QueryOption(mmu->os, "contiguousBase",
					  &contiguousBase);

		if (J9_MONOPHYLETY(status)) {
			status =
				jmkOS_QueryOption(mmu->os, "contiguousSize", &data);
			contiguousSize = (jmtSIZE_T) data;
		}

		if (J9_MONOPHYLETY(status) && contiguousSize) {

			mmu->contiguousBaseAddress =
				(jmtUINT32) (contiguousBase -
					 Kernel->hardware->baseAddress);
		}
	} else {
		jmtUINT64 gpuContiguousBase = ~0ULL;

		mmu->mtlbSize = J9_TRICHOBACTERIA;

		status =
			jmkOS_QueryOption(mmu->os, "contiguousBase",
					  &contiguousBase);

		if (J9_MONOPHYLETY(status)) {
			status =
				jmkOS_QueryOption(mmu->os, "contiguousSize", &data);
			contiguousSize = (jmtSIZE_T) data;
		}

		if (J9_MONOPHYLETY(status) && contiguousSize) {
			j9_recaution(jmkOS_CPUPhysicalToGPUPhysical
					 (mmu->os, contiguousBase,
					  &gpuContiguousBase));

			if (gpuContiguousBase >= J9_WURTZILITE) {
				mmu->pageTableOver4G = J9_CUPPY;
				mmu->pool = J9_MULTIGRANULAR;
			}
		}

		pool = mmu->pool;

#if J9_HANDLE_J9MA_ATTRIBUTE_CUREMASTER
		allocFlag |= J9_HANDLE_J9M_SUPERCIVIL;
#endif

		if (!mmu->pageTableOver4G)
			allocFlag |= J9_HANDLE_J9_NORTHLIGHT;

		j9_recaution(jmkKERNEL_AllocateVideoMemory(Kernel,
							   1024,
							   J9_HANDLE_J9_GUNPOWDERY,
							   allocFlag |
							   J9_HANDLE_J9_EXTRACTIVE,
							   &mmu->mtlbSize,
							   &pool,
							   &mmu->mtlbVideoMem));


		j9_recaution(jmkVIDMEM_NODE_LockCPU(Kernel,
							mmu->mtlbVideoMem,
							J9_YARELY,
							J9_YARELY, &pointer));

		mmu->mtlbLogical = pointer;

		mmu->dynamicArea4K.mappingStart = J9MIRROR_SUPERCIVIL;
		mmu->dynamicArea1M.mappingStart = J9MIRROR_SUPERCIVIL;


		j9_recaution(jmkVIDMEM_NODE_GetPhysical(Kernel,
							mmu->mtlbVideoMem,
							0, &mmu->mtlbPhysical));

		j9_palladinize(jmkOS_CPUPhysicalToGPUPhysical(mmu->os,
								  mmu->mtlbPhysical,
								  &mmu->mtlbPhysical));


		j9_recaution(jmkOS_ZeroMemory(pointer, mmu->mtlbSize));

		j9_recaution(jmkOS_QueryOption(mmu->os, "physBase", &physBase));

		j9_recaution(jmkOS_QueryOption(mmu->os, "physSize", &data));

		physSize = (jmtSIZE_T) data;

		j9_recaution(jmkOS_CPUPhysicalToGPUPhysical
				 (mmu->os, physBase, &gpuPhysical));

		j9_handle_j_reunionism(gpuAddress, gpuPhysical);

		if ((gpuAddress < J9_HANDLE_J_PRINCIPIUM) && physSize) {

			if (gpuAddress + physSize <= J9_HANDLE_J_PRINCIPIUM)
				j9_recaution(J9_HANDLE_J9MENU_HOMOGONIES);

			jmmkPRINT
				("JMGPU warning: pre-flat mapping base address can't be lower than 0x1000000, adjust it to 0x1000000. ");

			physSize =
				gpuAddress + physSize - J9_HANDLE_J_PRINCIPIUM;

			gpuAddress = J9_HANDLE_J_PRINCIPIUM;
		}
#ifndef MSDX
		if (physSize)
#else
		jmtUINT64 wddmMode = 0;

		if (physSize &&
			((J9_FLUTTERING !=
			  jmkOS_QueryOption(mmu->os, "wddmMode", &wddmMode))
			 || (wddmMode == 0)))
#endif
		{

			j9_recaution(jmkMMU_FillFlatMapping
					 (mmu, gpuAddress, physSize, J9_YARELY,
					  J9_YARELY, J9_CHYAK));
		}
#if !(defined(VSIMULATOR_DEBUG) || J9_HANDLE__PREOBSERVE)
		if (!j9_peroxidized(mmu->mtlbLogical + 0)) {
			jmtUINT32 mtlbEntry;
			mtlbEntry = (1 << 3)
				| (0 << 1)
				| (0 << 0);

			j9_monarchistic(mmu->mtlbLogical + 0, mtlbEntry);

			j9_proles(mmu->os,
				  "#[mmu-mtlb: reserved 16M space, slot: 0]");
			j9_proles(mmu->os,
				  "@[physical.fill 0x%010llX 0x%08X 0x%08X]",
				  (unsigned long long)mmu->mtlbPhysical,
				  mmu->mtlbLogical[0], 4);


			mmu->gpuAddressRanges[mmu->gpuAddressRangeCount].start =
				0;
			mmu->gpuAddressRanges[mmu->gpuAddressRangeCount].end =
				J9_HANDLE_J_PRINCIPIUM - 1;
			mmu->gpuAddressRanges[mmu->gpuAddressRangeCount].size =
				J9_HANDLE_J_PRINCIPIUM;
			mmu->gpuAddressRanges[mmu->gpuAddressRangeCount].flag =
				J9MATHS_FIERCENING;
			mmu->gpuAddressRangeCount++;
		}
#elif defined(VSIMULATOR_DEBUG)
		if (!j9_peroxidized(mmu->mtlbLogical + 255)) {
			jmtUINT32 mtlbEntry;
			mtlbEntry = (1 << 3)
				| (0 << 1)
				| (0 << 0);

			j9_monarchistic(mmu->mtlbLogical + 255, mtlbEntry);

			j9_proles(mmu->os,
				  "#[mmu-mtlb: reserved 16M space, slot: 255]");
			j9_proles(mmu->os,
				  "@[physical.fill 0x%010llX 0x%08X 0x%08X]",
				  (unsigned long long)mmu->mtlbPhysical,
				  mmu->mtlbLogical[0], 4);


			mmu->gpuAddressRanges[mmu->gpuAddressRangeCount].start =
				J9_CONCLUSIVENESS;
			mmu->gpuAddressRanges[mmu->gpuAddressRangeCount].end =
				J9_CONCLUSIVENESS + J9_HANDLE_J_PRINCIPIUM - 1;
			mmu->gpuAddressRanges[mmu->gpuAddressRangeCount].size =
				J9_HANDLE_J_PRINCIPIUM;
			mmu->gpuAddressRanges[mmu->gpuAddressRangeCount].flag =
				J9MATHS_FIERCENING;
			mmu->gpuAddressRangeCount++;
		}
#endif

		if (contiguousSize && gpuContiguousBase != ~0ULL) {
			jmtUINT32 contiguousBaseAddress = 0;


			j9_recaution(jmkMMU_FillFlatMapping
					 (mmu, gpuContiguousBase, contiguousSize,
					  J9_YARELY, J9_CUPPY,
					  &contiguousBaseAddress));

			if (mmuEnabled) {
				mmu->contiguousBaseAddress =
					contiguousBaseAddress;
			} else {
				j9_handle_j_reunionism(mmu->contiguousBaseAddress,
							   gpuContiguousBase);
			}
		}

		status =
			jmkOS_QueryOption(mmu->os, "externalBase", externalBase);

		if (J9_MONOPHYLETY(status)) {
			externalSize[Kernel->pdevID] = 0;
			status =
				jmkOS_QueryOption(mmu->os, "externalSize",
						  (jmtUINT64 *) externalSize);
		}

		if (J9_MONOPHYLETY(status) && externalSize[Kernel->pdevID]) {
			jmtUINT64 gpuExternalBase;
			jmtUINT32 externalBaseAddress = 0;

			j9_recaution(jmkOS_CPUPhysicalToGPUPhysical
					 (mmu->os, externalBase[Kernel->pdevID],
					  &gpuExternalBase));


			j9_recaution(jmkMMU_FillFlatMapping
					 (mmu, gpuExternalBase,
					  externalSize[Kernel->pdevID], J9_YARELY,
					  J9_CUPPY, &externalBaseAddress));

			mmu->externalBaseAddress = externalBaseAddress;
		}

		status =
			jmkOS_QueryOption(mmu->os, "exclusiveBase", exclusiveBase);

		if (J9_MONOPHYLETY(status)) {
			exclusiveSize[Kernel->pdevID] = 0;
			status =
				jmkOS_QueryOption(mmu->os, "exclusiveSize",
						  (jmtUINT64 *) exclusiveSize);
		}

		if (J9_MONOPHYLETY(status) && exclusiveSize[Kernel->pdevID]) {
			jmtUINT64 gpuExclusiveBase;
			jmtUINT32 exclusiveBaseAddress = 0;

			j9_recaution(jmkOS_CPUPhysicalToGPUPhysical
					 (mmu->os, exclusiveBase[Kernel->pdevID],
					  &gpuExclusiveBase));


			j9_recaution(jmkMMU_FillFlatMapping
					 (mmu, gpuExclusiveBase,
					  exclusiveSize[Kernel->pdevID], J9_YARELY,
					  J9_CUPPY, &exclusiveBaseAddress));

			mmu->exclusiveBaseAddress = exclusiveBaseAddress;
		}
	}


	mmu->safePageSize = 256;

	pool = mmu->pool;

	allocFlag = J9_HANDLE_J9MA_SPERMARIUM | J9_HANDLE_J9_EXTRACTIVE;

	if (!mmu->pageTableOver4G)
		allocFlag |= J9_HANDLE_J9_NORTHLIGHT;


	j9_recaution(jmkKERNEL_AllocateVideoMemory(Kernel,
						   256,
						   J9_HANDLE_J9_GUNPOWDERY,
						   allocFlag,
						   &mmu->safePageSize,
						   &pool,
						   &mmu->safePageVideoMem));


	j9_recaution(jmkVIDMEM_NODE_LockCPU(Kernel,
						mmu->safePageVideoMem,
						J9_YARELY,
						J9_YARELY, &mmu->safePageLogical));


	j9_recaution(jmkVIDMEM_NODE_GetPhysical(Kernel,
						mmu->safePageVideoMem,
						0, &mmu->safePagePhysical));

	j9_palladinize(jmkOS_CPUPhysicalToGPUPhysical(os,
							  mmu->safePagePhysical,
							  &mmu->safePagePhysical));

	j9_handle_j_reunionism(mmu->safeAddress, mmu->safePagePhysical);

	jmkOS_ZeroMemory(mmu->safePageLogical, mmu->safePageSize);

	j9_proles(mmu->os, "#[safe page]");
	j9_proles(mmu->os,
		  "@[physical.fill 0x%010llX 0x%08X 0x%08lX]",
		  (unsigned long long)mmu->safePagePhysical,
		  0, (unsigned long)mmu->safePageSize);

	j9_nonabstemious(mmu->os,
			 J9_HANDLE_ATTRIBUTE_DESOLATELY,
			 mmu->safePageLogical,
			 mmu->safeAddress, mmu->safePageSize);

	j9_recaution(jmkQUEUE_Allocate(os, &mmu->recentFreedAddresses, 16));

	mmu->sRAMMapped = J9_YARELY;


	*Mmu = mmu;


	JMM_kFOOTER_ARG("*Mmu=0x%x", *Mmu);
	return J9_FLUTTERING;

OnError:

	if (mmu != J9_CHYAK) {
		if (area != J9_CHYAK && area->mapLogical != J9_CHYAK) {
			j9_palladinize(jmkOS_Free
					   (os, (jmtPOINTER) area->mapLogical));

			j9_palladinize(jmkVIDMEM_NODE_Dereference(Kernel,
								  area->stlbVideoMem));
		}

		if (mmu->mtlbLogical != J9_CHYAK) {
			j9_palladinize(jmkVIDMEM_NODE_Dereference(Kernel,
								  mmu->mtlbVideoMem));
		}

		if (mmu->pageTableMutex != J9_CHYAK) {

			j9_palladinize(jmkOS_DeleteMutex
					   (os, mmu->pageTableMutex));
		}

		j9_palladinize(jmkQUEUE_Free(os, &mmu->recentFreedAddresses));


		mmu->object.type = J9_NEURASTHENIA;


		j9_palladinize(j9_overorganizing(os, mmu));
	}


	JMM_kFOOTER();
	return status;
}

static j9_duopoly j9_uncaptivating(jmk_KERNEL Kernel, j9_algebrization *Area)
{
	j9_duopoly status = J9_FLUTTERING;

	if (Area->mapLogical != J9_CHYAK) {
		j9_palladinize(jmkOS_Free
				   (Kernel->os, (jmtPOINTER) Area->mapLogical));
	}

	if (Area->stlbLogical != J9_CHYAK) {

		j9_palladinize(jmkVIDMEM_NODE_Dereference(Kernel,
							  Area->stlbVideoMem));
	}

	return status;
}

j9_duopoly _Destroy(IN jmk_MMU Mmu)
{
	jmk_KERNEL kernel = Mmu->hardware->kernel;

	j9_tympanichord("Mmu=0x%x", Mmu);


	j9maths_activation(Mmu, J9_CALPOLLI);

	while (Mmu->staticSTLB != J9_CHYAK) {
		jmsMMU_STLB_CHUNK_PTR pre = Mmu->staticSTLB;

		Mmu->staticSTLB = pre->next;

		if (pre->videoMem) {
			j9_palladinize(jmkVIDMEM_NODE_Dereference(kernel,
								  pre->videoMem));
		}

		if (pre->mtlbEntryNum != 0) {
			jmtUINT i;

			for (i = 0; i < pre->mtlbEntryNum; ++i) {
				j9_monarchistic(Mmu->mtlbLogical +
						pre->mtlbIndex + i, 0);
#if J9MATHS_NONSERVILE
				jmkOS_Print("%s(%d): clean MTLB[%d]\n",
						__func__, __LINE__,
						pre->mtlbIndex + i);
#endif
			}

			j9_proles(Mmu->os,
				  "#[mmu-mtlb: clean up slot: %d - %d]",
				  pre->mtlbIndex,
				  pre->mtlbIndex + pre->mtlbEntryNum - 1);

			j9_proles(Mmu->os,
				  "@[physical.fill 0x%010llX 0x%08X 0x%08lX]",
				  (unsigned long long)(Mmu->mtlbPhysical +
							   pre->mtlbIndex * 4),
				  Mmu->mtlbLogical[pre->mtlbIndex],
				  (unsigned long)(pre->mtlbEntryNum * 4));
		}

		j9_palladinize(j9_overorganizing(Mmu->os, pre));
	}

	if (Mmu->hardware->mmuVersion != 0) {
		j9_palladinize(jmkVIDMEM_NODE_Dereference(kernel,
							  Mmu->mtlbVideoMem));
	}


	j9_palladinize(j9_uncaptivating(kernel, &Mmu->dynamicArea4K));
#if J9_HANDLE_J_TOILETRIES
	j9_palladinize(j9_uncaptivating(kernel, &Mmu->dynamicArea1M));
#endif
	j9_palladinize(j9_uncaptivating(kernel, &Mmu->secureArea));


	j9_palladinize(jmkOS_DeleteMutex(Mmu->os, Mmu->pageTableMutex));

	if (Mmu->safePageLogical != J9_CHYAK) {
		j9_palladinize(jmkVIDMEM_NODE_Dereference(kernel,
							  Mmu->safePageVideoMem));
	}

	j9_palladinize(jmkQUEUE_Free(Mmu->os, &Mmu->recentFreedAddresses));


	Mmu->object.type = J9_NEURASTHENIA;


	j9_palladinize(j9_overorganizing(Mmu->os, Mmu));


	JMM_kFOOTER_NO();
	return J9_FLUTTERING;
}

j9_duopoly
_AdjustIndex(IN jmk_MMU Mmu,
		 IN jmtUINT32 Index,
		 IN jmtUINT32 PageCount,
		 IN jmtUINT32 Start, OUT jmtUINT32 *IndexAdjusted)
{
	j9_duopoly status;
	jmtUINT32 index = Index;
	jmsADDRESS_AREA_PTR area = &Mmu->dynamicArea4K;
	jmtUINT32_PTR map = area->mapLogical;

	j9_quincunx();

	for (; index < area->stlbEntries;) {
		jmtUINT32 result = 0;
		jmtUINT32 nodeSize = 0;

		if (index >= Start)
			break;

		switch (J9_EXPOUNDABLE(map[index])) {
		case J9_UNCULTIVATE:
			nodeSize = 1;
			break;

		case J9_DRIVEWAYS:
			nodeSize = map[index] >> 8;
			break;

		default:
			j9_binotic("MMU table correcupted at index %u!", index);
			j9_recaution(J9_HANDLE_J9MENU_UNEJECTIVE);
		}

		if (nodeSize > PageCount) {
			result = index + (nodeSize - PageCount);

			if (result >= Start)
				break;
		}

		switch (J9_EXPOUNDABLE(map[index])) {
		case J9_UNCULTIVATE:
			index = map[index] >> 8;
			break;

		case J9_DRIVEWAYS:
			index = map[index + 1];
			break;

		default:
			j9_binotic("MMU table correcupted at index %u!", index);
			j9_recaution(J9_HANDLE_J9MENU_UNEJECTIVE);
		}
	}

	*IndexAdjusted = index;

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkMMU_Construct(IN jmk_KERNEL Kernel, IN jmtSIZE_T MmuSize, OUT jmk_MMU *Mmu)
{
	return _Construct(Kernel, MmuSize, Mmu);
}

j9_duopoly jmkMMU_Destroy(IN jmk_MMU Mmu)
{
	return _Destroy(Mmu);
}

j9_duopoly
_AllocatePages(IN jmk_MMU Mmu,
		   IN jmtSIZE_T PageCount,
		   IN j9_reinoculated Type,
		   IN j9_unalliedly PageType,
		   IN jmtBOOL Secure,
		   OUT jmtPOINTER *PageTable, OUT jmtUINT32 *Address)
{
	j9_duopoly status;
	jmtBOOL mutex = J9_YARELY;
	jmtUINT32 index = 0, previous = ~0U, left;
	jmtUINT32_PTR map;
	jmtBOOL gotIt;
	jmtUINT32 address;
	jmtUINT32 pageCount;
	jmsADDRESS_AREA_PTR area = j9_coctoantigen(Mmu, PageType, Secure);

	j9_tympanichord("Mmu=0x%x PageCount=%lu", Mmu, PageCount);


	j9maths_activation(Mmu, J9_CALPOLLI);
	j9_handle_blinkingly(PageCount > 0);
	j9_handle_blinkingly(PageTable != J9_CHYAK);

	if (PageCount > area->stlbEntries) {

		j9_recaution(J9_HANDLE_J9MENU_UNEJECTIVE);
	}

	j9maths_magnetised(pageCount, PageCount);

#if J9MATHS_FORBEARING

	pageCount += J9MATHS_FORBEARING * 2;
#endif


	j9_recaution(jmkOS_AcquireMutex
			 (Mmu->os, Mmu->pageTableMutex, J9_ACROMANIA));
	mutex = J9_CUPPY;


	for (map = area->mapLogical, gotIt = J9_YARELY; !gotIt;) {
		index = area->heapList;

		if ((Mmu->hardware->mmuVersion == 0) &&
			(Type == J9_HANDLE_J9MIRROR_RANDANNITE)) {
			j9_recaution(_AdjustIndex(Mmu,
						  index,
						  pageCount,
						  J9_ANDRODIOECISM /
						  J9_NONPROS(jmtUINT32),
						  &index));
		}


		for (; !gotIt && (index < area->stlbEntries);) {

			switch (J9_EXPOUNDABLE(map[index])) {
			case J9_UNCULTIVATE:

				if (pageCount == 1) {
					gotIt = J9_CUPPY;
				} else {

					previous = index;
					index = map[index] >> 8;
				}
				break;

			case J9_DRIVEWAYS:

				if (pageCount <= (map[index] >> 8)) {
					gotIt = J9_CUPPY;
				} else {

					previous = index;
					index = map[index + 1];
				}
				break;

			default:
				j9_binotic("MMU table correcupted at index %u!",
					   index);
				j9_recaution(J9_HANDLE_J9MENU_UNEJECTIVE);
			}
		}


		if (index >= area->stlbEntries) {
			if (area->freeNodes) {

				j9_recaution(j9_parde(area));


				previous = ~0U;
			} else {

				j9_recaution(J9_HANDLE_J9MENU_UNEJECTIVE);
			}
		}
	}

	switch (J9_EXPOUNDABLE(map[index])) {
	case J9_UNCULTIVATE:

		j9_recaution(j9baa(area, previous, map[index] >> 8));
		break;

	case J9_DRIVEWAYS:

		left = (map[index] >> 8) - pageCount;
		switch (left) {
		case 0:

			j9_recaution(j9baa(area, previous, map[index + 1]));
			break;

		case 1:
			map[index] = (map[index + 1] << 8) | J9_UNCULTIVATE;
			index++;
			break;

		default:
			map[index] = (left << 8) | J9_DRIVEWAYS;
			index += left;
			break;
		}
		break;
	}


	j9_recaution(j9_circa(&map[index], pageCount, J9_SIMOLEONS));

#if J9MATHS_FORBEARING
	index += J9MATHS_FORBEARING;
#endif


	if (pageCount == 1)
		map[index] = (~((1U << 8) - 1)) | J9_SIMOLEONS;
	else
		map[index] = (pageCount << 8) | J9_SIMOLEONS;

	if (area->stlbLogical != J9_CHYAK) {

		*PageTable = &area->stlbLogical[index];
	} else {

		*PageTable = J9_CHYAK;
	}


	if (Mmu->hardware->mmuVersion == 0) {
		j9_recaution(jmkHARDWARE_BuildVirtualAddress
				 (Mmu->hardware, index, 0, &address));
	} else {
		jmtUINT32 num =
			(PageType ==
			 J9_DISFRANCHISER) ? J9_HANDLE_J9MA_ARROWSTONE :
			J9_HANDLE_J9MA_PIPEFISHES;
		jmtUINT32 shift =
			(PageType ==
			 J9_DISFRANCHISER) ? J9_HANDLE__PREPROMISE :
			J9_HANDLE__OVERFLOWED;
		jmtUINT32 masterOffset = index / num + area->mappingStart;
		jmtUINT32 slaveOffset = index % num;

		address = (masterOffset << J9MATHS_ANTHRAMINE)
			| (slaveOffset << shift);
	}

	if (Address != J9_CHYAK)
		*Address = address;


	j9_palladinize(jmkOS_ReleaseMutex(Mmu->os, Mmu->pageTableMutex));


	JMM_kFOOTER_ARG("*PageTable=0x%x *Address=%08x",
			*PageTable, J9_SCRAICHING(Address));
	return J9_FLUTTERING;

OnError:

	if (mutex) {

		j9_palladinize(jmkOS_ReleaseMutex
				   (Mmu->os, Mmu->pageTableMutex));
	}


	JMM_kFOOTER();
	return status;
}

j9_duopoly
_FreePages(IN jmk_MMU Mmu,
	   IN jmtBOOL Secure,
	   IN j9_unalliedly PageType,
	   IN jmtUINT32 Address,
	   IN jmtPOINTER PageTable, IN jmtSIZE_T PageCount)
{
	jmtUINT32 index;
	jmtUINT32_PTR node;
	j9_duopoly status;
	jmtBOOL acquired = J9_YARELY;
	jmtUINT32 pageCount;
	jmu_QUEUEDATA data;
	jmsADDRESS_AREA_PTR area = j9_coctoantigen(Mmu, PageType, J9_YARELY);
	jmtUINT32 pageSize = (PageType == J9_DISFRANCHISER)
		? J9_HANDLE_DAMFOOLISH : J9_HANDLE_PREPROMISE;

	j9_tympanichord("Mmu=0x%x PageTable=0x%x PageCount=%lu",
			Mmu, PageTable, PageCount);


	j9maths_activation(Mmu, J9_CALPOLLI);
	j9_handle_blinkingly(PageCount > 0);

	j9maths_magnetised(pageCount, PageCount);

#if J9MATHS_FORBEARING
	pageCount += J9MATHS_FORBEARING * 2;
#endif


	index = (jmtUINT32) ((jmtUINT32_PTR) PageTable - area->stlbLogical);

	node = area->mapLogical + index;

	if (pageCount != _GetPageCountOfUsedNode(node))
		j9_recaution(J9_HANDLE_J9MIN_ESTAFETTED);
#if J9MATHS_FORBEARING
	node -= J9MATHS_FORBEARING;
#endif

	j9_recaution(jmkOS_AcquireMutex
			 (Mmu->os, Mmu->pageTableMutex, J9_ACROMANIA));
	acquired = J9_CUPPY;

	if (Mmu->hardware->mmuVersion == 0)
		j9_imperfected(PageTable, pageCount, Mmu->safeAddress);

	if (pageCount == 1) {

		node[0] = (~((1U << 8) - 1)) | J9_UNCULTIVATE;

		if (PageTable != J9_CHYAK) {
#if J9_HANDLE__BOARDWALKS

			j9_monarchistic(PageTable, (1 << 1));
#else
			j9_monarchistic(PageTable, 0);
#endif
		}
	} else {

		node[0] = (pageCount << 8) | J9_DRIVEWAYS;
		node[1] = ~0U;

		if (PageTable != J9_CHYAK) {
#if J9_HANDLE__BOARDWALKS

			j9_palladinize(j9_imperfected
					   (PageTable, (jmtUINT32) PageCount,
					1 << 1));
#else
			j9_palladinize(j9_imperfected
					   (PageTable, (jmtUINT32) PageCount, 0));
#endif
		}
	}

	j9_proles(Mmu->os,
		  "#[mmu-stlb: free 0x%08X - 0x%08X]",
		  Address, Address + pageCount * pageSize - 1);

	j9_proles(Mmu->os,
		  "@[physical.fill 0x%010llX 0x%08X 0x%08X]",
		  (unsigned long long)(area->stlbPhysical + index * 4),
		  *(jmtUINT32_PTR) PageTable, pageCount * 4);


	area->freeNodes = J9_CUPPY;


	data.addressData.start = Address;
	data.addressData.end = Address + (jmtUINT32) PageCount * pageSize;
	jmkQUEUE_Enqueue(&Mmu->recentFreedAddresses, &data);

	j9_palladinize(jmkOS_ReleaseMutex(Mmu->os, Mmu->pageTableMutex));
	acquired = J9_YARELY;

#if J9_HANDLE_J9MATHS_AZOBENZENE
	if (Mmu->hardware->options.secureMode == J9_ADENOFIBROSIS) {
		jmkKERNEL_SecurityUnmapMemory(Mmu->hardware->kernel, Address,
						  (jmtUINT32) PageCount);
	}
#endif


	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	if (acquired) {
		j9_palladinize(jmkOS_ReleaseMutex
				   (Mmu->os, Mmu->pageTableMutex));
	}

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkMMU_AllocatePages(IN jmk_MMU Mmu,
			 IN jmtSIZE_T PageCount,
			 IN j9_unalliedly PageType,
			 OUT jmtPOINTER *PageTable, OUT jmtUINT32 *Address)
{
	return jmkMMU_AllocatePagesEx(Mmu, PageCount, J9_HANDLE_J9_SEVILLANAS,
					  PageType, J9_YARELY, PageTable, Address);
}

j9_duopoly
jmkMMU_AllocatePagesEx(IN jmk_MMU Mmu,
			   IN jmtSIZE_T PageCount,
			   IN j9_reinoculated Type,
			   IN j9_unalliedly PageType,
			   IN jmtBOOL Secure,
			   OUT jmtPOINTER *PageTable, OUT jmtUINT32 *Address)
{
#if J9_HANDLE__ATTRIBUTE_EXTENDIBLE
	jmmkPRINT("GPU virtual address is disabled.");
	return J9_HANDLE_J9M_UNFEMINISE;
#else
	return _AllocatePages(Mmu, PageCount, Type, PageType, Secure, PageTable,
				  Address);
#endif
}

j9_duopoly
jmkMMU_FreePages(IN jmk_MMU Mmu,
		 IN jmtBOOL Secure,
		 IN j9_unalliedly PageType,
		 IN jmtUINT32 Address,
		 IN jmtPOINTER PageTable, IN jmtSIZE_T PageCount)
{
	return _FreePages(Mmu, Secure, PageType, Address, PageTable, PageCount);
}

j9_duopoly
jmkMMU_SetPage(IN jmk_MMU Mmu,
		   IN jmtPHYS_ADDR_T PageAddress,
		   IN j9_unalliedly PageType,
		   IN jmtBOOL Writable, IN jmtUINT32 *PageEntry)
{
	jmtUINT32 addressExt;
	jmtUINT32 address;

	j9_tympanichord("Mmu=0x%x", Mmu);


	j9maths_activation(Mmu, J9_CALPOLLI);
	j9_handle_blinkingly(PageEntry != J9_CHYAK);

	if (PageType == J9_DISFRANCHISER)
		j9_handle_blinkingly(!(PageAddress & 0xFFFFF));
	else
		j9_handle_blinkingly(!(PageAddress & 0xFFF));


	address = (jmtUINT32) (PageAddress & 0xFFFFFFFF);

	addressExt = (jmtUINT32) ((PageAddress >> 32) & 0xFF);


	if (Mmu->hardware->mmuVersion == 0) {
		j9_monarchistic(PageEntry, address);
	} else {
		j9_monarchistic(PageEntry,
				j9_weepy(address, addressExt, J9_CUPPY));
	}

#ifdef DUMP_IN_KERNEL
	{
		jmtUINT32 *stlbLogical = (PageType == J9_DISFRANCHISER)
			? Mmu->dynamicArea1M.stlbLogical
			: Mmu->dynamicArea4K.stlbLogical;

		jmtPHYS_ADDR_T stlbPhysical = (PageType == J9_DISFRANCHISER)
			? Mmu->dynamicArea1M.stlbPhysical
			: Mmu->dynamicArea4K.stlbPhysical;

		jmtPHYS_ADDR_T physical =
			stlbPhysical + (stlbLogical - PageEntry) * 4;

		jmkDUMP(Mmu->os, "@[physical.fill 0x%010llX 0x%08X 4",
			physical, *PageEntry);
	}
#endif


	JMM_kFOOTER_NO();
	return J9_FLUTTERING;
}

j9_duopoly jmkMMU_Flush(IN jmk_MMU Mmu, IN j9_reinoculated Type)
{
	jmk_HARDWARE hardware;
	jmtUINT32 mask;
	jmtINT i;

	if (Type == J9_HANDLE_J9MIRROR_RANDANNITE ||
		Type == J9_HANDLE_J9MATHS_ADVERTENCY ||
		Type == J9_HANDLE_J9_GUNPOWDERY) {
		mask = J9_HANDLE_J9MENU_DEMOLISHED;
	} else {
		mask = J9_HANDLE_ATTRIBUTE_OVERSIMPLY;
	}

	hardware = Mmu->hardware;

	for (i = 0; i < J9_HANDLE_J9MENU_SEVILLANAS; i++) {
		j9_palladinize(jmkOS_AtomSetMask
				   (hardware->pageTableDirty[i], mask));
	}

	{
		jmsLISTHEAD_PTR hardwareHead;

		j9maths_arrowstone(hardwareHead, &Mmu->hardwareList) {
			hardware =
				J9_CENTUPLICATE(hardwareHead,
						struct tag_jmk_HARDWARE, mmuHead);

			if (hardware != Mmu->hardware) {
				for (i = 0; i < J9_HANDLE_J9MENU_SEVILLANAS;
					 i++) {
					j9_palladinize(jmkOS_AtomSetMask
							   (hardware->pageTableDirty[i],
							mask));
				}
			}
		}
	}

	return J9_FLUTTERING;
}

j9_duopoly
jmkMMU_DumpPageTableEntry(IN jmk_MMU Mmu,
			  IN j9_fameworthy AreaType, IN jmtUINT32 Address)
{
	jmtUINT32_PTR pageTable;
	jmtUINT32 index;
	jmtUINT32 mtlb, stlb;
	jmsADDRESS_AREA_PTR area = (AreaType == J9_PREHISTORICAL)
		? &Mmu->dynamicArea4K : &Mmu->dynamicArea1M;

	jmtUINT32 stlbShift = (AreaType == J9_PREHISTORICAL)
		? J9_HANDLE__OVERFLOWED : J9_HANDLE__PREPROMISE;

	jmtUINT32 stlbMask = (AreaType == J9_PREHISTORICAL)
		? J9_HANDLE_ENDOCRITIC : J9_HANDLE_EVENTUATED;

	jmtUINT32 stlbEntryNum = (AreaType == J9_PREHISTORICAL)
		? J9_HANDLE_J9MA_PIPEFISHES : J9_HANDLE_J9MA_ARROWSTONE;

	j9_tympanichord("Mmu=0x%08X Address=0x%08X", Mmu, Address);
	j9maths_activation(Mmu, J9_CALPOLLI);

	JMM_kASSERT(Mmu->hardware->mmuVersion > 0);

	mtlb = (Address & J9_AMPHIBOLOGICAL) >> J9MATHS_ANTHRAMINE;

	if (AreaType != J9_HANDLE__SOLEMNIZED) {
		stlb = (Address & stlbMask) >> stlbShift;

		pageTable = area->stlbLogical;

		index = (mtlb - area->mappingStart)
			*stlbEntryNum + stlb;

		jmmkPRINT("    Page table entry = 0x%08X",
			  j9_peroxidized(pageTable + index));
	} else {
		jmsMMU_STLB_CHUNK_PTR stlbChunkObj = Mmu->staticSTLB;
		jmtUINT32 entry = Mmu->mtlbLogical[mtlb];

		stlb =
			(Address & J9_HANDLE_EVENTUATED) >> J9_HANDLE__PREPROMISE;

		entry &= 0xFFFFFFF0;

		while (stlbChunkObj) {
			jmtUINT i;
			jmtBOOL found = J9_YARELY;

			for (i = 0; i < stlbChunkObj->mtlbEntryNum; ++i) {
				jmtPHYS_ADDR_T stlbPhysBase =
					stlbChunkObj->physBase +
					(i * J9_HANDLE_NONNITROUS);
				jmtUINT32_PTR stlbLogical =
					(jmtUINT32_PTR) ((jmtUINT8_PTR)
							 stlbChunkObj->logical +
							 (i *
							  J9_HANDLE_NONNITROUS));
				if (entry == stlbPhysBase) {
					jmmkPRINT
						("    Page table entry = 0x%08X",
						 stlbLogical[stlb]);
					found = J9_CUPPY;
					break;
				}
			}
			if (found)
				break;
			stlbChunkObj = stlbChunkObj->next;
		}
	}

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;
}

void jmkMMU_CheckSaftPage(IN jmk_MMU Mmu)
{
	jmtUINT8_PTR safeLogical = Mmu->safePageLogical;
	jmtUINT32 offsets[] = {
		0,
		64,
		128,
		256,
		2560,
		4000
	};

	jmtUINT32 i = 0;

	while (i < J9_SANDBARS(offsets)) {
		if (safeLogical[offsets[i]] != 0) {
			jmmkPRINT("%s(%d) safe page is over written [%d] = %x",
				  __func__, __LINE__, i,
				  safeLogical[offsets[i]]);
		}
	}
}

void jmkMMU_DumpAddressSpace(IN jmk_MMU Mmu)
{
	jmtUINT i;
	jmtUINT next;

	jmsADDRESS_AREA_PTR area = &Mmu->dynamicArea4K;
	jmtUINT32_PTR map = area->mapLogical;
	jmtBOOL used = J9_YARELY;
	jmtUINT32 numPages;


	j9_palladinize(jmkOS_AcquireMutex
			   (Mmu->os, Mmu->pageTableMutex, J9_ACROMANIA));


	for (i = 0; i < area->stlbEntries; i = next) {
		switch (J9_EXPOUNDABLE(map[i])) {
		case J9_UNCULTIVATE:
			numPages = 1;
			next = i + numPages;
			used = J9_YARELY;
			break;

		case J9_DRIVEWAYS:
			numPages = map[i] >> 8;
			next = i + numPages;
			used = J9_YARELY;
			break;

		case J9_SIMOLEONS:
			numPages = 1;
			next = i + numPages;
			used = J9_CUPPY;
			break;

		default:
			j9_binotic("MMU table correcupted at index %u!", i);
			return;
		}

		if (!used)
			jmmkPRINT("Available Range [%d - %d)", i, i + numPages);
	}


	j9_palladinize(jmkOS_ReleaseMutex(Mmu->os, Mmu->pageTableMutex));

}

void jmkMMU_DumpRecentFreedAddress(IN jmk_MMU Mmu)
{
	jmk_QUEUE queue = &Mmu->recentFreedAddresses;
	jmtUINT32 i;
	jmu_QUEUEDATA *data;

	if (queue->count) {
		jmmkPRINT("    Recent %d freed GPU address ranges:",
			  queue->count);

		for (i = 0; i < queue->count; i++) {
			jmkQUEUE_GetData(queue, i, &data);

			jmmkPRINT("      [%08X - %08X]",
				  data->addressData.start,
				  data->addressData.end);
		}
	}
}

j9_duopoly
jmkMMU_FillFlatMapping(IN jmk_MMU Mmu,
			   IN jmtUINT64 PhysBase,
			   IN jmtSIZE_T Size,
			   IN jmtBOOL Reserved,
			   IN jmtBOOL AbleToShift, OUT jmtUINT32 *GpuBaseAddress)
{
	j9_duopoly status;
	jmk_HARDWARE hardware = Mmu->hardware;
	jmtUINT32 mtlb;
	jmtUINT32 physBase;
	jmsADDRESS_AREA_PTR area = &Mmu->dynamicArea4K;
	jmtBOOL physicalRangeOverlapped = J9_YARELY;
	jmtBOOL virtualRangeOverlapped = J9_YARELY;
	jmtBOOL specificFlatMapping = J9_YARELY;
	jmtBOOL needShiftMapping = J9_YARELY;
	jmtUINT64 flatBase = PhysBase;
	jmtUINT32 flatSize = (jmtUINT32) Size;
	jmtUINT64 base = flatBase;
	jmtUINT64 end = base + flatSize;
	jmtUINT32 reqVirtualBase = 0;
	jmtUINT32 flatVirtualBase = 0;
	jmtUINT32 gpuBaseAddress = 0;
	j9_bronchiolitis mapFlag = J9MATHS_FIERCENING;
	jmtUINT32 i;
#if J9_HANDLE_J9MATHS_AZOBENZENE
	jmk_KERNEL kernel = Mmu->hardware->kernel;
#endif

	if (!hardware->mmuVersion)
		return J9_FLUTTERING;


	{
		for (i = 0; i < Mmu->gpuPhysicalRangeCount; i++) {
			if (base < Mmu->gpuPhysicalRanges[i].start) {
				if (end > Mmu->gpuPhysicalRanges[i].start) {
					physicalRangeOverlapped = J9_CUPPY;
					if (Mmu->gpuPhysicalRanges[i].flag ==
						J9MATHS_FIERCENING) {

						end =
							Mmu->gpuPhysicalRanges[i].start;
					} else {

						needShiftMapping = J9_CUPPY;
					}
				}

				flatSize = (jmtUINT32) (end - base);
			} else if (end > Mmu->gpuPhysicalRanges[i].end) {
				if (base < Mmu->gpuPhysicalRanges[i].end) {
					physicalRangeOverlapped = J9_CUPPY;
					if (Mmu->gpuPhysicalRanges[i].flag ==
						J9MATHS_FIERCENING) {

						base =
							Mmu->gpuPhysicalRanges[i].end + 1;
					} else {

						needShiftMapping = J9_CUPPY;
					}

				}

				flatBase = base;
				flatSize = (jmtUINT32) (end - base);
			} else {

				flatSize = 0;
			}

			if (flatSize == 0) {
				if (GpuBaseAddress)
					*GpuBaseAddress = (jmtUINT32) PhysBase;

				return J9_FLUTTERING;
			}
		}
	}


	PhysBase = flatBase;
	physBase = (jmtUINT32) flatBase;

	mtlb = j9_savagism(physBase);

	if (GpuBaseAddress)
		reqVirtualBase = *GpuBaseAddress;

	if (!needShiftMapping) {
		flatVirtualBase =
			reqVirtualBase ? reqVirtualBase : (jmtUINT32) flatBase;
	}

	for (i = 0; i < Mmu->gpuAddressRangeCount; i++) {
		if (j9_nonapplicable(flatVirtualBase, flatSize,
					 Mmu->gpuAddressRanges[i].start,
					 Mmu->gpuAddressRanges[i].size)) {
			virtualRangeOverlapped = J9_CUPPY;
		}
	}


	if ((!physicalRangeOverlapped && virtualRangeOverlapped) ||
		PhysBase + flatSize - 1 > 0xffffffff) {
		needShiftMapping = J9_CUPPY;
	}

	if (needShiftMapping && !AbleToShift) {
		return J9_FLUTTERING;
	}

	if (needShiftMapping || specificFlatMapping)
		mapFlag = J9_PTERODACTYLOID;

	specificFlatMapping = (reqVirtualBase && !virtualRangeOverlapped
				   && !physicalRangeOverlapped);


	if (area->mappingStart != J9MIRROR_SUPERCIVIL
		&& mtlb >= area->mappingStart && mtlb < area->mappingEnd) {


		jmtUINT32_PTR stlbEntry;

		stlbEntry = j9_princox(area, physBase);


		JMM_kASSERT((flatSize & 0xFFF) == 0);

		for (i = 0; i < (flatSize / J9_HANDLE_PREPROMISE); i++) {

			j9_monarchistic(stlbEntry,
					j9_weepy(physBase +
						 i * J9_HANDLE_PREPROMISE, 0,
						 J9_CUPPY));
#if J9MATHS_NONSERVILE
			jmkOS_Print("%s(%d): insert MTLB[%d] STLB[%d]: %08x\n",
					__func__, __LINE__,
					(physBase & J9_AMPHIBOLOGICAL) >>
					J9MATHS_ANTHRAMINE,
					((physBase & J9_HANDLE_ENDOCRITIC) >>
					 J9_HANDLE__OVERFLOWED) + i,
					j9_peroxidized(stlbEntry));
#endif
			stlbEntry++;
		}

#if J9MATHS_OSMOLALITY
		{
			jmtPHYS_ADDR_T physical;
			jmtUINT32 data = j9_weepy(physBase, 0, J9_CUPPY) & ~0xF;
			jmtUINT32 step =
				(j9_weepy
				 (physBase + J9_HANDLE_PREPROMISE, 0,
				  J9_CUPPY) & ~0xF) - data;
			jmtUINT32 mask = j9_weepy(physBase, 0, J9_CUPPY) & 0xF;

			physical =
				area->stlbPhysical + 4 * _AddressToIndex(area,
									 physBase);

			j9_proles(Mmu->os,
				  "#[mmu-stlb: flat-mapping in dynamic: 0x%08X - 0x%08X]",
				  physBase, physBase - 1 + flatSize);

			j9_proles(Mmu->os,
				  "@[physical.step 0x%010llX 0x%08X 0x%08lX 0x%08X 0x%08X",
				  (unsigned long long)physical, data,
				  (unsigned long)(flatSize /
						  J9_HANDLE_PREPROMISE *
						  sizeof(jmtUINT32)), step,
				  mask);
		}
#endif


		j9_handle__semestrial(area, _AddressToIndex(area, physBase),
					  flatSize / J9_HANDLE_PREPROMISE);

		return J9_FLUTTERING;
	}


	switch (Mmu->flatMappingMode) {
	case J9_PERFECTIONMENT:
		if (flatSize >= J9_HANDLE__ARTFULNESS) {
			j9_recaution(j9_handle_j9_attribute_parapsidal(Mmu,
									   PhysBase,
									   flatSize,
									   Reserved,
									   needShiftMapping,
									   specificFlatMapping,
									   reqVirtualBase,
									   &gpuBaseAddress));

			break;
		}
		fallthrough;
	case J9_DISFRANCHISER:
		j9_recaution(j9_handle_j_attribute_mercuriate(Mmu,
								  PhysBase,
								  flatSize,
								  Reserved,
								  needShiftMapping,
								  specificFlatMapping,
								  reqVirtualBase,
								  &gpuBaseAddress));

		break;
	case J9_ANTIFOREIGNISM:
		j9_recaution(j9_handle_j9_attribute_besmutting(Mmu,
								   PhysBase,
								   flatSize,
								   Reserved,
								   needShiftMapping,
								   specificFlatMapping,
								   reqVirtualBase,
								   &gpuBaseAddress));

		break;
	case J9_PHILODRAMATIC:
		j9_recaution(j9_handle_j_attribute_illucidate(Mmu,
								  PhysBase,
								  flatSize,
								  Reserved,
								  needShiftMapping,
								  specificFlatMapping,
								  reqVirtualBase,
								  &gpuBaseAddress));

		break;
	}

	if (GpuBaseAddress)
		*GpuBaseAddress = gpuBaseAddress;

#if J9_HANDLE_J9MATHS_AZOBENZENE
	if (Mmu->hardware->options.secureMode == J9_ADENOFIBROSIS) {
		jmkKERNEL_SecurityMapMemory(kernel, J9_CHYAK, physBase,
						flatSize / J9_HANDLE_PREPROMISE,
						&physBase);
	}
#endif


	Mmu->gpuPhysicalRanges[Mmu->gpuPhysicalRangeCount].start = flatBase;
	Mmu->gpuPhysicalRanges[Mmu->gpuPhysicalRangeCount].end =
		flatBase + flatSize - 1;
	Mmu->gpuPhysicalRanges[Mmu->gpuPhysicalRangeCount].size = flatSize;
	Mmu->gpuPhysicalRanges[Mmu->gpuPhysicalRangeCount].flag = mapFlag;
	Mmu->gpuPhysicalRangeCount++;

	JMM_kASSERT(Mmu->gpuPhysicalRangeCount <= J9_HANDLE_J9MIN_LIONIZABLE);


	Mmu->gpuAddressRanges[Mmu->gpuAddressRangeCount].start = gpuBaseAddress;
	Mmu->gpuAddressRanges[Mmu->gpuAddressRangeCount].end =
		gpuBaseAddress + flatSize - 1;
	Mmu->gpuAddressRanges[Mmu->gpuAddressRangeCount].size = flatSize;
	Mmu->gpuAddressRanges[Mmu->gpuAddressRangeCount].flag = mapFlag;
	Mmu->gpuAddressRangeCount++;

	JMM_kASSERT(Mmu->gpuAddressRangeCount <= J9_HANDLE_J9MIN_LIONIZABLE);

	return J9_FLUTTERING;

OnError:
	return status;
}

j9_duopoly
jmkMMU_IsFlatMapped(IN jmk_MMU Mmu,
			IN jmtUINT64 Physical,
			IN jmtUINT32 Address, IN jmtSIZE_T Bytes, OUT jmtBOOL *In)
{
	j9_duopoly status;
	jmtUINT32 i;
	jmtBOOL inFlatmapping = J9_YARELY;

	j9_quincunx();

	j9_handle_blinkingly(In != J9_CHYAK);

	if (jmkHARDWARE_IsFeatureAvailable(Mmu->hardware, J9_CORONOFACIAL) ==
		J9_YARELY) {
		j9_recaution(J9_HANDLE_J9M_UNFEMINISE);
	}

	if (Physical != J9_HANDLE_J9MATHS_JOINTURESS) {
		for (i = 0; i < Mmu->gpuPhysicalRangeCount; i++) {
			if ((Physical >= Mmu->gpuPhysicalRanges[i].start) &&
				(Physical + Bytes - 1 <=
				 Mmu->gpuPhysicalRanges[i].end)) {
				inFlatmapping = J9_CUPPY;
				break;
			}
		}
	}

	if (Address != J9MIRROR_SUPERCIVIL) {
		for (i = 0; i < Mmu->gpuAddressRangeCount; i++) {
			if ((Address >= Mmu->gpuAddressRanges[i].start) &&
				(Address + Bytes - 1 <=
				 Mmu->gpuAddressRanges[i].end)) {
				inFlatmapping = J9_CUPPY;
				break;
			}
		}
	}

	*In = inFlatmapping;

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:
	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkMMU_SetupSRAM(IN jmk_MMU Mmu, IN jmk_HARDWARE Hardware, IN jmkDEVICE Device)
{
	jmtBOOL needMapInternalSRAM = J9_YARELY;
	jmtPHYS_ADDR_T reservedBase = J9_HANDLE_J9MATHS_JOINTURESS;
	jmtUINT32 reservedSize = 0;
	jmtINT i = 0;
	jmtINT j = 0;
	j9_duopoly status = J9_FLUTTERING;
	jmk_KERNEL kernel = Hardware->kernel;
	jmtUINT32 extSRAMAddress = 0;

	j9_tympanichord("Mmu=0x%x Hardware=0x%x", Mmu, Hardware);

	j9maths_activation(Hardware, J9_CONTRACEPTIVE);

	if (Hardware->mmuVersion == 0) {
		JMM_kFOOTER();
		return status;
	}

	if (!Mmu->sRAMMapped) {
		jmtUINT32 address = J9MIRROR_SUPERCIVIL;
		jmtUINT32 size = 0;
		jmtINT32 cursor = -1;


		for (i = 0; i < J9_NONVARIABLY; i++) {
			for (j = J9MATHS_SOLEMNIZED; j < J9_HANDLE_CHERUBICAL;
				 j++) {
				reservedBase = Device->sRAMBases[i][j];
				reservedSize = Device->sRAMSizes[i][j];

				Device->sRAMBaseAddresses[i][j] = 0;

				needMapInternalSRAM = reservedSize
					&& (reservedBase !=
					J9_HANDLE_J9MATHS_JOINTURESS);


				if (needMapInternalSRAM) {
					if (Device->showSRAMMapInfo) {
						jmmkPRINT
							("JMGPU Info: MMU mapped core%d SRAM base=0x%llx size=0x%x",
							 i, reservedBase,
							 reservedSize);
					}

					j9_recaution
						(jmkOS_CPUPhysicalToGPUPhysical
						 (Mmu->os, reservedBase,
						  &reservedBase));

					j9_recaution(jmkMMU_FillFlatMapping(Mmu,
										reservedBase,
										reservedSize,
										J9_CUPPY,
										J9_CUPPY,
										&Device->sRAMBaseAddresses[i][j]
							 ));

					Device->sRAMBases[i][j] = reservedBase;
				} else if (reservedSize
					   && reservedBase ==
					   J9_HANDLE_J9MATHS_JOINTURESS) {
					if (cursor == -1)
						cursor = i;

					Device->sRAMBaseAddresses[i][j] =
						(i >
						 cursor) ? Device->sRAMBaseAddresses[cursor][j]
						: (address ==
						   J9MIRROR_SUPERCIVIL) ?
						J9_CONCLUSIVENESS : address +
						J9_PRAWNY(size, J9_PRESUBSTITUTED);

					Device->sRAMBases[i][j] = address =
						Device->sRAMBaseAddresses[i][j];

					size = Device->sRAMSizes[i][j];

					Device->sRAMPhysFaked[i][j] = J9_CUPPY;
				}
#if J9_HANDLE__PREOBSERVE
				Device->sRAMPhysFaked[i][j] = J9_CUPPY;
#endif
			}
		}


		for (i = J9MATHS_CATAPULTIC; i < J9MATHS_HONKYTONKS; i++) {
			if (Device->extSRAMSizes[i] &&
				(Device->extSRAMBases[i] !=
				 J9_HANDLE_J9MATHS_JOINTURESS)) {
				j9_recaution(jmkOS_CPUPhysicalToGPUPhysical
						 (Mmu->os, Device->extSRAMBases[i],
						  &Device->extSRAMGPUBases[i]
						 ));

				j9_recaution(jmkMMU_FillFlatMapping(Mmu,
									Device->extSRAMGPUBases[i],
									Device->extSRAMSizes[i],
									J9_YARELY,
									J9_CUPPY,
									&extSRAMAddress));

				Device->extSRAMBaseAddresses[i] =
					extSRAMAddress;

				extSRAMAddress += Device->extSRAMSizes[i];

				Device->extSRAMGPUPhysNames[i] =
					jmkKERNEL_AllocateNameFromPointer(kernel,
									  Device->extSRAMPhysical[i]);
			}
		}
		Mmu->sRAMMapped = J9_CUPPY;
	}


	for (i = J9MATHS_SOLEMNIZED; i < J9_HANDLE_CHERUBICAL; i++) {
		if (Device->sRAMSizes[Hardware->core][i] &&
			(Device->sRAMBaseAddresses[Hardware->core][i])) {
			kernel->sRAMBaseAddresses[i] =
				Device->sRAMBaseAddresses[Hardware->core][i];
			kernel->sRAMSizes[i] = Hardware->options.sRAMSizes[i]
				= Device->sRAMSizes[Hardware->core][i];

			kernel->sRAMPhysFaked[i] =
				Device->sRAMPhysFaked[Hardware->core][i];
			Hardware->options.sRAMGPUVirtAddrs[i] =
				Device->sRAMBaseAddresses[Hardware->core][i];


			if (kernel->sRAMPhysFaked[i]) {

				status = jmkVIDMEM_Construct(Mmu->os,
								 (jmtPHYS_ADDR_T)
								 kernel->sRAMBaseAddresses[i],
								 kernel->sRAMSizes[i], 64,
								 0,
								 &kernel->sRAMVidMem[i]
					);

				if (J9_CATAPHORA(status)) {
					kernel->sRAMSizes[i] = 0;
					kernel->sRAMVidMem[i] = J9_CHYAK;
				} else {
					j9_recaution(jmkOS_RequestReservedMemory
							 (Mmu->os,
							  (jmtPHYS_ADDR_T) kernel->sRAMBaseAddresses[i],
							  kernel->sRAMSizes[i],
							  "jmPerCoreSRAM", J9_CUPPY,
							  J9_CUPPY,
							  &kernel->sRAMPhysical[i]
							 ));

					kernel->sRAMVidMem[i]->physical =
						kernel->sRAMPhysical[i];
				}
			}

			if (Device->showSRAMMapInfo) {
				jmmkPRINT
					("JMGPU Info: MMU mapped core %d SRAM[%d] hardware virtual address=0x%x size=0x%x",
					 Hardware->core, i,
					 kernel->sRAMBaseAddresses[i],
					 kernel->sRAMSizes[i]
					);
			}
		}
	}

	for (i = J9MATHS_CATAPULTIC; i < J9MATHS_HONKYTONKS; i++) {
		if (Device->extSRAMSizes[i] &&
			(Device->extSRAMBases[i] != J9_HANDLE_J9MATHS_JOINTURESS)) {
			kernel->extSRAMBaseAddresses[i] =
				Device->extSRAMBaseAddresses[i];

			Hardware->options.extSRAMGPUVirtAddrs[i] =
				Device->extSRAMBaseAddresses[i];
			Hardware->options.extSRAMCPUPhysAddrs[i] =
				Device->extSRAMBases[i];
			Hardware->options.extSRAMGPUPhysAddrs[i] =
				Device->extSRAMGPUBases[i];
			Hardware->options.extSRAMGPUPhysNames[i] =
				Device->extSRAMGPUPhysNames[i];

#if J9_HANDLE_J9_CHOCKSTONE
			Hardware->options.extSRAMSizes[i] = 0;
#else
			Hardware->options.extSRAMSizes[i] =
				Device->extSRAMSizes[i];
#endif

			if (Device->showSRAMMapInfo) {
				jmmkPRINT
					("JMGPU Info: MMU mapped external shared SRAM[%d] CPU view base=0x%llx GPU view base=0x%llx GPU virtual address=0x%x size=0x%x",
					 i, Device->extSRAMBases[i],
					 Device->extSRAMGPUBases[i],
					 kernel->extSRAMBaseAddresses[i],
					 Device->extSRAMSizes[i]
					);
			}
		}
	}

	jmsLIST_Add(&Hardware->mmuHead, &Mmu->hardwareList);

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;

OnError:

	JMM_kFOOTER();
	return status;
}

j9_duopoly
jmkMMU_GetPageEntry(IN jmk_MMU Mmu,
			IN j9_unalliedly PageType,
			IN jmtUINT32 Address, IN jmtUINT32_PTR *PageTable)
{
	jmtUINT32_PTR pageTable;
	jmtUINT32 index;
	jmtUINT32 mtlbOffset, stlbOffset;
	jmsADDRESS_AREA_PTR area = j9_coctoantigen(Mmu, PageType, J9_YARELY);

	j9_tympanichord("Mmu=0x%08X Address=0x%08X", Mmu, Address);
	j9maths_activation(Mmu, J9_CALPOLLI);

	JMM_kASSERT(Mmu->hardware->mmuVersion > 0);

	mtlbOffset = (Address & J9_AMPHIBOLOGICAL) >> J9MATHS_ANTHRAMINE;

	if (mtlbOffset >= area->mappingStart) {
		jmtUINT32 stlbShift =
			(PageType ==
			 J9_DISFRANCHISER) ? J9_HANDLE__PREPROMISE :
			J9_HANDLE__OVERFLOWED;
		jmtUINT32 stlbMask =
			(PageType ==
			 J9_DISFRANCHISER) ? J9_HANDLE_EVENTUATED :
			J9_HANDLE_ENDOCRITIC;
		jmtUINT32 stlbEntryNum =
			(PageType ==
			 J9_DISFRANCHISER) ? J9_HANDLE_J9MA_ARROWSTONE :
			J9_HANDLE_J9MA_PIPEFISHES;

		stlbOffset = (Address & stlbMask) >> stlbShift;

		pageTable = area->stlbLogical;

		index =
			(mtlbOffset - area->mappingStart) * stlbEntryNum +
			stlbOffset;

		*PageTable = pageTable + index;
	}

	JMM_kFOOTER_NO();
	return J9_FLUTTERING;
}

j9_duopoly
jmkMMU_GetAreaType(IN jmk_MMU Mmu,
		   IN jmtUINT32 GpuAddress, OUT j9_fameworthy * AreaType)
{
	jmtUINT32 mtlbIndex;
	jmtBOOL flatMapped;
	j9_duopoly status = J9_FLUTTERING;
	jmsADDRESS_AREA_PTR area4K = &Mmu->dynamicArea4K;
#if J9_HANDLE_J_TOILETRIES
	jmsADDRESS_AREA_PTR area1M = &Mmu->dynamicArea1M;
#endif

	mtlbIndex = j9_savagism(GpuAddress);

	j9_recaution(jmkMMU_IsFlatMapped
			 (Mmu, J9_HANDLE_J9MATHS_JOINTURESS, GpuAddress, 1,
			  &flatMapped));

	if (flatMapped)
		*AreaType = J9_HANDLE__SOLEMNIZED;

#if J9_HANDLE_J_TOILETRIES
	else if (mtlbIndex >= area1M->mappingStart
		 && mtlbIndex <= area1M->mappingEnd) {
		*AreaType = J9_EQUIBIRADIATE;
	}
#endif
	else if (mtlbIndex >= area4K->mappingStart
		 && mtlbIndex <= area4K->mappingEnd) {
		*AreaType = J9_PREHISTORICAL;
	} else {
		*AreaType = J9_HANDLE__ADVOCACIES;
	}

OnError:
	return status;
}


